Attribute VB_Name = "modulDX"
' DirectX classes and Pie Slicer rendering code.
'
Option Explicit


' Code switch for the PCX loader.
#Const DebugPCX = 0

Private Const MyLag = 0.9
Private Const UseZBuffer As Boolean = True
Private Const MyFlags As Long = DDPF_RGB Or DDPF_ALPHAPIXELS

Public Const GUID As String = "IID_IDirect3DHALDevice"

Public DoRender As Integer

Private DevCaps As D3DDEVICEDESC7

Type DxCaps
  ColorKeyBlend As Boolean
End Type

Public AllCaps As D3DDEVICEDESC7
Public MyCaps As DxCaps

' Enumeration of all formats.  Needs be called only once when
' Pie Slicer is running.
Type RenderModes
  ' Enumeration class objects.
  ' Need be called only once when Pie Slicer is running.
  TexEnum As Direct3DEnumPixelFormats
  ZEnum As Direct3DEnumPixelFormats
  
  ' Texture format used by DirectX
  TexFormatID As Integer
  TexFormat As DDPIXELFORMAT
  
  ' Z-buffer format used by DirectX
  ZFormatID As Integer
  ZFormat As DDPIXELFORMAT
End Type

Private Formats As RenderModes

Private HasZBuffer As Boolean

Private DistCamera As Single

Public LastTex As Integer

' Material associated with this PIE file.
' During rendering, the texture colors are modulated by
' the material's alpha value.  The object's overall transparency can
' be controled by adjusting the material's alpha value.
'
Private PieMtrl As D3DMATERIAL7



' Used during the RenderPie routine,
' (helps minimize SetRenderState calls).
'
'  1 = muzzle-flash type rendering was set:
'       z-buffer write protected
'       highlight type (src/dest blend = 1) alpha blending
'       x0.5 alpha modulation
'
'  0 = normal rendering was set
'
' -1 = Force a SetRenderState at start of RenderPie routine
Private Transp As Integer


' Vars to record render states during rendering.
Private Filter As Long
Private TwoSided As Long
Private FilterOn1 As Long, FilterOn2 As Long
Private FilterOff1 As Long, FilterOff2 As Long

' Vertex type used for this rendering
Private Type MyFVF
  vPosition As D3DVECTOR
  vNormal As D3DVECTOR
End Type

' DirectX compatible polygon format
Public Type DxPolygon
  Submodel As Integer ' Submodel this polygon is in (for ani)
  Texture As Integer
  
  PolyFlags As Long ' Type flags as inherited from PIE data
  MaxFrames As Integer ' # team-colors
  
  Num_Vxs As Integer   ' # points in polygon
  Vertex1() As Integer ' Pointer to first DX vertex
                       ' (one element per team color)
End Type

' Animation modes for Rendered View.
'
' Note that you cannot animate team-colors
' and submodels at the same time.
'
Public Enum AnimateModes
  Ani_Stable = 0 ' No play
  
  ' Bitflag for continuous loop play
  Ani_Loop = 1
  
  ' Bitflags indicating single-lap play modes
  Ani_PlayColors = 2
  Ani_PlaySubs = 4
  
  ' Complete status for various loop modes
  Ani_LoopColors = Ani_PlayColors Or Ani_Loop
  Ani_LoopSubs = Ani_PlaySubs Or Ani_Loop

  ' Values used to identify toolbar buttons pressed during rendering.
  ' They DON'T overlap with the other flags of this enum.
  '
  AniClass_Prev = 16
  AniClass_Next = 32
  AniClass_Play = 64
  AniClass_Loop = 128
  AniClass_TeamColors = 256
  AniClass_Submodels = 512
End Enum

' DirectX compatible PIE file format
Public Type DxPieFile
  Vertexes() As D3DVERTEX
  Polys() As DxPolygon
  Ccs() As D3DVECTOR
  
  Type As Integer
  Origin() As D3DVECTOR  '
  HorzAngle As Integer   ' Camera info for this file
  VertAngle As Integer   '
  
  FileName As String
  
  TeamColor As Integer      '
  MaxTeamColors As Integer  ' Derived from PIE data
  
  
  AnimateMode As Integer
  ' = any of:
  '   Ani_Stable (0),
  '   Ani_PlayColors, Ani_PlaySubs
  '   Ani_LoopColors, Ani_LoopSubs
  
  Anim As AnimData ' Copied directly from PIE file's ani data
  
  SubFrame As Integer
  
  TexPageName As String
  TexPage As Integer ' Indexes to DxTextures() array
  PieType As Integer ' Enumerated type of PIE file
  
  Matrix As D3DMATRIX ' Created from angles and origin
End Type


' DirectX compatible PIE model structure
'
Public Type DxModel
  MaxFrames As Integer
  Files() As DxPieFile
  PieOrigin As D3DVECTOR
  TurretH As Integer
  TurretV As Integer
  TurretRota As D3DMATRIX
  DoRender As Boolean ' Forces a render pass when True
  
End Type

Public PieModel As DxModel


' Standard DX stuff, mostly
'

Public Dx_7 As New DirectX7
Public Dx_Draw As DirectDraw7
Public Dx_3D As Direct3D7
Public Dx_Device3D As Direct3DDevice7

Public Const Page1 As Integer = 0
Public Const Page2 As Integer = 1
Public Dx_Surf(2) As DirectDrawSurface7
Public Dx_Screen As DirectDrawSurface7

Public Const MAX_TexMaps As Long = 16
Public VidMem(MAX_TexMaps) As DirectDrawSurface7

Public Dx_ZBuffer As DirectDrawSurface7

Public Dx_Texture As DirectDrawSurface7
Public Dx_Alpha As DirectDrawSurface7

Public SrcRec As RECT, MidRec As RECT, DestRec As RECT

Public Dx_RecViewPort(0) As D3DRECT

Public DoneInit As Boolean


Public Sub Init_DirectX()

Screen.MousePointer = 11
InitD3D
DoneInit = True
Screen.MousePointer = 0

End Sub

Public Sub InitD3D()
Dim ddsd As DDSURFACEDESC2

' Initialize DirectDraw.
Set Dx_Draw = Dx_7.DirectDrawCreate("")
Dx_Draw.SetCooperativeLevel formMaster.hWnd, DDSCL_NORMAL

' Initialize Direct3D.
Set Dx_3D = Dx_Draw.GetDirect3D

' Initialize list of z-buffer formats
Set Formats.ZEnum = Dx_3D.GetEnumZBufferFormats(GUID$)


' Set up the primary (display) surface
'
  ddsd.lFlags = DDSD_CAPS
  ddsd.ddsCaps.lCaps = DDSCAPS_PRIMARYSURFACE
  
  Set Dx_Screen = Dx_Draw.CreateSurface(ddsd)

  ' Do not support 256-color modes...yet?
  '
  Dx_Draw.GetDisplayMode ddsd
  Debug.Assert (ddsd.ddpfPixelFormat.lRGBBitCount > 8)

  ' Create a clipper for the display surface
  Dim pcClipper As DirectDrawClipper
     
  Set pcClipper = Dx_Draw.CreateClipper(0)
  pcClipper.SetHWnd FormMain.pic(2).hWnd
      
  Dx_Screen.SetClipper pcClipper

' Initialize the z-buffer formats
'
Sys.DirectX_ZFormat = GetZBufferFormat()

' Set up the off-screen, render surface, z-buffer
SetDrawingSurface Dx_Surf(Page1), 0

' Prepare middle surface
SetDrawingSurface Dx_Surf(Page2), 1

'On Error Resume Next
 
'If Dx_Device3D Is Nothing Then
  Set Dx_Device3D = Dx_3D.CreateDevice(GUID$, Dx_Surf(Page1))
'Else
'  Dx_Device3D.SetRenderTarget Dx_Surf(Page1)
'End If

Debug.Assert Not (Dx_Device3D Is Nothing)

' Initialize the viewport for the render-target surface
SetViewport Dx_Surf(Page1)

' Initialize the texture format enum.
Set Formats.TexEnum = Dx_Device3D.GetTextureFormatsEnum()

' Determine the texture format Pie Slicer prefers by default
Sys.DirectX_TexFormat = GetAutoTexFormat()

' Record antialiasing capabilities
Dim desc As D3DDEVICEDESC7
Dx_Device3D.GetCaps desc
Sys.CanAlias = desc.dpcTriCaps.lRasterCaps And D3DPRASTERCAPS_ANTIALIASSORTINDEPENDENT

' Record paletted texture capabilities
'   * reference only, no implementation of this yet
Dim Caps1 As DDCAPS, Caps2 As DDCAPS
Dx_Draw.GetCaps Caps1, Caps2
If Caps1.lCaps And DDCAPS_PALETTE Then
  Sys.PalMode = 2
ElseIf Caps2.lCaps And DDCAPS_PALETTE Then
  Sys.PalMode = 1
End If

End Sub




Public Sub DrawLoop()
'
' Begins Ax rendering loop for all PIE files in the current model.
'

Const NumFPS = 1024
Dim vPos As D3DVECTOR

Dim Ax As Integer
For Ax = 1 To 4
  If formLists.barMode.buttons(Ax).Value = 1 Then
    formLists.barMode.buttons(5).Tag = Ax
    Exit For
  End If
Next Ax

formLists.barMode_ButtonClick formLists.barMode.buttons(5)

' Now moved to Pie Slicer's own startup
'
' Init_DirectX

' For later reference
Dx_Device3D.GetCaps DevCaps

MyCaps.ColorKeyBlend = DevCaps.dpcTriCaps.lTextureCaps And D3DPTEXTURECAPS_COLORKEYBLEND

Dim icon As Integer

With formMaster
  If .Timer.Enabled Then
    .timeCount = 0
    .Timer_Timer
    .Timer.Enabled = False
  End If
  
  Dim GottaSave As Boolean
  GottaSave = .timeSave.Enabled
  .timeSave.Enabled = False
End With

' Prepare the PIE file for rendering.
'
Dim MaxFiles As Integer
MaxFiles = InitPie()

' This command for debug purposes only
'
'VerifyTexture Sys.VidMem(1).Surf

' List all files being rendered.
Dim List As ListItems, Item As ListItem
Set List = formLists.lstvFiles.ListItems

List.Clear
Dim Entry As String
For Ax = 1 To UBound(PieModel.Files)
  Entry = PieModel.Files(Ax).FileName
  Debug.Print PieModel.Files(Ax).MaxTeamColors, PieModel.MaxFrames
  Set Item = List.Add()
  Debug.Assert Not Item Is Nothing
  Item.Text = Mid$(Entry, InStrRev(Entry, "\") + 1)
  If Item.Text = "" Then Item.Text = "(Untitled PIE File)"
Next Ax

' List all texpages uesd for rendering.
Dim ListBox As ListBox
Set ListBox = formLists.lstTexpages
ListBox.Clear
For Ax = 0 To UBound(Sys.VidMem)
  Entry = Sys.VidMem(Ax).FileName
  If Ax = 0 Then Entry = "(No Texture)"
  
  ListBox.AddItem Mid$(Entry, InStrRev(Entry, "\") + 1)
Next Ax

Dim UsesWhat As Integer
Sys.DTCMenuCode = FormMain.DoTeamColorsMenu(-1)

For Ax = 1 To 4
  FormMain.barDX.buttons(Ax).Enabled = (Sys.DTCMenuCode > 0)
Next Ax

Dim matView As D3DMATRIX
Dim Dx_RecTemp As RECT
Dim J As Long

' Initialize all render-states for the render device.
SetRenderStates


Dim St As Single

' All this used for some testing of the TickCount() function
'
'Dim StTime As Long, St2Time As Long, EndTime As Long
'StTime = Dx_7.TickCount
'St2Time = StTime
'
'St! = Timer
'
'' One-second timing
'Do While (Timer - St) < 10
'  Do: Loop While Dx_7.TickCount = St2Time
'  St2Time = Dx_7.TickCount
'  Ax = Ax + 1
'Loop
'EndTime = Dx_7.TickCount
'Debug.Print "Tick="; (EndTime - StTime)
'Debug.Print "Timer="; (Timer - St!)
'Debug.Print "# loops="; Ax

Transp = -1
Dim Stage As Integer, LIcon As Integer
Dim ResetCount As Boolean
Stage = 1
Ax = 0

Dim ClearFlag As Long
ClearFlag = D3DCLEAR_TARGET Or IIf(HasZBuffer, D3DCLEAR_ZBUFFER, 0)

Dim picBox As PictureBox
Set picBox = FormMain.pic(2)

'
' Main draw loop.
'
PieModel.DoRender = True ' Indicate the model needs be rendered now

Do Until ViewIndex < VIEW_RENDERED
  ' Increment the rendered frame count (to calculate FPS later)
  Ax = Ax + 1
  
  ' "LIcon" tracks "icon" variable, delayed by one loop
  LIcon = icon
  icon = IIf(CLng(8 * Timer) And 1, 4, 0)
  
  ' Apply Pie Slicer's viewing angles.
  DoRender = UpdateAngles(Sys.HorzAngle, Sys.VertAngle) Or UpdateZooms(DestRec.Right - DestRec.Left, DestRec.Bottom - DestRec.Top, Sys.ZoomScalar)
    
  ' Once every tick . . . .
  If icon <> LIcon Then
    If NextTeamColor() Then DoRender = True
  End If
  
  ' Model-level render override switch.
  If PieModel.DoRender Then
    DoRender = True
    PieModel.DoRender = False
  End If
  
  ' Begin the scene, render the model
  If DoRender Then
    ' Clear the rendering surface for Ax new loop.
    On Error Resume Next
    Dim BackColor As Long
    BackColor = IIf(FormMain.mnuRenderBG(1).Checked, &H80&, 0)
    
    Dx_Device3D.Clear 1, Dx_RecViewPort(), ClearFlag, BackColor, 1, 0
    
    If Err.Number <> 0 Then
      formMaster.SetStatus icon & "|Error " & Hex$(Err.Number) & " -- Z-Buffer Malfunction", FormMain.statBar
      Dx_Device3D.Clear 1, Dx_RecViewPort(), D3DCLEAR_TARGET, 0&, 1, 0
    End If
    On Error GoTo 0
  
    Dx_Device3D.BeginScene
      ResetCount = RenderPie()
    Dx_Device3D.EndScene
    
    Screen.MousePointer = 13
  Else
    Screen.MousePointer = 0
  End If
  
  ' Draw status info
  
  J& = Dx_Surf(Page2).Blt(MidRec, Dx_Surf(Page1), SrcRec, DDBLT_WAIT)
  
  ' Draw screenshot rectangle, where applicable
  Dim w As Long, H As Long, X1 As Long, Y1 As Long
  w = FormMain.picShot.Width
  H = FormMain.picShot.Height
  X1 = (picBox.Width - w) \ 2
  Y1 = (picBox.Height - H) \ 2
  
  If Not formLists.ActiveControl Is Nothing Then
  If TypeOf formLists.ActiveControl Is TextBox Then
  If w > 1 And H > 1 Then
  
    Dx_Surf(Page2).DrawBox X1, Y1, (X1 + w - 1), (Y1 + H - 1)
    Dx_Surf(Page2).DrawText X1, Y1 + H + 1, "Screenshot region", 0
  
  End If
  End If
  End If
  
  Dx_7.GetWindowRect picBox.hWnd, DestRec
  
  With Dx_Surf(Page2)
    .SetFont picBox.Font
    .SetForeColor &HFFFF00
    
'    If Model.Files(ThisFile).FileName <> "" Then
'      Dim N As Integer
'      N = InStrRev(Model.Files(ThisFile).FileName, "\")
'      .DrawText 1, DestRec.Bottom - DestRec.Top - 45, Mid$(Model.Files(ThisFile).FileName, N + 1), False
'    End If
    
    Dim FPS As Single, Dur As Single
    Dim B As Integer
    
    ' FPS counter. Count how many frames have incremented since the last
    ' iteration.
    'Dur = Timer - St
    If icon <> LIcon Then ' every 0.125 second
      If B < 5 Then B = B + 1
      FPS = 8 * Ax 'FPS * (B - 1) / B + (4 * Ax) * (1 / B)
      Ax = 0
      St = Timer
    End If
    
'    .DrawText 1, DestRec.Bottom - DestRec.Top - 60, "Framerate = " & FPS, False
    
    
    ' Display turret rotation angles, where applicable
    Dim F As Integer
    For F = 1 To UBound(PieModel.Files)
      If PieModel.Files(F).PieType = TYPE_GNDTUR Or PieModel.Files(F).PieType = TYPE_AIRTUR Then
        .DrawText 1, MidRec.Bottom - 15, "Turret: H " & IIf(PieModel.TurretH > 180, (360 - PieModel.TurretH) & IIf(PieModel.TurretH Mod 180, "R", ""), PieModel.TurretH & IIf(PieModel.TurretH Mod 180, "L", "")) & ", V " & IIf(PieModel.TurretV > 0, "+", "") & PieModel.TurretV, 0
        Exit For
      End If
    Next F
  End With
  
    
'  J& = Dx_Screen.Blt(DestRec, Texture, Dx_RecTemp, DDBLT_WAIT)
  J& = Dx_Screen.Blt(DestRec, Dx_Surf(Page2), MidRec, DDBLT_WAIT)
  If J& <> DD_OK Then
    MsgBox "Couldn't blit to the primary surface. (" & Hex(J) & ")"
    Debug.Assert False
    Exit Do
  End If

  DoEvents
  
Loop

formMaster.timeSave.Enabled = GottaSave

' Switch back to normal toolbar on List window.
Ax = formLists.barMode.buttons(5).Tag
formLists.barMode_ButtonClick formLists.barMode.buttons(Ax)

' Round the viewing angles to the nearest 5-degree increment.
Sys.HorzAngle = Sys.HorzAngle - (Sys.HorzAngle Mod 5)
Sys.VertAngle = Sys.VertAngle - (Sys.VertAngle Mod 5)

FormMain.ShowMouseCoords VIEW_RENDERED, Sys.HorzAngle, Sys.VertAngle

End Sub

Public Function UpdateAngles(Optional ByVal AngleH As Integer = -97, Optional ByVal AngleV As Integer = -460, Optional Delay As Single = MyLag, Optional File As Integer = 0) As Integer
' Updates the camera angles.
Const Dist As Long = 500

Static LastH As Single, LastV As Single, LastT As Single

Dim Diff As Single
Dim Mat1 As D3DMATRIX, Mat2 As D3DMATRIX, Mat3 As D3DMATRIX, Mat4 As D3DMATRIX
Dim matRes As D3DMATRIX
Dim LightV As D3DVECTOR

' Allow for midnight timer reset
If Timer < LastT Then LastT = LastT - 86400

If AngleH <> -460 Then
  Diff = LastH - AngleH
  If Abs(Diff) > 300 Then Diff = Sgn(Diff) * (Abs(Diff) - 360)
  If Delay = 0 Or Abs(Diff) < 0.1 Then
    If Diff Then UpdateAngles = True
    LastH = AngleH
  Else
    UpdateAngles = True
    LastH = AngleH + (Diff) * Delay ^ ((Timer - LastT) * 100)
  End If
End If

If Abs(AngleV) <= 90 Then
  Diff = LastV - AngleV
  If Delay = 0 Or Abs(Diff) < 0.1 Then
    If Diff Then UpdateAngles = True
    LastV = AngleV
  Else
    UpdateAngles = True
    LastV = AngleV + Diff * Delay ^ ((Timer - LastT) * 100)
  End If
End If

' Apply model rotation for horizontal/vertical angles.
Dim matCc As D3DMATRIX, matCam As D3DMATRIX
' Set up the rotation point

' Apply connector & turret rotation data to the matrix
Dx_7.IdentityMatrix matCc

matCc = GetRotaMatrix(PieModel.Files(File).PieType, 1)

' Set the world matrix
'
'Dx_7.MatrixMultiply Mat1, matCc, matCam
Dx_Device3D.SetTransform D3DTRANSFORMSTATE_WORLD, matCc

' Apply viewing analges into Mat3.
Dx_7.RotateYMatrix Mat1, (LastH * m_Deg2Rad)
Dx_7.RotateXMatrix Mat2, (LastV * m_Deg2Rad)
Dx_7.MatrixMultiply Mat3, Mat1, Mat2

' Put camera offset then angle into Mat4.
'
Dim Temp As D3DVECTOR, IsVTOL As Integer
Temp = PieModel.PieOrigin
Dx_7.IdentityMatrix Mat1
Mat1.rc41 = -CamCentre.X - Temp.X
Mat1.rc42 = -CamCentre.Y - Temp.Y
Mat1.rc43 = -CamCentre.Z - Temp.Z
Dx_7.MatrixMultiply Mat4, Mat1, Mat3 ' Camera offset then angle

' Add distance offset to the Z coordinate (for the z-buffer)
'
matRes = Mat4
matRes.rc43 = 2 * DistCamera

Dx_Device3D.SetTransform D3DTRANSFORMSTATE_VIEW, matRes

LightV = Project3D(P3D_MyAngles, v_tempVector(0, 0, 1), CamCentre, -LastH, 0)
Dim LightDesc As D3DLIGHT7
Dim C1 As Single
With LightDesc
  
  C1 = Cos((LastV + 30) * m_Deg2Rad)
  .Direction.X = Sin(LastH * m_Deg2Rad) * C1
  .Direction.Y = -Sin((LastV + 30) * m_Deg2Rad)
  .Direction.Z = Cos(LastH * m_Deg2Rad) * C1
  .dltType = D3DLIGHT_DIRECTIONAL
  
  .diffuse.R = 1
  .diffuse.G = 1
  .diffuse.B = 1
  .Ambient.R = 0.2
  .Ambient.G = 0.2
  .Ambient.B = 0.2
End With

Dx_Device3D.SetLight 0, LightDesc

LastT = Timer

End Function

Public Function UpdateZooms(Optional ByVal Width As Single = 0, Optional ByVal Height As Single = 0, Optional ByVal Zoom As Single = -1, Optional ByVal Delay As Single = MyLag) As Integer
' Updates the window size and zoom factor.
Dim matProj As D3DMATRIX
Static StepZ As Single, SizeW As Single, SizeH As Single, LastT As Single

If ViewIndex < VIEW_RENDERED Then Exit Function
If LastT > Timer Then LastT = LastT - 86400

If Zoom > 0 Then
  If Delay <= 0 Then
    StepZ = Zoom
  ElseIf Delay < 1 Then
    StepZ = Zoom + (StepZ - Zoom) * Delay ^ ((Timer - LastT) * 100)
  End If
End If

If Abs(Zoom - StepZ) > 0.01 Then UpdateZooms = True

If Width > 0 Then SizeW = Width
If Height > 0 Then SizeH = Height
Dim Zn As Single, Zf As Single

Zn = -4 * DistCamera
Zf = Zn + (8 * DistCamera)

' Apply the zoom factor and scaling.
With matProj
  .rc11 = 20 * StepZ / SizeW
  .rc22 = 20 * StepZ / SizeH
  .rc33 = 1 / (Zf - Zn)
  .rc43 = -Zn * .rc33
  .rc34 = 0 '(((8 * Timer) Mod 10)) / 100
  .rc44 = 10

'  formMaster.SetStatus "3|Z values = " & Format$(-DistCamera * .rc33 + .rc43, "0.0#") & " ~ " & Format$(DistCamera * .rc33 + .rc43, "0.0#")
End With

On Error Resume Next
Dx_Device3D.SetTransform D3DTRANSFORMSTATE_PROJECTION, matProj
LastT = Timer

End Function

Private Function InitPie() As Integer
'
' Translates Pie Slicer polygon data into a directX compatible format.
'
' Returns the # of pie files being rendered.
'

Dim SinglePoly As DxPolygon

Screen.MousePointer = 13
Dim Py As Integer, PtIndx As Integer
Dim NVx As Integer, A As Integer, F As Integer
Dim Point As D3DVECTOR, Norm As D3DVECTOR
Dim Frame As Integer
Dim MaxFrames As Integer, OffsV As D3DVECTOR
Dim PtData As PointData, Polygon As PolygonData
PieModel.MaxFrames = 0
Dim DxPy As Integer

With GetRotaMatrix(Model.Files(ThisFile).PieType, 0)
  PieModel.PieOrigin.X = .rc41
  PieModel.PieOrigin.Y = .rc42
  PieModel.PieOrigin.Z = .rc43
End With

ReDim Preserve PieModel.Files(UBound(Model.Files))

' Begin copying PIE data for rendering.
For F = 1 To UBound(Model.Files)
With Model.Files(F)
  With PieModel.Files(F)
    ReDim .Polys(0)
    ReDim .Vertexes(0)
  End With
  
  ' Copy submodel animation data, if necessary
  If .Anim.AnimType = 0 Then
    PieModel.Files(F).Anim = .Anim
    PieModel.Files(F).Anim.MaxFrames = 0
  Else
    PieModel.Files(F).Anim = .Anim
  End If
  
  ' Allow use of no-texture (page 0) rendering
  If GetTexPageNum(.TexPage, "") = 0 Then
    PieModel.Files(F).TexPage = 0
  Else
    PieModel.Files(F).TexPage = .TexFile
  End If
  
  ' Assign filename & logical type for later reference
  PieModel.Files(F).FileName = .FileName
  PieModel.Files(F).PieType = .PieType
  PieModel.Files(F).MaxTeamColors = 0
  
  Dim MinL As Integer, MaxL As Integer
  MinL = 1: MaxL = 10 'UBound(.Levs)
  
  Dim SizeDist As Single
  SizeDist = getPieSize(.Levs)
  If SizeDist > DistCamera Then DistCamera = SizeDist
    
  For Py% = 1 To UBound(.Levs.Polys)
    Polygon = .Levs.Polys(Py%)
    
    With PieModel.Files(F)
      DxPy = UBound(.Polys) + 1
      ReDim Preserve .Polys(DxPy)
    
      ' Transfer polygon options
      With .Polys(DxPy)
        ' Inherit the polygon's type data
        .PolyFlags = Polygon.Flags
        .Submodel = Polygon.Submodel
        .Num_Vxs = UBound(Polygon.PtData)
    
        ' Record the max # of color variations on this polygon
        MaxFrames = IIf(Polygon.Flags And &H4000, Polygon.TeamC.Animated, 1)
        If MaxFrames < 1 Then MaxFrames = 1
        
        .MaxFrames = MaxFrames
        
        ' Initialize the vertex index ptrs
        ReDim .Vertex1(MaxFrames)
        
      End With
      
      ' Keep track of maximum team-colors in this PIE file...
      If MaxFrames > .MaxTeamColors Then .MaxTeamColors = MaxFrames
      
      ' ...and in the overall model
      If .MaxTeamColors > PieModel.MaxFrames Then PieModel.MaxFrames = .MaxTeamColors
    End With
      
    ' Loop through each team-color (or frame of animation)
    For Frame% = 1 To MaxFrames
      ' Grab this team-color
      py_getTeamOffset Polygon, Frame%, OffsV
      
      ' Assign starting vertex for this team-color
      With PieModel.Files(F)
        .Polys(DxPy).Vertex1(Frame%) = UBound(.Vertexes) + 1
      End With
      
      ' Create the polygon vertexes
      For A% = 1 To UBound(Polygon.PtData)
        PtData = Polygon.PtData(A%)
        
        Norm = py_getNormalV(.Levs, Py%, A%)
        PtIndx = PtData.Index
        Point = .Levs.Points(PtIndx).Loc
      
        With PieModel.Files(F)
          NVx = UBound(.Vertexes) + 1
          ReDim Preserve .Vertexes(NVx)
        
          With .Vertexes(NVx)
            .X = Int(Point.X + 0.5)
            .Y = Int(Point.Y + 0.5)
            .Z = Int(Point.Z + 0.5)
            .nx = Norm.X / 25 ' v_Normalize the vectors
            .ny = Norm.Y / 25
            .nz = Norm.Z / 25
            .tu = Int(PtData.Loc.X + 0.5 + OffsV.X) / 256
            .tv = Int(PtData.Loc.Y + 0.5 + OffsV.Y) / 256
          End With
        End With
        Next A%
    Next Frame%
  Next Py%

End With
Next F

Dim TempF As DxPieFile
Dim MaxF As Integer
MaxF = UBound(PieModel.Files)

For F = 1 To MaxF
  
  ' Initialize ANI data for the pie file.
  '
  InitAnims PieModel.Files(F).Anim
  
  ' Move muzzleflash-type PIE files to the end of the rendering order
  ' (special effects purposes)
  If F < MaxF And PieModel.Files(F).PieType = TYPE_FLASH Then
    If PieModel.Files(F + 1).PieType <> TYPE_FLASH Then
      TempF = PieModel.Files(F)
      For A = F To MaxF - 1
        PieModel.Files(A) = PieModel.Files(A + 1)
      Next A
      PieModel.Files(MaxF) = TempF
    End If
  End If

Next F

' Return the # of pie files being rendered.
InitPie = MaxF

' Sort two-sided polygons to the front of the array
' (not currently needed)
'
'For Py% = 1 To UBound(PieModel.Polys) - 1
'  ' If a two-sided polygon follows a one-sided polygon...
'  If Not PieModel.Polys(Py%).TwoSided And PieModel.Polys(Py% + 1).TwoSided Then
'    ' ...then move it to the front of the array.
'    For A% = Py% To 1 Step -1
'      SinglePoly = PieModel.Polys(A%)
'      PieModel.Polys(A%) = PieModel.Polys(A% + 1)
'      PieModel.Polys(A% + 1) = SinglePoly
'    Next A%
'  End If
'Next Py%

Screen.MousePointer = 0

End Function

Private Function RenderPie() As Integer
' Renders DirectX-fomatted PIE data.
'
' No return value at present.

Dim WorldMat As D3DMATRIX
Dx_Device3D.GetTransform D3DTRANSFORMSTATE_WORLD, WorldMat

Dim Cnt As Integer, ThisC As Integer
Dim NotLastFrame As Boolean

'On Error Resume Next
Dim icon As Integer
Dim St As Single
St = Timer
icon = IIf(4 * Timer <> Int(4 * Timer), 0, 4)
Dim F As Integer
    
For F = 1 To UBound(PieModel.Files)
Debug.Assert PieModel.Files(F).PieType <> TYPE_HIDE
  
  ' Special effect:  Don't z-buffer the muzzle-flash GFX.
  ' Enhances rendering of muzzle-flash GFX, but is also less true to WZ.
  ' (If not desired, set ZWRITEENABLE to True)
  '
  If PieModel.Files(F).PieType = TYPE_FLASH Then
    Dx_Device3D.SetRenderState D3DRENDERSTATE_ZWRITEENABLE, D3DZB_FALSE
  Else
    Dx_Device3D.SetRenderState D3DRENDERSTATE_ZWRITEENABLE, True
  End If
  
  Dx_Device3D.SetRenderState D3DRENDERSTATE_CULLMODE, D3DCULL_CCW
  TwoSided = 0
  
  Err.Clear
  
  ' Update the view matrixes
  Dim H As Integer, V As Integer
  getViewAngles H, V, 0
  UpdateAngles H, V, , F
  
  With PieModel.Files(F)
    
    If .PieType = TYPE_FLASH Then
      ' Make transparent.  Also note the z-buffer is write protected.
      ' (These PIE files are always last in the render order)
      If Transp <> 1 Then
        Dx_Device3D.SetRenderState D3DRENDERSTATE_DESTBLEND, D3DBLEND_ONE
        
        ' Write-protect the Z buffer (not like WZ, but improves quality of explosion GFX)
'        Dx_Device3D.SetRenderState D3DRENDERSTATE_ZWRITEENABLE, 0

        ' Set muzzle-flashes to 50% added alpha, like WZ
        PieMtrl.diffuse.A = 0.5
        Dx_Device3D.SetMaterial PieMtrl
        Transp = 1
      End If
    Else
      If Transp <> 0 Then
        ' Set solid-alpha material...
        PieMtrl.diffuse.A = 1 'AlphaV / 10
        Dx_Device3D.SetMaterial PieMtrl
        
        ' and normal alpha-blending.
        Dx_Device3D.SetRenderState D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA
        
        ' Re-activate the Z buffer
 '       Dx_Device3D.SetRenderState D3DRENDERSTATE_ZWRITEENABLE, 1
        Transp = 0
      End If
    End If
    
    Dim TexPage As Integer
    If PieModel.Files(F).TexPage <> LastTex Then
      If PieModel.Files(F).TexPage = 0 Then
        Dx_Device3D.SetTexture 0, Nothing
      Else
        On Error Resume Next
        Dx_Device3D.SetTexture 0, VidMem(Sys.VidMem(PieModel.Files(F).TexPage).SurfPtr)
      
        If Err.Number > 0 Then
          formMaster.SetStatus icon & "|Error " & Hex$(Err.Number) & " -- could not texture this model", FormMain.statBar
          Dx_Device3D.SetTexture 0, Nothing
        End If
      End If
      On Error GoTo 0
      LastTex = PieModel.Files(F).TexPage
    End If
    
    For Cnt = 1 To UBound(.Polys) 'UBound(.Polys) To 1 Step -1
      Dim Submodel As Integer
      Submodel = .Polys(Cnt).Submodel
      
      ' Grab submodel info from polygon data
      Dim SubVisi As Boolean
      If .Anim.AnimType = ANIM3DFRAMES Then
        SubVisi = (Submodel = .SubFrame)
      Else
        SubVisi = True
        If .Anim.AnimType = ANIM3DTRANS Then
        If Submodel <= UBound(.Anim.Segments) Then
          Dx_Device3D.SetTransform D3DTRANSFORMSTATE_WORLD, WorldMat
          Dx_Device3D.MultiplyTransform D3DTRANSFORMSTATE_WORLD, .Anim.Segments(Submodel).Frames(.SubFrame).Matrix
        End If
        End If
      End If
      
      ' Toggle texture filtering if necessary
      If (.Polys(Cnt).PolyFlags Xor Filter) And &H800& Then
        Filter = (.Polys(Cnt).PolyFlags And &H800&)
        Dx_Device3D.SetTextureStageState 0, D3DTSS_MAGFILTER, IIf(Filter, FilterOff1, FilterOn1)
        Dx_Device3D.SetTextureStageState 0, D3DTSS_MINFILTER, IIf(Filter, FilterOff2, FilterOn2)
      End If
    
      ' Toggle two-sided if necessary
      If (.Polys(Cnt).PolyFlags Xor TwoSided) And &H2000 Then
        TwoSided = (.Polys(Cnt).PolyFlags And &H2000&)
        Dx_Device3D.SetRenderState D3DRENDERSTATE_CULLMODE, IIf(TwoSided, D3DCULL_NONE, D3DCULL_CCW)
      End If
    
    ' Get the associated team-color for this polygon.
      ThisC = .TeamColor
        
      ' Animated polygons only:  Wrap around the team-colors value.
      If .Polys(Cnt).MaxFrames > 1 Then
        ' If specified team-color is > maximum team-color, wrap it.
        ' Would use Modulus operator, but team-color index is one-based
        ' not zero-based.  Do this instead
        '
        Do While ThisC > .Polys(Cnt).MaxFrames
          ThisC = ThisC - .Polys(Cnt).MaxFrames ' Cycle back one set
        Loop
      Else
        ThisC = 1
      End If
      
      If ThisC < 1 Then ThisC = 1
      
      Debug.Assert ThisC >= LBound(.Polys(Cnt).Vertex1)
      Debug.Assert ThisC <= UBound(.Polys(Cnt).Vertex1)
      
      'On Error Resume Next
      ' Render polygons from selected submodel only.
      If SubVisi And (.Polys(Cnt).PolyFlags And &H8200) Then
        Dx_Device3D.DrawPrimitive D3DPT_TRIANGLEFAN, D3DFVF_VERTEX, .Vertexes(.Polys(Cnt).Vertex1(ThisC)), .Polys(Cnt).Num_Vxs, D3DDP_DEFAULT
      End If
      
      If Err.Number <> 0 Then
        formMaster.SetStatus IIf(CLng(2 * Timer) And 1, 4, 0) & "|Error 0x" & Hex$(Err.Number) & " when rendering polygon " & Cnt & ".  Notify Strata for assistance.", FormMain.statBar
        Exit For
      End If
    Next Cnt
  End With

  Static LastR As Long
  Debug.Assert Err.Number = LastR
  LastR = Err.Number

Next F

End Function


Public Function DrawMatrix(ByRef Matrix As D3DMATRIX) As Integer
' Debug use only!
Debug.Print "--------"
Debug.Print Format$(Time, "h:n:s am/pm")
Debug.Print Matrix.rc11 & ",", Matrix.rc12 & ",", Matrix.rc13 & ",", Matrix.rc14
Debug.Print Matrix.rc21 & ",", Matrix.rc22 & ",", Matrix.rc23 & ",", Matrix.rc24
Debug.Print Matrix.rc31 & ",", Matrix.rc32 & ",", Matrix.rc33 & ",", Matrix.rc34
Debug.Print Matrix.rc41 & ",", Matrix.rc42 & ",", Matrix.rc43 & ",", Matrix.rc44
Debug.Print "--------"

End Function

Public Sub ChangeTexture(VidPage As DxTexture, ByVal sFile As String, Optional ByVal pFormat As Integer = 0)
'
' Changes the texture on a DirectX surface.
' May be called in or out of the rendering loop.
'
Syntax:
'
' ChangeTexture(VidPage, sFile, pFormat)
'
' VidPage - Pie Slicer vidmem page to affect
'            * Includes filename, H/W, DX handle
' sFile   - Path/filename of texture to load.
'            * May be either BMP or PCX format
' pFormat - Pixel format index to use for this surface
'            * 0 = auto (ARGB8888 or ARGB1555)


Dim ddsTexture As DirectDrawSurface7
Dim I As Long, bFound As Integer
Dim ddsd As DDSURFACEDESC2


If sFile = "" Then Debug.Assert False: Exit Sub
If Dir(sFile) = "" Then Debug.Assert False: Exit Sub

' Apply user-specified pixel format index if necessary
'   * User.TexFormat may = 0, which means "auto"
If pFormat = 0 Then pFormat = User.TexFormat

' File detection.  Not much need for anything complex.
'
Select Case LCase$(Right$(sFile, 4))
Case ".bmp" ' BMP textures
  
  ' Basic DX flag declares
  '
  ddsd.lFlags = DDSD_CAPS Or DDSD_HEIGHT Or DDSD_WIDTH Or DDSD_PIXELFORMAT Or DDSD_TEXTURESTAGE
  
  ' Apply color-keyed alphablending if requested by user
  '
  If User.UseColorKey Then ddsd.lFlags = ddsd.lFlags Or DDSD_CKSRCBLT
  
  ' Set up texture options
  '
  With ddsd
    .ddsCaps.lCaps = DDSCAPS_TEXTURE
    .ddsCaps.lCaps2 = DDSCAPS2_TEXTUREMANAGE
    .lTextureStage = 0
  End With
  
  ' Find an appropriate pixel pFormat (outputs to PieSlicer.log).
  Open App.Path & "\PieSlicer.Log" For Output As #1
  
  If pFormat > 0 Then
    bFound = pFormat
  Else
    bFound = Sys.DirectX_TexFormat
  End If
  Close #1
  
  If bFound = 0 Then
    MsgBox "Could not find ARGB 8888/1555 pixel Format for texturing this model."
    Debug.Assert False
    Exit Sub
  End If
    
  ' Create the texture surface using the pFormat index.
  Formats.TexEnum.GetItem bFound, ddsd.ddpfPixelFormat

  'ddsd.ddpfPixelFormat.lFlags = DDPF_RGB Or DDPF_PALETTEINDEXED8
  'ddsd.ddpfPixelFormat.lRGBBitCount = 8
 
  ' Load the bitmap data
  Set ddsTexture = Dx_Draw.CreateSurfaceFromFile(sFile, ddsd)
  
  ' Set texture alphas
  SetTextureAlphas ddsTexture, Sys.DirectX_TexColorGreen

Case ".pcx" ' PCX textures.

  ' Load the PCX bitmap into the surface.
  '
  Dim Code As Long
  Set ddsTexture = LoadPCXTexture(sFile, Code)
  Debug.Assert Code >= 0

End Select

' Apply the color key if hardware supports it
Dim ddCK As DDCOLORKEY
If MyCaps.ColorKeyBlend Then
 With ddCK
  .High = Sys.DirectX_TexColorGreen
  .low = Sys.DirectX_TexColorGreen
 End With
 ddsTexture.SetColorKey DDCKEY_SRCBLT, ddCK
End If

' Return the newly created texture
' and set the reference info
'
Set VidMem(VidPage.SurfPtr) = Nothing
Set VidMem(VidPage.SurfPtr) = ddsTexture

If ddsTexture Is Nothing Then Exit Sub

ddsTexture.GetSurfaceDesc ddsd
VidPage.FileName = sFile
VidPage.SurfWidth = ddsd.lWidth
VidPage.SurfHeight = ddsd.lHeight

Debug.Assert Err.Number = 0
Err.Clear

End Sub

Public Function CreateZBuffer(ByRef Source As DirectDrawSurface7)
Dim ddsd As DDSURFACEDESC2
Dim Src_sd As DDSURFACEDESC2

Dim d3dEnumPFs As Direct3DEnumPixelFormats
Dim Dx_PXLFMT As DDPIXELFORMAT

Source.GetSurfaceDesc Src_sd
Debug.Assert Src_sd.lWidth
Debug.Assert Src_sd.lHeight

' Set up a z-buffer description
With ddsd
  .lFlags = DDSD_CAPS Or DDSD_WIDTH Or DDSD_HEIGHT Or DDSD_PIXELFORMAT
   
  .ddsCaps.lCaps = DDSCAPS_ZBUFFER 'Or DDSCAPS_SYSTEMMEMORY
  If InStr(GUID$, "RGB") > 0 Then
    .ddsCaps.lCaps = .ddsCaps.lCaps Or DDSCAPS_SYSTEMMEMORY
  End If
  
  ' Make sure to use the same width & height values
  .lWidth = Src_sd.lWidth
  .lHeight = Src_sd.lHeight
  
  If Formats.ZFormatID = 0 Then
    Debug.Assert Formats.ZFormatID
    MsgBox "Internal error: No z-buffer format determined."
    CreateZBuffer = 0
    Exit Function
  End If
  
  .ddpfPixelFormat = Formats.ZFormat

'  .ddsCaps.lCaps = ddsCaps.lCaps Or DDSCAPS_ZBUFFER
End With
  
' Activate the z-buffer
Debug.Assert ddsd.lWidth
Debug.Assert ddsd.lHeight
Set Dx_ZBuffer = Dx_Draw.CreateSurface(ddsd)

' Attach the z-buffer
Source.AddAttachedSurface Dx_ZBuffer
CreateZBuffer = 1
End Function

Private Sub Misc()
' Temporary use only
End Sub

Private Sub WriteFormat(ByVal FileN As Integer, ByRef Pixels As DDPIXELFORMAT)

With Pixels

Print #FileN, "Flags = 0x" & Hex$(.lFlags)

If .lFlags And DDPF_RGB Then
  Print #FileN, "- RGB pixels"
  Print #FileN, "Red channel = 0x" & Hex$(.lRBitMask)
  Print #FileN, "Green channel = 0x" & Hex$(.lGBitMask)
  Print #FileN, "Blue channel = 0x" & Hex$(.lBBitMask)
  Print #FileN, "Alpha channel = 0x" & Hex$(.lRGBAlphaBitMask)
ElseIf .lFlags And DDPF_ZBUFFER Then
  Print #FileN, "- Z-Buffer surface"
ElseIf .lFlags And DDPF_ALPHA Then
  Print #FileN, "- Alpha surface"
  Print #FileN, "Alpha Depth = " & .lAlphaBitDepth
End If

End With

End Sub

Private Function SetTextureAlphas(ByRef ddsTexture As DirectDrawSurface7, ByVal ColorKey As Long) As Long
Dim R1 As RECT, Found As Boolean, ddsd As DDSURFACEDESC2
Dim X As Long, Y As Long, C As Long
Dim X2 As Long, Y2 As Long

' Processes the texture surface and applies alpha values to each pixel.
'

ddsTexture.GetSurfaceDesc ddsd
' R1.(Top, Left) = (0,0)
R1.Bottom = ddsd.lHeight
R1.Right = ddsd.lWidth

ddsTexture.Lock R1, ddsd, DDLOCK_WAIT, 0

      
Dim Px As Long, Alpha As Long, green As Long
Alpha = ddsd.ddpfPixelFormat.lRGBAlphaBitMask
green = ddsd.ddpfPixelFormat.lGBitMask

For Y2 = 0 To ddsd.lHeight - 1
For X2 = 0 To ddsd.lWidth - 1
    
  Px = ddsTexture.GetLockedPixel(X2, Y2)
  
  ' Set alpha=0 for the green color
  '
  If Px = green Then
    If Not User.UseColorKey Then ddsTexture.SetLockedPixel X2, Y2, green
  
  ' Adjust natural black to avoid color-keyed transparency glitch
  '
  ElseIf Px = 0 And User.UseColorKey Then
    If Alpha = &H8000& Then ' ARGB1555
      ddsTexture.SetLockedPixel X2, Y2, &H8421&
    ElseIf Alpha = &HF000& Then ' ARGB4444
      ddsTexture.SetLockedPixel X2, Y2, &HF111&
    ElseIf Alpha = &HFF000000 Then ' ARGB8888
      ddsTexture.SetLockedPixel X2, Y2, &HFF010101
    Else
      
      Debug.Assert False
    End If
 ' Anything else: apply alpha = 1
  Else
    ddsTexture.SetLockedPixel X2, Y2, (Px Or Alpha) ' Set opaque
  End If

Next X2
Next Y2

ddsTexture.Unlock R1
 
End Function

Public Function GetFormats() As String()
Dim TextureEnum As Direct3DEnumPixelFormats
Dim ddsd As DDSURFACEDESC2
Dim I As Integer, NumF As Integer
Dim Stri As String
Dim Formats() As String: ReDim Formats(0)

If Dx_Device3D Is Nothing Then InitD3D

Set TextureEnum = Dx_Device3D.GetTextureFormatsEnum()

For I = 1 To TextureEnum.GetCount()
  
  TextureEnum.GetItem I, ddsd.ddpfPixelFormat
  
  With ddsd.ddpfPixelFormat
    If (.lFlags And MyFlags) = MyFlags Then
      NumF = UBound(Formats) + 1
      ReDim Preserve Formats(NumF)
      If .lRGBAlphaBitMask = &HFF000000 Then
        Stri$ = "32-bit ARGB 8888"
      ElseIf .lRGBAlphaBitMask = &HF000& Then
        Stri$ = "16-bit ARGB 4444"
      ElseIf .lRGBAlphaBitMask = &H8000& Then
        Stri$ = "16-bit ARGB 1555"
      End If
      Formats(NumF) = I & ":" & Stri$
    End If
  End With
Next I

GetFormats = Formats()

End Function

Public Function GetRotaMatrix(ByVal PieType As Integer, ByRef Flags As Long) As D3DMATRIX
'
' Calculates the rotation/origin matrix for a PIE file.
'
Syntax:
'
' GetRotaMatrix(PieType, Flags)
'
'   PieType - Type of PIE file being evaluated
'
'   Flags   - +1 = (input) Apply turret rotation values
'             +2 = (output) Indicates an aerial component
'
Dim Dest As D3DMATRIX
Static LastTurH As Integer, LastTurV As Integer

' Lots of matrixes needed here
'
Dim Temp0 As D3DMATRIX
Dim Temp As D3DMATRIX
Dim Temp2 As D3DMATRIX
Dim TurretM As D3DMATRIX
Dx_7.IdentityMatrix Temp0
Dx_7.IdentityMatrix Temp
Dx_7.IdentityMatrix Temp2
Dx_7.IdentityMatrix TurretM


Dim FoundCc(1 To 4) As Boolean
Dim CC(1 To 4) As D3DVECTOR

Dim Failed As Boolean, F As Integer
For F = 1 To UBound(Model.Files)
With Model.Files(F).Levs

  If Model.Files(F).PieType = TYPE_BODY Then
    If UBound(.Ccs) > 0 Then
      FoundCc(1) = True
      CC(1) = .Ccs(1).Loc
      
      If UBound(.Ccs) > 1 Then
        FoundCc(2) = True
        CC(2) = .Ccs(2).Loc
      End If
      Exit For
    End If
  End If

End With
Next F

Dim HorzMat As D3DMATRIX, VertMat As D3DMATRIX
If Flags And 1 Then ' Render mode: create rotation matrixes
  Dx_7.RotateYMatrix HorzMat, PieModel.TurretH * (m_Deg2Rad)

  Dx_7.RotateXMatrix VertMat, PieModel.TurretV * (m_Deg2Rad)
Else ' Wireframe mode: ignore rotations
  Dx_7.IdentityMatrix HorzMat
  Dx_7.IdentityMatrix VertMat
End If

Select Case PieType
Case TYPE_SINGLE, TYPE_BODY, TYPE_PROP ' Single / Chassis / Prop
  ' No matrix info needed for these types
  
  GetRotaMatrix = Temp0

  Exit Function
  
Case TYPE_GNDTUR ' Ground turrets
  
  If FoundCc(1) Then
    If Flags And 1 Then
      TurretM = HorzMat
      TurretM.rc41 = TurretM.rc41 + CC(1).X
      TurretM.rc42 = TurretM.rc42 + CC(1).Y
      TurretM.rc43 = TurretM.rc43 + CC(1).Z
    Else
      TurretM.rc41 = CC(1).X
      TurretM.rc42 = CC(1).Y
      TurretM.rc43 = CC(1).Z
    End If
  Else ' If no chassis found, return zero
    GetRotaMatrix = Temp
    Exit Function
  End If
  
Case TYPE_AIRTUR ' Aircraft turrets
  Temp.rc11 = -1
  Temp.rc22 = -1
  Flags = Flags Or 2
  
  If FoundCc(2) Then
    If Flags And 1 Then
      Dx_7.MatrixMultiply TurretM, Temp, HorzMat
      TurretM.rc41 = TurretM.rc41 + CC(2).X
      TurretM.rc42 = TurretM.rc42 + CC(2).Y
      TurretM.rc43 = TurretM.rc43 + CC(2).Z
    Else
      TurretM = Temp
      TurretM.rc41 = CC(2).X
      TurretM.rc42 = CC(2).Y
      TurretM.rc43 = CC(2).Z
    End If
  Else
    TurretM = Temp
  End If
 
Case TYPE_MUZL ' Vehicle muzzle...
  For F = 1 To UBound(Model.Files)
  With Model.Files(F)
    
    Select Case .PieType
    Case TYPE_GNDTUR ' ...on tank mount
      If FoundCc(1) Then
        Temp2 = GetRotaMatrix(.PieType, Flags)
      Else
        Dx_7.IdentityMatrix Temp2
      End If
      
      ' Apply vertical rotation during render mode
      If Flags And 1 Then
        Dx_7.MatrixMultiply TurretM, VertMat, Temp2
      Else
        TurretM = Temp2
      End If
      
      F = UBound(Model.Files) ' Exits for loop
    
    Case TYPE_AIRTUR ' ...on aircraft mount
    
      If FoundCc(2) Then
        Temp2 = GetRotaMatrix(.PieType, Flags)
      Else
        Dx_7.IdentityMatrix Temp2
        Temp2.rc22 = -1
        Temp2.rc11 = -1
      End If
      
      ' Apply vertical rotation during render mode
      If Flags And 1 Then
        Dx_7.RotateXMatrix VertMat, -PieModel.TurretV * (m_Deg2Rad)
        Dx_7.MatrixMultiply TurretM, VertMat, Temp2
      Else
        TurretM = Temp2
      End If
      
      F = UBound(Model.Files) ' Exits for loop
    End Select
  End With
  Next F
  
Case TYPE_MUZLCYB ' Cyborg type turret muzzle
  '
  ' Temp is matrix for turret muzzle
  ' Temp2 is matrix for turret base

  For F = 1 To UBound(Model.Files)
  With Model.Files(F)
  
    If .PieType = TYPE_GNDTUR Or .PieType = TYPE_AIRTUR Then
      If UBound(.Levs.Ccs) > 0 Then
        FoundCc(3) = True
        CC(3) = .Levs.Ccs(1).Loc
      Else
        FoundCc(3) = True
        ' Locate about turret origin
      End If
      Temp2 = GetRotaMatrix(.PieType, Flags)  ' Get the turret base matrix
      
      F = UBound(Model.Files)
    End If
  
  End With
  Next F
  
  ' Rendered mode: Include vertical rotation
  If Flags And 1 Then
    Temp = VertMat
  Else ' Wireframe mode: no rotation
    Dx_7.IdentityMatrix Temp
  End If
  
  Temp.rc41 = Temp.rc41 + CC(3).X
  Temp.rc42 = Temp.rc42 + CC(3).Y
  Temp.rc43 = Temp.rc43 + CC(3).Z
  
  Dx_7.MatrixMultiply TurretM, Temp, Temp2

Case TYPE_FLASH ' Muzzle-flash

  For F = 1 To UBound(Model.Files)
  With Model.Files(F)
    If .PieType = TYPE_MUZL Or .PieType = TYPE_MUZLCYB Then
      If UBound(.Levs.Ccs) > 0 Then
        FoundCc(4) = True
        CC(4) = .Levs.Ccs(1).Loc
        Temp2 = GetRotaMatrix(.PieType, Flags)  ' Get the turret muzzle matrix
        Exit For
      Else
        Failed = True
      End If
    End If
  End With
  Next F
  
  ' Offset muzzle-flash graphic.
  Dx_7.IdentityMatrix Temp
  If FoundCc(4) Then
    Temp.rc41 = Temp.rc41 + CC(4).X
    Temp.rc42 = Temp.rc42 + CC(4).Y
    Temp.rc43 = Temp.rc43 + CC(4).Z
  End If
  
  Dx_7.MatrixMultiply TurretM, Temp, Temp2

Case TYPE_PROJ ' Projectile

  For F = 1 To UBound(Model.Files)
  With Model.Files(F)
    If .PieType = TYPE_MUZL Or .PieType = TYPE_MUZLCYB Then
      If UBound(.Levs.Ccs) > 0 Then
        FoundCc(4) = True
        CC(4) = .Levs.Ccs(1).Loc
        Temp2 = GetRotaMatrix(.PieType, Flags)  ' Get the turret muzzle matrix
        Exit For
      Else
        Failed = True
      End If
    End If
  End With
  Next F
  
  ' Offset muzzle-flash graphic.
  Dx_7.IdentityMatrix Temp
  If FoundCc(4) Then
    Temp.rc41 = Temp.rc41 + CC(4).X
    Temp.rc42 = Temp.rc42 + CC(4).Y
    Temp.rc43 = Temp.rc43 + CC(4).Z + 64 ' 0.5 sq.
  Else
    Temp.rc43 = 64  ' 0.5 sq.
  End If
  
  Dx_7.MatrixMultiply TurretM, Temp, Temp2
'  Debug.Assert Not Failed

End Select

GetRotaMatrix = TurretM

End Function

Public Function GetAutoTexFormat() As Integer
Dim ddsd As DDSURFACEDESC2
Dim bFound As Integer
Dim Pri As Byte, I As Integer

Debug.Assert Not (Dx_Device3D Is Nothing)

Debug.Assert Not (Formats.TexEnum Is Nothing)

For I = 1 To Formats.TexEnum.GetCount()
  Call Formats.TexEnum.GetItem(I, ddsd.ddpfPixelFormat)

'  Print #1, ""
'  Print #1, "Pixel Format #" & I
  
'  WriteFormat 1, ddsd.ddpfPixelFormat
  
  With ddsd.ddpfPixelFormat
    
    ' Prefer specific modes.  ARGB 8888 > 1555 > 4444
    If (.lFlags And MyFlags) = MyFlags Then
      If .lRGBAlphaBitMask = &HFF000000 Then
        ' ARGB 8888
        Pri = 8: bFound = I
        Sys.DirectX_TexColorGreen = .lGBitMask
        Exit For
      ElseIf .lRGBAlphaBitMask = &H8000& Then
        ' ARGB 1555
        Pri = 5: bFound = I
        Sys.DirectX_TexColorGreen = .lGBitMask
      ElseIf .lRGBAlphaBitMask = &HF000& And Pri < 5 Then
        ' ARGB 4444
        Pri = 4: bFound = I
        Sys.DirectX_TexColorGreen = .lGBitMask
      End If
    End If
  End With
Next I

GetAutoTexFormat = bFound

End Function

Private Function NextTeamColor() As Integer
' Increments the current animation status on selected PIE files.
'
' Should be called only at a regular interval
' (currently 1/8 sec)
'
' Returns True if any files were updated to a new frame,
' or zero if nothing happened.

Dim F As Integer
Dim CallSecondary As Boolean
For F = 1 To UBound(PieModel.Files)
With PieModel.Files(F)
  Dim Item As ListItem
  Set Item = formLists.lstvFiles.ListItems(F)
  
  Select Case .AnimateMode
  Case Ani_PlayColors, Ani_LoopColors
    If .MaxTeamColors > 1 Then
    If .TeamColor < .MaxTeamColors Then ' Mid loop
      .TeamColor = .TeamColor + 1
      NextTeamColor = True
    ElseIf .AnimateMode = Ani_LoopColors Then ' continuous loop
      .TeamColor = 1
      NextTeamColor = True
    Else ' single loop end
      .AnimateMode = 0
      CallSecondary = True
    End If
    End If
    
  Case Ani_PlaySubs, Ani_LoopSubs
    If .SubFrame < .Anim.MaxFrames Then ' Mid loop
      .SubFrame = .SubFrame + 1
      NextTeamColor = True
    ElseIf .AnimateMode = Ani_LoopSubs Then ' continuous loop
      .SubFrame = 1
      NextTeamColor = True
    Else ' single loop end
      .AnimateMode = 0
      CallSecondary = True
    End If
    
  Case 0 ' Stable
    
  Case Else
    Debug.Assert False
    Exit Function
  End Select

  If .MaxTeamColors > 1 Then
    If CStr(.TeamColor) <> Item.SubItems(1) Then Item.SubItems(1) = .TeamColor
  End If
  
  If .Anim.MaxFrames > 1 Then
    If CStr(.SubFrame) <> Item.SubItems(2) Then Item.SubItems(2) = .SubFrame
  End If
End With
Next F

If CallSecondary Then
  CheckPlayingStatus
End If

End Function

Public Sub CheckPlayingStatus()
'
' Checks to see which PIE files are playing/looping or not,
' and sets the buttons on formMain's DX toolbar accordingly (on/mixed/off)
'


' Local constants (bitflags).
' Keeps track of which PIE files are playing, looping, whatnot.
Const SomeStopped As Integer = 1, SomeNotStopped As Integer = 2
Const SomePlaying As Integer = 4, SomeNotPlaying As Integer = 8
Const SomeLooping As Integer = 16, SomeNotLooping As Integer = 32

' Derived constants.
'
' If ALL rendered PIE files are in the same status,
' the combined status flags will EXACTLY match one of these.
Const GotStopped As Integer = SomeStopped Or SomeNotPlaying Or SomeNotLooping
Const GotPlaying As Integer = SomePlaying Or SomeNotStopped Or SomeNotLooping
Const GotLooping As Integer = SomeLooping Or SomeNotStopped Or SomeNotPlaying

Dim File As Integer
Dim Status As Byte

' Loop through the render files
For File = 1 To UBound(PieModel.Files)
With PieModel.Files(File)
  Select Case .AnimateMode
  Case Ani_Stable ' Nothing playing on this PIE file
    If .MaxTeamColors > 1 Or .Anim.MaxFrames > 1 Then Status = Status Or GotStopped
  
  Case Ani_PlaySubs ' Submodel animation (play once)
    If .Anim.MaxFrames > 1 Then Status = Status Or GotPlaying
  
  Case Ani_LoopSubs ' Submodel animation (play/loop)
    If .Anim.MaxFrames > 1 Then Status = Status Or GotLooping
  
  Case Ani_LoopColors ' Team-colors animation (play once)
    If .MaxTeamColors > 1 Then Status = Status Or GotLooping
  
  Case Ani_PlayColors ' Team-colors animation (play/loop)
    If .MaxTeamColors > 1 Then Status = Status Or GotPlaying
  
  Case Else
    Debug.Assert False
  End Select
  
End With
Next File

Dim Button As Button

' Set the "Play Once" button status
Set Button = FormMain.barDX.buttons("Play1")

If Status = GotPlaying Then
  Button.Value = 1
  Button.MixedState = False
ElseIf Status And SomePlaying Then
  Button.MixedState = True
  Button.Value = 1
Else
  Button.Value = 0
  Button.MixedState = False
End If

' Set the "Play All" button status
Set Button = FormMain.barDX.buttons("PlayAll")
If Status = GotLooping Then
  Button.Value = 1
  Button.MixedState = False
ElseIf Status And SomeLooping Then
  Button.MixedState = True
  Button.Value = 1
Else
  Button.Value = 0
  Button.MixedState = False
End If

End Sub

Public Sub SetDrawingSurface(ByRef Source As DirectDrawSurface7, ByVal Index As Integer)
Dim DestRec As RECT
Dim ddsd As DDSURFACEDESC2

' Sets up one of Pie Slicer's DirectDraw surfaces.
'
Syntax:
'
' SetDrawingSurface(surf, index)
'
'  surf  -- Variable to store this surface in
'  index -- 0 = back buffer: 3D render target, z-buffer, optional high res
'           1 = mid buffer: used for blits & stuff only

Dx_7.GetWindowRect FormMain.pic(2).hWnd, DestRec

'If Source Is Nothing Then
  ddsd.lFlags = DDSD_HEIGHT Or DDSD_WIDTH Or DDSD_CAPS
  
  ddsd.ddsCaps.lCaps = DDSCAPS_OFFSCREENPLAIN Or IIf(Index = 0, DDSCAPS_3DDEVICE, 0)

  If Index = 0 Then
    ddsd.ddsCaps.lCaps = ddsd.ddsCaps.lCaps Or DDSCAPS_3DDEVICE
    
    ' Request antialiasing permissions for the render target
    If Sys.CanAlias Then ddsd.ddsCaps.lCaps2 = DDSCAPS2_HINTANTIALIASING
  End If

'Else
'  Source.GetSurfaceDesc ddsd
'  Debug.Assert ddsd.lFlags And (DDSD_HEIGHT Or DDSD_WIDTH Or DDSD_CAPS)
'  Debug.Assert ddsd.ddsCaps.lCaps And (DDSCAPS_OFFSCREENPLAIN Or DDSCAPS_3DDEVICE)
'End If

Dim Scalar As Integer: Scalar = 1

' Use double-resolution rendering as an alternative to antialiasing,
' only when the card does not already support antialiasing features
If Index = 0 Then
  If User.Settings And Set_2xRes Then
    If Sys.CanAlias = 0 Then Scalar = 2
  End If
End If

ddsd.lWidth = (DestRec.Right - DestRec.Left) * Scalar
ddsd.lHeight = (DestRec.Bottom - DestRec.Top) * Scalar
Debug.Assert ddsd.lWidth
Debug.Assert ddsd.lHeight

Dim Temp As DirectDrawSurface7
Set Temp = Dx_Draw.CreateSurface(ddsd)

If Not Temp Is Nothing Then
  Set Source = Temp
  Set Temp = Nothing
End If

' Cache the dimensions of the render target. We'll use
' it later for blitting operations.
Dim Rec As RECT
Rec.Left = 0
Rec.Top = 0
Rec.Right = ddsd.lWidth
Rec.Bottom = ddsd.lHeight

If Index = 0 Then
  SrcRec = Rec
  HasZBuffer = CreateZBuffer(Source)
Else
  MidRec = Rec
End If

End Sub

Public Sub SetViewport(ByRef Surf As DirectDrawSurface7)
' Set up the D3D viewport.
Dim ddsd As DDSURFACEDESC2
Dim VPDesc As D3DVIEWPORT7
 
Surf.GetSurfaceDesc ddsd

VPDesc.lWidth = ddsd.lWidth
VPDesc.lHeight = ddsd.lHeight
VPDesc.minz = 0!
VPDesc.maxz = 1!

' Activate the D3D viewport.
Dx_Device3D.SetViewport VPDesc

' Save the results for future reference
With Dx_RecViewPort(0)
  .X1 = 0: .Y1 = 0
  .X2 = VPDesc.lWidth
  .Y2 = VPDesc.lHeight
End With

End Sub


Private Function GetZBufferFormat()
Dim Dx_PXLFMT As DDPIXELFORMAT

Debug.Assert Not (Formats.ZEnum Is Nothing)
Debug.Assert Formats.ZEnum.GetCount

Dim I As Long
For I = 1 To Formats.ZEnum.GetCount()
  Call Formats.ZEnum.GetItem(I, Dx_PXLFMT)
  If Dx_PXLFMT.lFlags = DDPF_ZBUFFER Then
    Formats.ZFormat = Dx_PXLFMT
    Formats.ZFormatID = I
    GetZBufferFormat = I
    Exit Function
  End If
Next I&
  
Debug.Assert False
MsgBox "Could not determine an appropriate z-buffer format out of " & Formats.ZEnum.GetCount() & " formats available."

End Function

Public Sub SetRenderStates()
' Enable gray ambient light.
Dx_Device3D.SetRenderState D3DRENDERSTATE_AMBIENT, &H808080

' Apply solid, white material.  The alpha value will be mixed into
' the texture during rendering.
Dim Colr As D3DCOLORVALUE
Dim LightDesc As D3DLIGHT7

Colr.R = 1
Colr.G = 1
Colr.B = 1
Colr.A = 1 ' Solid material (a must-have for alpha modulation)

With PieMtrl
  PieMtrl.Ambient = Colr
  PieMtrl.diffuse = Colr
End With

Dx_Device3D.SetMaterial PieMtrl

' Assign material-controled alpha modulation
' Color = (texture alpha) * (material alpha)
Call Dx_Device3D.SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE) ' Multiply...
Call Dx_Device3D.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE) ' texture's alpha
Call Dx_Device3D.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE) ' by material's (diffuse) alpha

' Use the same color settings for all emitted light color.
Dim C As D3DCOLORVALUE
With C
  .R = 1#: .G = 1#: .B = 1#
  .A = 1# ' The alpha component isn't used for lights.
End With

With LightDesc
  .Direction.X = 0
  .Direction.Y = -0.7
  .Direction.Z = 0.8
  .dltType = D3DLIGHT_DIRECTIONAL
  
  ' Don't set alpha value for the directional lighting
  .Ambient.R = 0
  .Ambient.G = 0
  .Ambient.B = 0
  
  .diffuse.R = 1
  .diffuse.G = 1
  .diffuse.B = 1
  
  .specular.R = 0
  .specular.G = 0
  .specular.B = 0

End With

Dx_Device3D.SetLight 0, LightDesc

' Enable light #0.
Call Dx_Device3D.LightEnable(0, True)

' Apply initial angles.
UpdateAngles Sys.HorzAngle, Sys.VertAngle, 0

With FormMain.pic(2)
  UpdateZooms .ScaleWidth, .ScaleHeight, Sys.ZoomScalar, 0
End With

LastTex = -1

' Enable the z-buffer.
With Dx_Device3D

  If HasZBuffer Then
    .SetRenderState D3DRENDERSTATE_ZENABLE, D3DZB_TRUE
  Else
    .SetRenderState D3DRENDERSTATE_ZENABLE, D3DZB_FALSE
  End If
  
  ' Enable transparency
  .SetRenderState D3DRENDERSTATE_ALPHABLENDENABLE, True
  .SetRenderState D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA
  .SetRenderState D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA
  
  ' Set the lower alpha limit
  .SetRenderState D3DRENDERSTATE_ALPHATESTENABLE, True
  .SetRenderState D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATEREQUAL
  .SetRenderState D3DRENDERSTATE_ALPHAREF, (User.AlphaRef - 1)

  ' Allow use of color keys, depending on user settings
  .SetRenderState D3DRENDERSTATE_COLORKEYENABLE, User.UseColorKey
  .SetRenderState D3DRENDERSTATE_COLORKEYBLENDENABLE, User.UseColorKey
  
  ' Initialize texture filtering.
  .SetTextureStageState 0, D3DTSS_MAGFILTER, IIf(Filter, FilterOff1, FilterOn1)
  .SetTextureStageState 0, D3DTSS_MINFILTER, IIf(Filter, FilterOff2, FilterOn2)
  
'  .SetTextureStageState 0, D3DTSS_MAGFILTER, FilterOff1
'  .SetTextureStageState 0, D3DTSS_MINFILTER, FilterOff2
'  Filter = 0
  
  ' Switch on cull-facing.  It'll also be toggled on
  ' as necessary during the actual rendering.
  .SetRenderState D3DRENDERSTATE_CULLMODE, D3DCULL_CCW
  TwoSided = 0
  
  ' Set antialiasing, if available
  If User.Settings And Set_2xRes And Sys.CanAlias > 0 Then
    .SetRenderState D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_SORTINDEPENDENT
  Else
    .SetRenderState D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_NONE
  End If
  
End With


End Sub

Private Sub InitAnims(ByRef Data As AnimData)
If Data.AnimType = 0 Then Exit Sub

' Initializes rotation matrixes for the ANI data.
Dim M1 As D3DMATRIX, M2 As D3DMATRIX
Dim M3 As D3DMATRIX, M4 As D3DMATRIX

Dim S As Integer, F As Integer
For S = 1 To UBound(Data.Segments)
With Data.Segments(S)
  For F = 1 To UBound(.Frames)
  With .Frames(F)
    ' Initialize matrixes.
    Dx_7.IdentityMatrix M1
    M2 = M1: M3 = M1: M4 = M1
    
    Dx_7.RotateXMatrix M1, -.Rotation.X * m_Deg2Rad ' First rotation
    Dx_7.RotateZMatrix M2, -.Rotation.Z * m_Deg2Rad ' Second rotation
    Dx_7.RotateYMatrix M3, .Rotation.Y * m_Deg2Rad ' Third rotation
    
    Dx_7.MatrixMultiply M4, M1, M2 ' Stash M1*M2 in M4
    Dx_7.MatrixMultiply .Matrix, M4, M3 ' Stash (M1*M2)*M3 in result
    
    ' Apply offset values.
    .Matrix.rc41 = .Offset.X
    .Matrix.rc42 = .Offset.Y
    .Matrix.rc43 = .Offset.Z
  End With
  Next F
End With
Next S

End Sub

Public Function SetAnimateMode(ByRef RenderFile As DxPieFile, ByVal AniType As String, ByVal How As String) As Integer
'
' Changes the animation settings for a PIE file while rendering.
'
' Returns TRUE if any files were changed (e.g. needs re-rendering)
'

Select Case AniType
Case "Team", "2"
  Select Case How
  Case "Prev", "0"
    ' Shift team-colors back one frame
    If (RenderFile.AnimateMode And Ani_PlayColors) = 0 And RenderFile.TeamColor > 1 Then
      RenderFile.TeamColor = RenderFile.TeamColor - 1
      SetAnimateMode = True
    End If
    
  Case "Next", "1"
    ' Shift team-colors forward one frame
    If (RenderFile.AnimateMode And Ani_PlayColors) = 0 And RenderFile.TeamColor < RenderFile.MaxTeamColors Then
      RenderFile.TeamColor = RenderFile.TeamColor + 1
      SetAnimateMode = True
    End If
  
  Case "Play1", "2"
    
    SetAnimateMode = True
    ' Start/stop single-loop animation.
    If RenderFile.AnimateMode <> Ani_PlayColors Then
      RenderFile.AnimateMode = Ani_PlayColors
      If RenderFile.TeamColor > 1 Then RenderFile.TeamColor = 1
    Else
      RenderFile.AnimateMode = 0
    End If
    
  Case "PlayAll", "3"
    SetAnimateMode = True
    ' Start/stop looping animation.
    If RenderFile.AnimateMode <> Ani_LoopColors Then
      RenderFile.AnimateMode = Ani_LoopColors
    Else
      RenderFile.AnimateMode = 0
    End If
  
  Case Else
    Debug.Assert False
  End Select

Case "Subs", "3"
  Select Case How
  Case "Prev", "0"
    ' Shift submodels back one frame
    If (RenderFile.AnimateMode And Ani_PlaySubs) = 0 And RenderFile.SubFrame > 1 Then
      RenderFile.SubFrame = RenderFile.SubFrame - 1
      PieModel.DoRender = True
    End If
    
  Case "Next", "1"
    ' Shift submodels forward one frame
    If (RenderFile.AnimateMode And Ani_PlaySubs) = 0 And RenderFile.SubFrame < RenderFile.Anim.MaxFrames Then
      RenderFile.SubFrame = RenderFile.SubFrame + 1
      PieModel.DoRender = True
    End If
  
  Case "Play1", "2"
    ' Start/stop single-loop animation.
    If RenderFile.AnimateMode <> Ani_PlaySubs Then
      RenderFile.AnimateMode = Ani_PlaySubs
      If RenderFile.SubFrame > 1 Then RenderFile.SubFrame = 1
    Else
      RenderFile.AnimateMode = 0
    End If
    PieModel.DoRender = True
    
  Case "PlayAll", "3"
    ' Start/stop looping animation.
    If RenderFile.AnimateMode <> Ani_LoopSubs Then
      RenderFile.AnimateMode = Ani_LoopSubs
    Else
      RenderFile.AnimateMode = 0
    End If
    PieModel.DoRender = True
  
  Case Else
    Debug.Assert False
  End Select
  
Case Else
  Debug.Assert False
End Select

End Function


Public Function LoadPCXTexture(ByVal FileName As String, lCode As Long) As DirectDrawSurface7
'
' Loads a PCX file into a DirectX surface.
'
' * Only works with user textures
'

If FileName = "" Then lCode = -1: Debug.Assert False: Exit Function
If Dir(FileName) = "" Then lCode = -1: Exit Function



' General vars.
Dim Ax As Long, Bx As Long, Cx As Long, Dx As Long

' Masks & planes (for multiplane i.e. true-color PCX files).
  Const pl_Red As Integer = 1   '
  Const pl_Green As Integer = 2 ' RGB constants must be in this order
  Const pl_Blue As Integer = 3  '
  Const pl_Alpha As Integer = 4
  Const pl_Black As Integer = 5 ' Not a mask, but a color value
Dim Masks(0 To pl_Black) As Long

'On Error GoTo Catcher

Dim Byt As Byte, Int2 As Integer, Int4 As Long

Dim File As Integer: File = FreeFile
Open FileName For Binary Access Read Lock Write As #File
  
Screen.MousePointer = 13

'
' Step 1: Verify PCX header information
'
''''''''''''''''
  
  Dim Data() As Byte, Ptr As Long
  ReDim Data(0 To 127)

  Get #File, 1, Data()
  
  ' Verify PCX file header characteristics
  If Data(0) <> 10 Then
    lCode = -2 ' Not a PCX format.
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End If

  Select Case Data(1)
  Case 5 ' Supported formats: RGB true color and 8-bit paletted
  
  Case 0, 2 To 4 ' Not supported by Pie Slicer
    lCode = -3
    Screen.MousePointer = 0
    Close #File
    Exit Function
  
  Case Else ' Invalid format
    lCode = -2
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End Select
  
  Dim FileVer As Long
  FileVer = Data(1)
  
  If Data(2) <> 1 Then
    lCode = -2 ' Bad PCX format
    Close #File
    Exit Function
  End If
  
  Select Case Data(3)
  Case 8
  Case 1, 2, 4
    lCode = -3 ' Not supported
    Screen.MousePointer = 0
    Close #File
    Exit Function
  
  Case Else
    lCode = -2 ' Bad PCX format
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End Select
  
  ' PCX dimensions - 8 bytes total
  Dim PCXRect(0 To 3) As Integer
  Get #File, 5, PCXRect()
  
  Dim Width As Integer, Height As Integer
  Width = PCXRect(2) - PCXRect(0) + 1
  Height = PCXRect(3) - PCXRect(1) + 1
  
  'Get #File, 5, Width: Get #File, , Height ' (0,0) point
  '
  'Get #File, , Int2: Width = Int2 - Width + 1   ' (x,y) point --
  'Get #File, , Int2: Height = Int2 - Height + 1 ' use to derive width/height
  
  ' Reserved = 0
  If Data(64) Then
    lCode = -2
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End If
  
  ' # of color planes to use
  ' 1=typical (paletted)
  ' 3=true color (when Filever=5)
  Dim NumPlanes As Long
  NumPlanes = Data(65)
  
  ' Scanline pitch - verify that it's an EVEN number
  Get #File, 67, Int2
  Debug.Assert Int2 = 256 * Data(67) Or Data(66)
  If Int2 And 1 Then ' Data(66) and 1
    lCode = -2
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End If
  
  ' Color table type.  Must be 1 (color) or 2 (mono)
  Get #File, 69, Int2
  If Int2 <> 1 And Int2 <> 2 Then
    lCode = -2
    Screen.MousePointer = 0
    Close #File
    Exit Function
  End If
  
'
' Step 1a: DirectX surface creation
'
''''''''''''''''
Dim ddsd As DDSURFACEDESC2
  
  ' Flags
  ddsd.lFlags = DDSD_CAPS Or DDSD_HEIGHT Or DDSD_WIDTH Or DDSD_PIXELFORMAT Or DDSD_TEXTURESTAGE
  If User.UseColorKey Then ddsd.lFlags = ddsd.lFlags Or DDSD_CKSRCBLT
  
  ' Caps
  ddsd.ddsCaps.lCaps = DDSCAPS_TEXTURE
  ddsd.ddsCaps.lCaps2 = DDSCAPS2_TEXTUREMANAGE
  ddsd.lTextureStage = 0
  
  ' Dims
  ddsd.lHeight = IIf(Height < 256, 256, Height)
  ddsd.lWidth = IIf(Width < 256, 256, Width)
  
  ' Pixel format
  Formats.TexEnum.GetItem Sys.DirectX_TexFormat, ddsd.ddpfPixelFormat
  
  ' Cache the channel masks
  With ddsd.ddpfPixelFormat
    Masks(pl_Red) = .lRBitMask
    Masks(pl_Green) = .lGBitMask
    Masks(pl_Blue) = .lBBitMask
    Masks(pl_Alpha) = .lRGBAlphaBitMask
  End With
  
  ' Cache WZ's transparency color (green)
  Dim c_Green As Long: c_Green = Masks(pl_Green)
  
  ' Locate the LSB in each of the RGB masks
  ' (this will simply computations later)
  For Cx = pl_Red To pl_Blue
    If Masks(Cx) > 0 Then
      Dx = 1
      Do Until Masks(Cx) And Dx
        Dx = Dx * 2
      Loop
      Masks(Cx) = Dx
    End If
  Next Cx
  
  Dim Div As Long
  Select Case Masks(pl_Alpha)
    Case &H8000& ' 16bit ARGB1555
                 ' Only 5 bits of each color value are relevant, so \8
                 
      Masks(pl_Black) = &H8421&
      Div = 8
    
    Case &HF000& ' 16bit ARGB4444
                 ' Only 4 bits of each color value are relevant, so \16
      
      Masks(pl_Black) = &HF111&
      Div = 16
    
    Case &HFF000000 ' 32-bit surface
     
      Masks(pl_Black) = &HFF010101
      Div = 0
      
  End Select
  
'
' Step 2 - Load PCX palette info, if needed
'
''''''''''''''''
  
  Dim TrueRGB As Boolean
  If FileVer = 5 Then
    Select Case NumPlanes
    Case 1 ' Paletted PCX info
      ' Skip to end of file for palette info
    
      ReDim Data(0 To 768)
      Seek #File, LOF(File) - 768
      Get #File, , Data()
      Ptr = 0
      
      If Data(0) <> 12 Then
        Debug.Assert False
        lCode = -2 ' Bad format
        Screen.MousePointer = 0
        Close #File
        Exit Function
      End If
    
      ' Convert color table data to the pixel format
      Dim R As Long, G As Long, B As Long
      
      Dim Pal(0 To 255) As Long
      
      For Cx = 0 To 255
        If Div Then
          ' 16-bit surfaces: ignore less significant bits
          R = Data(Ptr + 1) \ Div * Masks(pl_Red)
          G = Data(Ptr + 2) \ Div * Masks(pl_Green)
          B = Data(Ptr + 3) \ Div * Masks(pl_Blue)
        Else
          ' 32-bit surfaces
          R = Data(Ptr + 1) * Masks(pl_Red)
          G = Data(Ptr + 2) * Masks(pl_Green)
          B = Data(Ptr + 3) * Masks(pl_Blue)
        End If
        If Cx Then
          Pal(Cx) = R Or G Or B Or Masks(pl_Alpha)
        Else
          Pal(Cx) = R Or G Or B 'And Not Masks(pl_Black)
        End If
        
        Ptr = Ptr + 3
      Next Cx
    
      ' Prepare a 8-bit palettized DX surface
'      ddsd.ddpfPixelFormat.lFlags = DDPF_RGB Or DDPF_PALETTEINDEXED8
'      ddsd.ddpfPixelFormat.lRGBBitCount = 8
      
'      Set LoadPCXTexture = Dx_Draw.CreateSurface(ddsd)
'      LoadPCXTexture.SetPalette DxPal
    
    
    Case 3 ' True-color PCX info - prepare a 16-bit DX surface
      
      ' Just note that this is a true-color PCX.  We'll do the rest later.
      TrueRGB = True
      
      ' Create a buffer to hold the scanline data.
      ' This way, we won't need to call .GetLockedPixel later.
      '
      Dim ScanLine() As Long
      ReDim ScanLine(ddsd.lWidth)
      
      
    Case Else
      lCode = -3 ' Unsupported
      Close #File
      Exit Function
    End Select
  End If
  
  ' Finally, initialize the surface
  Set LoadPCXTexture = Dx_Draw.CreateSurface(ddsd)
  
'
' Step 3 - Load PCX image data
'
''''''''''''''''
  
  Screen.MousePointer = 13
  
  ' The image begins at byte offset 128
  Seek #File, 129
  
  ReDim Data(0 To LOF(File) - 128 - IIf(TrueRGB, 0, 769))
  Get #File, , Data()
Close #File
  
Ptr = 0

' Access the DX surface
Dim Area As RECT, Lockdsd As DDSURFACEDESC2
  Area.Left = 0
  Area.Right = ddsd.lWidth
  Area.Top = 0
  Area.Bottom = ddsd.lHeight
LoadPCXTexture.Lock Area, Lockdsd, 0, 0


' Initialize the current plane mask & draw coords
Dim Plane As Integer, Px As Long, Py As Long
Plane = pl_Red ' =1
Px = 0: Py = 0

'
' Then, start decompressing the scanline data
'
Dim RLE As Long, Colr As Long
Do
  ' Read first byte
  Byt = Data(Ptr)
  If (Byt And 192) = 192 Then ' RLE encode flag: read another byte for color
    Debug.Assert Byt > 192
    
    RLE = Byt And 63
    Colr = Data(Ptr + 1)
    Ptr = Ptr + 2
  
  Else ' No RLE for this pixel
    RLE = 1
    Colr = Byt
    Ptr = Ptr + 1
  End If
    
  ' Process according to true-color or paletted modes
  If TrueRGB Then
    Ax = Ax + RLE
    
    ' Begin writing the pixels
    For Cx = 1 To RLE
      '
      ' Determine RGB pixel value, since each plane is encoded separately
      '
      
      ' Convert color value to DX pixel format
      If Div Then
        Ax = Colr \ Div * Masks(Plane)
      Else
        Ax = Colr * Masks(Plane)
      End If
      
      Select Case Plane
      Case pl_Red ' First pass (red) - nothing special
      
      Case pl_Green ' Second pass (green)
        Ax = Ax Or ScanLine(Px)
        
      Case pl_Blue ' Final pass (blue) - add alpha values
        Ax = Ax Or ScanLine(Px)
        If Ax <> c_Green Then Ax = Ax Or Masks(pl_Alpha)
        If Ax = Masks(pl_Alpha) And User.UseColorKey Then Ax = Masks(pl_Black)
      End Select

      ' Cache the color value as needed
      ScanLine(Px) = Ax
      
      ' Set the surface pixel
      Debug.Assert Px >= 0 And Px < Width
      Debug.Assert Py >= 0 And Py < Height
      LoadPCXTexture.SetLockedPixel Px, Py, ScanLine(Px)
      
      ' Allow for multiplane wrapping at end of line
      Px = Px + 1
      If Px >= Width Then ' End of line
        Px = 0
        
        If Plane < pl_Blue Then ' Next plane
          Plane = Plane + 1
        
        Else ' End of scanline
          Plane = pl_Red
          Py = Py + 1
          If Py >= Height Then Exit Do ' End of file
        End If
      End If
      
    Next Cx
    
  Else
    
    ' 8-bit palette mode.  MUCH easier than true color!
    
    For Cx = 1 To RLE
      ' Set the pixel color
      Debug.Assert Px >= 0 And Px < Width
      Debug.Assert Py >= 0 And Py < Height
      LoadPCXTexture.SetLockedPixel Px, Py, Pal(Colr)
      
      Px = Px + 1
      If Px >= Width Then ' End of line
        Px = 0
        
        Py = Py + 1
        If Py >= Height Then Exit Do ' End of file
      End If
    Next Cx
  
  End If
  
Loop Until Ptr > UBound(Data)


Done:

LoadPCXTexture.Unlock Area

Screen.MousePointer = 0
Exit Function

Catcher:
  
  Debug.Print "Error #" & Err.Number & " - " & Err.Description
  Debug.Assert False
  Set LoadPCXTexture = Nothing
  
On Error GoTo 0
Resume

End Function


Public Sub VerifyTexture(Data As DirectDrawSurface7)
Dim X As Long, Y As Long, P As Long
Dim Area As RECT, ddsd As DDSURFACEDESC2
If Data Is Nothing Then Exit Sub
Data.GetSurfaceDesc ddsd

Data.Lock Area, ddsd, 0, 0
For Y = 0 To 31
 For X = 0 To 31
  P = Data.GetLockedPixel(X, Y)
  Debug.Print Left$(Hex$(P), 1);
 Next X: Debug.Print
Next Y
Data.Unlock Area

Debug.Assert False

End Sub

Public Sub SetFilters(ByVal Allow As Integer)
'
' Records whether texture filtering will be allowed or not.
'
FilterOn1 = IIf(Allow And 1, D3DTFG_POINT, D3DTFG_LINEAR)
FilterOn2 = IIf(Allow And 1, D3DTFN_POINT, D3DTFN_LINEAR)
FilterOff1 = IIf(Allow And 2, D3DTFG_LINEAR, D3DTFG_POINT)
FilterOff2 = IIf(Allow And 2, D3DTFG_LINEAR, D3DTFG_POINT)

If Not Dx_Device3D Is Nothing Then
  Dx_Device3D.SetTextureStageState 0, D3DTSS_MAGFILTER, IIf(Filter, FilterOff1, FilterOn1)
  Dx_Device3D.SetTextureStageState 0, D3DTSS_MINFILTER, IIf(Filter, FilterOff2, FilterOn2)
End If

End Sub

Public Sub DoCaps()

Dim Caps1 As DDCAPS, Caps2 As DDCAPS
Dx_Draw.GetCaps Caps1, Caps2

Debug.Print Caps1.lCaps And DDCAPS_PALETTE, Caps2.lCaps And DDCAPS_PALETTE

End Sub
