Attribute VB_Name = "Module1"

' Version spec for PCK files.
'   100 - Original PCK spec
'   110 - New texture management system
'   120 - Support for embedded PIE data
Private Const VerPCK As Integer = 120

' Subformat ID for embedded PIE data.
'   0   - Original PCK_Pie version
'   110 - New submodel format (per-polygon spec.)
'   111 - Bugfix: # polygons stored as hex not decimal
Private Const VerPCK_Pie As Integer = 111

' Subformat ID for config.dtb data, [Options] section
'
Private Const VerCFG_Options As Integer = 100


' Was in FormMain, moved here now.
Public ViewIndex As Integer


'
' Constants related to WZ operation.
'
' Max # of points WZ allows in a polygon.  DNF!
Public Const WZ_MaxPolySize As Integer = 10

' Values for passing to the gotSelected() function
'
Public Enum GotSelectStates
  ' Values for the Query argument
  '
  Get_Pts = &H10
  Get_Pys = &H8
  Get_QueryMask = Get_Pts Or Get_Pys ' Bitmask for the above
  
  Get_CurOnly = 0
  Get_AllVisi = 1
  Get_WithCur = 2
  Get_WithSel = 3
  Get_InSub = 4
  Get_AllSel = 5
  Get_HowMask = 7 ' Bitmask for the above
End Enum


Public ThisFile As Integer


' Not used currently, but keep around just in case
'Public Declare Function ShellExecute Lib "shell32.dll" Alias "ShellExecuteA" ( _
    ByVal hWnd As Long, _
    ByVal lpOperation As String, _
    ByVal lpFile As String, _
    ByVal lpParameters As String, _
    ByVal lpDirectory As String, _
    ByVal nShowCmd As Long) As Long

' For the 'Always On Top' function
Declare Sub SetWindowPos Lib "User32" (ByVal hWnd As Long, ByVal _
    hWndInsertAfter As Long, ByVal X As Long, ByVal Y As Long, ByVal Cx _
    As Long, ByVal CY As Long, ByVal wFlags As Long)

Public Const HWND_TOPMOST = -1
Public Const HWND_NOTTOPMOST = -2
Public Const SWP_NOACTIVATE = &H10



' This, because of a VB crash.
Option Explicit

' Viewing angles/modes for Pie Slicer.
Public Enum ViewModes
  VIEW_BACK = 0
  VIEW_TOP ' 1
  VIEW_SIDE ' 2
  VIEW_3D ' 3
  VIEW_RENDERED ' 4
End Enum

Public Enum PieTypes
  TYPE_SINGLE = 0
  TYPE_HIDE ' 1
  TYPE_BODY ' 2
  TYPE_PROP ' 3
  TYPE_GNDTUR ' 4
  TYPE_AIRTUR ' 5
  TYPE_MUZL ' 6
  TYPE_MUZLCYB ' 7
  TYPE_FLASH ' 8
  TYPE_PROJ ' 9
End Enum


' Generic PIE Slicer status message codes.
'
Public Enum StatMsgs
  Msg_NeedPts = 1
  Msg_NeedCc
  Msg_NeedPtOrCc
  Msg_NeedPys
  Msg_Sel
  Msg_SelPts
  Msg_SelCc
  Msg_SelPtOrCc
  Msg_SelPys
  
  Msg_ClickPy
  Msg_ClickVx
  Msg_NoSplitTri
End Enum



' BitBlt routine, and sister function StretchBlt
' Who needs "PaintPicture" anymore?
Declare Function BitBlt Lib "gdi32" (ByVal hDestDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal dwRop As Long) As Long
Declare Function StretchBlt Lib "gdi32" (ByVal hdc As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal nSrcWidth As Long, ByVal nSrcHeight As Long, ByVal dwRop As Long) As Long


' Cached variables for quicker access
'Public ThisLev As Integer
Public ThisPoint As Integer
Public ThisPoly As Integer
Public ThisCc As Integer

' Camera offset to utilize.
Public CamCentre As D3DVECTOR

Public SkipEvents As Boolean


' This for future implementation, should I need it(?)
'Declare Function TransparentBlt Lib "msimg32" (ByVal hDestDC As Long, ByVal X As Long, ByVal Y As Long, ByVal nWidth As Long, ByVal nHeight As Long, ByVal hSrcDC As Long, ByVal xSrc As Long, ByVal ySrc As Long, ByVal nWidthSrc As Long, ByVal nHeightSrc As Long, ByVal crTransparent As Long) As Long

' Quick-reference window ID codes.
' Each form has a local constant called "win_Me", set to one of these.
Public Enum WindowEnums
  win_Screen = 0
  win_First = 1
    win_Master = win_First
    win_Model
    win_Lists
    win_Tools
    win_Texture
  win_Last = win_Texture
End Enum

' Status flags.  Used to manage point and polygon data
'
Public Enum StatFlags
  
  ' Point/polygon select flags
  '
  stat_Hide = &H8000 ' Hidden from view (overrides Stat_Sel flag)
  Stat_Normal = 0 ' Normal/locked
  stat_Sel = 1 ' Selected for editing
  
  ' Bitmasks.
  ' (StatFlag and Stat_Hide) returns visibility
  ' (StatFlag and Stat_SelMask) = Stat_Sel returns select status.
  '
  stat_SelMask = stat_Sel Or stat_Hide
  stat_NotSelMask = Not stat_SelMask
  
  ' Submodel flags - currently used mostly by enumSubmodels() function
  '
  stat_Sub1 = &H10 ' Bitflag for the submodel #1
  Stat_SubsAll = &H3FF0 ' 10-bit mask - one for each submodel
  Stat_AutoSubs = &H4000 ' Specifies to let Pie Slicer handle submodel point data
  Stat_SubFlagMask = Stat_SubsAll Or Stat_AutoSubs
End Enum


' Texture array format, used for both DirectX & GDI
Public Type DxTexture
  Page As Integer ' Page # for PIE file reference
  FileName As String ' Texture being used here
  
  ' Reference data for this surface
  SurfPtr As Integer ' Ptr to vidmem array
  SurfWidth As Integer
  SurfHeight As Integer
End Type

' PIE type point structure.  Also used for polygon vertex data.
Public Type PointData
  ' Coordinate information
  Loc As D3DVECTOR
  
  Index As Integer ' Vertex data only; not used otherwise

  'Status flag and miscellaneous data:
  StatFlags As Integer
End Type

' Team-colors suboptions for polygon data
Public Type TeamSettings
  Animated As Integer
  AnimTime As Integer ' Time delay (ms) -- firing GFX only
  AnimW As Integer ' SizeX of frame (px.)
  AnimH As Integer ' SizeY of frame (px.)
End Type

Public Type PolygonData
  
  ' Status flags
  StatFlags As Integer
  
  ' Submodel identification
  Submodel As Integer
  
  ' PIE type flags for this polygon. See enumeration, next
  Flags As Long
  
  ' Set of polygon vertexes (includes texturing).
  ' Use UBound(PtData) to return # of points.
  PtData() As PointData
  
  ' Animated / team-colored polygons use this.
  ' Use PtData().Loc to grab first frame's texture.
  TeamC As TeamSettings
End Type

Public Enum PolyFlags
  ' Supported flags
  Py_Textur = &H200&
  Py_Pixel = &H800&
  Py_2xSide = &H2000&
  Py_TeamC = &H4000&
  
  ' Unsupported flags (named by value)
  Py_Flag1 = &H100&
  Py_Flag4 = &H400&
  Py_Flag10 = &H1000&
  Py_Flag80 = &H8000&
  Py_Flag100 = &H10000
  Py_Flag200 = &H20000
  Py_Flag400 = &H40000
  Py_Flag800 = &H80000
End Enum

' Submodels data structure.
'
Public Type LevelType

  ' Points array
  Points() As PointData

  ' Polygons array
  Polys() As PolygonData

  ' Connectors array
  Ccs() As PointData

  ' BSP array (IGNORED)
  BSP() As String
End Type

' ANI data structures.
Public Type FrameData
  Frame As Integer
  Offset As D3DVECTOR ' default=0
  
  Rotation As D3DVECTOR ' about x/y/z axes
  
  Scalar As D3DVECTOR ' Present, but not used by WZ
  Matrix As D3DMATRIX ' For rendering (computed from offset/rotations)
End Type
Public Type SegData
  Name As String
  Parent As Integer ' Segment heirarchy, inherits offset/rotation from them
  
  RotaCenter As D3DVECTOR ' default=0, indicates center-point for rotation
  RelRotate As Integer ' default=0, otherwise rotation angles are relative to parent
  
  Frames() As FrameData
End Type
Public Type AnimData
  FileName As String ' Name of associated ANI file
  AnimType As Integer ' See Anim_Types, below
  
  MaxFrames As Integer
  FrameRate As Integer ' fps
  Segments() As SegData
End Type

Public Enum Anim_Types
  NOANIM = 0
  ANIM3DFRAMES = 1
  ANIM3DTRANS = 2
End Enum

' PIE data structure.
Public Type PieType
  FileName As String
  Flags As Long
  
  ' These values control whether a file's been modified or not
  Loaded As Single ' Time of last save/load
  Changed As Single ' Time of last change
  
  TexPage As String ' As specified in the PIE format, for WZ
  
  TexFile As Integer ' For Pie Slicer reference
  
  Anim As AnimData ' Animation data for multiple submodels
  
  ' Enumerated logical PIE type (for rendering)
  PieType As Integer
  
  ' Reference values calculated from the above
  PieOrigin As D3DVECTOR
  P3DFlags As Long ' = 0 or P3D_IsVTOL
  
  ' Levs() as LevelType
  Levs As LevelType

End Type

' PCK data structure.
Public Type PieModel
  FileName As String
  Files() As PieType
  
  Anims() As AnimData
  
  ' Temporary use only -- use Sys.VidMem() for the rest
  Tex_Files() As String
  
  Loaded As Single
  Changed As Single
End Type

' Main data array.
Public Model As PieModel

' UDT support for an "Undo" feature.
'
Public Type UndoState
  
  ' Backup data for the model.
  PieStatus As PieModel
  
  ' Other backup data
  '
  CurFile As Integer ' Important when editing multiple files!
  CurLev As Integer
  CurPoint As Integer
  CurCc As Integer
  CurPolygon As Integer
  
  Action As Integer ' Type of change made to the data
  Done As Integer   ' Allows repeated operations of the same type
                    ' (e.g. move/rotate) to be considered a single step.
                    ' TRUE forces the next change to be a different step.
End Type


' Pie Slicer system constant.
'
' Defines the maximum # of steps supported by the Undo/Redo system
'
Const MAX_UNDO As Long = 5

Public Enum UndoOps
  
  ' Special action codes go first.
  Undo_Clear = 0 ' Erases the Undo history.  Careful!
  Undo_Undo ' Edit/Undo command
  Undo_Redo ' Edit/Redo command
  
  ' Next come the groupings.
  '
  ' First group are actions which cache the PIE data each and every time
  Undo_FirstReCache
    Undo_CopyPyFlags = Undo_FirstReCache
    Undo_CopyPyAni
    Undo_AddPt
    Undo_DelPt
    Undo_DelPts
    Undo_AddCc
    Undo_DelCc
    Undo_AddPy
    Undo_DelPy
    Undo_DelPys
    Undo_AttachPt
    Undo_AttachPts
    Undo_DoExtrude
    Undo_SlicePoly
    Undo_ShatterPoly
    Undo_JoinPolys
    Undo_PasteData
    Undo_Mirror
  Undo_LastReCache = Undo_PasteData
  
  ' Next come the groups that can be considered 'continuous'.
  '
  ' 'Continuous' means that repeated actions of the same type
  ' will be considered one edit and PIE data won't be re-cached each time.
  '
  ' Group #2 is for actions continuous across the same selection
  '
  Undo_FirstSameSel
    Undo_MovePts = Undo_FirstSameSel
    Undo_MoveSel = Undo_MovePts
    Undo_ScaleSel
    Undo_RotateSel
    Undo_ChangeSub
  Undo_LastSameSel = Undo_ChangeSub
  
  ' Group #3 is for actions continuous for the same polygon only
  '
  Undo_FirstSamePoly
    Undo_EditPy = Undo_FirstSamePoly
    'Undo_FlipPy
    Undo_ReorderPy
    Undo_SetTriFan
    Undo_AddVertex
    Undo_MoveVertex
    Undo_DetachVertex
    Undo_KillVertex
  Undo_LastSamePoly = Undo_KillVertex
  
  ' Then come the actions which aren't grouped at all,
  ' either because of special handling or NYI.
  '
  Undo_EditPt
  Undo_EditCc
  
  ' User changed current selection.
  ' Doesn't actually change the PIE data, and therefore
  ' isn't counted as an Undo step, but it instead
  ' flags the previous operation (whatever it was)
  ' as being complete, thus forcing a re-cache for the next edit
  ' even if the next edit is the same type as the last edit.
  '
  Undo_ChangeSel
  
  Undo_TexturePy ' NYI
  
  ' No longer needed because of new submodel management
  'Undo_AddSub
  'Undo_DeleteSub
  
End Enum

' Array fonts for quick Pie Slicer reference.
'
Public Enum StyleEnums
  STYLE_First = 1
    STYLE_6pt = STYLE_First
    
    STYLE_7pt
    STYLE_7ptB
    
    STYLE_8pt
    STYLE_8ptB
  STYLE_Last = STYLE_8ptB
End Enum

' PIE Slicer's options and other system data.
Public Type Keywords
  Pie As String
  Tex As String
End Type

' Handy UDT for various Pie Slicer app settings.
'
' There can be only ONE variable of this type!
'
Public Type SystemType
 
  ' General variables.
  '
  WindowPos(win_Last) As D3DRECT
 
  ' Array of fonts for quick reference
  '
  MyFonts(STYLE_First To STYLE_Last) As New StdFont
 
  ' formMaster related options
  '
  RecentFiles() As String ' Recent files.
  UseLineFilter As Integer ' How to load PIE files (by line or by word)
  
  ' formMain related settings
  '
  HorzAngle As Integer: VertAngle As Integer ' 3D view angles
  
  ZoomStep As Integer ' Zoom level (in dB)
  ZoomScalar As Single ' Zoom scalar
  
  MoveStep As Integer ' step level from 1-9
  MoveRate As Integer
  
  ' formTex related settings
  '
  TexPage As String ' Texpage used for alignment.
  TexZoom As Integer ' Texture zoom factor.
  
  Keyword() As Keywords
  ' Auto-texturemapping options
  TexCenter As D3DVECTOR
  TexRotation As Single
  TexScalar As Single
  
  
  ' formTools related settings.
  '
  PtRotation As Single
  PtScalar As Single
  PtCenter As D3DVECTOR ' Center / axis base point
  PtAxis As D3DVECTOR ' Secondary axis point
  
  
  ' 0 = 16x16
  ' 1 = 24x24
  ' 2 = 32x32
  IconSet As Integer
  IconSize As Integer
  
  ' Other general options.
  '
  
  ' Return value from formMain.DoTeamColors(-1)
  DTCMenuCode As Long
  
  DirectX_TexFormat As Integer  ' Texture format used by the video card
  DirectX_TexColorGreen As Long ' Color key, dictated by format
  DirectX_ZFormat As Integer    ' Z buffer format used by video card

  ' Presence/absence of antialiasing caps.
  CanAlias As Long

  ' Level of paletted texture support.
  '   2 = hardware
  '   1 = emulated
  '   0 = none
  PalMode As Long
  
  ' Array of DirectX surfaces available for texturing
  VidMem() As DxTexture
  
  ' Index to current vidmem page.
  ' Reference var used when NOT rendering.
  CurPage As Integer

  
  ' Vars required for the Undo/Redo feature.
  '
  Backups(MAX_UNDO) As UndoState ' Buffer of PIE model data
  
  BackupPtr As Integer ' Indexes to the above array.
                       ' Backups(BackupPtr) records most recent change
  
  MaxUndoStep As Long ' Maximum # of Undo steps currently in memory
                      '   * Cannot exceed MAX_UNDO
  CurUndoStep As Long ' Current position within the Undo history.
                      ' Usually = MaxUndoStep when changes are being made,
                      ' but decrements with each Undo command used.
                      '
                      ' If CurUndoStep < MaxUndoStep, the Redo command
                      ' will be available for use.
End Type

' Constants for Sys.MoveStep,
' from which Sys.MoveRate is calculated
'
Public Const MoveStep_Min As Integer = 0
Public Const MoveStep_Init As Integer = 3
Public Const MoveStep_Max As Integer = 8


Public Enum Settings
  ' GUI related options.
  Set_ShowMap = &H1       ' Show Keymappings
  Set_ShowGrid1 = &H2     ' Show Editing grid
  Set_ShowGrid2 = &H4     ' Show WZ tile grid
  Set_NumberGrid2 = &H8   ' Show tile #'s in 2D WZ grids
  Set_ShowCompass = &H100 ' Show compass
  
  ' User config options.
  Set_MoveBinary = &H10 ' Binary move rate
  
  Set_2xRes = &H200     ' Antialiasing.
                        ' If requested, but not available on the card,
                        ' Pie Slicer will render at double resolution
                        
                        
  ' Flags for PCX support (2 bits)
  ' (No longer needed since PCX loader is fixed)
'    Set_NoPCX = 0
'    Set_LowPCX = &H400
'    Set_FullPCX = &HC00
'  Set_PCXMask = &HC00
  
  Set_Default = Set_ShowMap Or Set_ShowGrid1 Or Set_ShowGrid2 Or Set_NumberGrid2 Or Set_ShowCompass Or Set_MoveBinary 'Or Set_FullPCX
End Enum

' Secondary options, user configurable and saved to disk
Public Type UserType
  GridSpacing As D3DVECTOR ' Grid spacing options
  
  AlwaysOnTop As Integer
  
  AutoSave As Integer
  
  ' See above enumeration.
  Settings As Long
  
  ' Maximum # of recent files in the Recent Files list.
  ' Actual filenames are stored in the System datatype however.
  MaxFiles As Integer
  
  ' Point/polygon options.
  NewOffset As D3DVECTOR
  
  ' DirectX texture pixel format (indexes to the video card)
  TexFormat As Long
  
  UseColorKey As Integer
  AlphaRef As Integer ' 0-255
  
  IconSet As Integer ' 0-2
End Type

Public Sys As SystemType
Public User As UserType

' Flags for Project3D()
'
Public Enum ProjectionFlags
  P3D_MakeInts = &H10& ' Round to integers (WZ precision)
  P3D_IntMath = P3D_MakeInts ' An alias
  
  P3D_InvertY = &H20&   ' Return screen-compatible Y coordinate
  P3D_UseCamPt = &H40& ' Rotate around supplied center-point
  
  P3D_MyAngles = &H80&  ' Use Angl/Angl2 params
  
  P3D_AddZoom = &H100& ' Include zoom factor (commonly used)
  
  P3D_IsVTOL = &H200& ' Aircraft coords: invert X/Y axes before rotation
  
  P3D_UseFilePt = &H400& ' Use specified file point
  
  ' Typical flags setup for calling Project3D()
  P3D_Typ = P3D_MakeInts Or P3D_UseCamPt Or P3D_InvertY Or P3D_AddZoom
  
  P3D_MyFile = P3D_Typ
  P3D_OtherFile = P3D_Typ Or P3D_UseFilePt
  
  ' Other derived flags
  P3D_TypVTOL = P3D_Typ Or P3D_IsVTOL
  
  ' Default flags setup = perform rotation based on current cam angles only
  P3D_Math = 0
End Enum

' Flags enum for DrawPolygon()
'
Public Enum DrawPolyFlags
  
  ' Primary flags, passed to the Flags parameter
  '
  HidePoints = 1&
  HidePtNums = 2&
  HideNormal = 4&
  DrawTriFan = 8&
  DrawPolyNum = &H10&
  SoloColor = &H20& ' Ignore SetColor1 and SetBackColor subflags
  
  DoTexproject = &H100& ' Used for the DrawTexPoly() function
  
  ' Secondary flags (with params), passed to the SubFlags() array
  '
  SetColor1 = &H1000& ' Next element is first color
  SetColor2 = &H2000& ' Next element is second color
  SetBackColor = &H4000& ' Next element is backside color
  SetColorMask = SetColor1 Or SetColor2 Or SetBackColor ' All of the above
  
  ' Other subflags.  May not be combined.
  '
  SetWidth = &H8000& ' Next element is draw width
  WarpVertex = &H10000  ' Next two elements are vertex & point for polygon
  
  ' Common flag combinations.
  '
  Flags_Default = HideNormal Or HidePoints ' usual polygons
  Flags_CurPoly = DrawTriFan Or DrawPolyNum ' Current Polygon
  Flags_SelPoly = HideNormal Or HidePoints Or HidePtNums Or SoloColor ' Selected polys
  Flags_StdPoly = HidePoints Or HideNormal Or SoloColor ' Standard (simple) polygons
  Flags_ExtrudePy = HidePtNums Or HideNormal Or SoloColor ' Extrusion polys
  
  ' Flags combination for the autotexturemapping mode
  Flags_py_AutoTexMap = DoTexproject Or HidePoints Or HidePtNums Or SoloColor
End Enum

' Flags for scale/rotate ops
'
Public Enum ScaRotaFlags
  ' Main op flags. Also used for sel_GetPointList()
  '
  SCR_DoPoints = 1 ' Affect points
  SCR_DoPolys = 2  ' Affect polygon vertexes
  SCR_DoCcs = 4    ' Include connectors
  SCR_DataMask = SCR_DoPoints Or SCR_DoPolys Or SCR_DoCcs
  
  ' Axis flags.
  '  * MAY be combined for scaling, but NOT for rotations
  '
  SCR_DoAxisX = 8
  SCR_DoAxisY = 16
  SCR_DoAxisZ = 32
  SCR_AxisMask = SCR_DoAxisX Or SCR_DoAxisY Or SCR_DoAxisZ
  
  ' Compute relative to Sys.PtCenter when supplied
  '
  SCR_UseCenterPt = 64
  
  ' Perform axial scale/rotate when supplied.
  ' Takes priority over most of the above flags.
  SCR_DoAxial = 128
  SCR_DoAxialMask = SCR_DoAxial Or SCR_DataMask
End Enum


' Flags that don't belong much elsewhere.  Used by a few select functions.
'
Public Enum MiscFlags
  
  ' Constants used by Pie Slicer's Open/Save As dialog boxes.
  '
  cdl_SaveFlags = cdlOFNOverwritePrompt
  cdl_OpenFlags = cdlOFNFileMustExist
  cdl_IsReadOnly = cdlOFNReadOnly
  
  ' For the GetPieSize() function
  GetSize_ZoomLevel = 1 ' Return zoom level instead of size
  GetSize_AxialComp = 2 ' Measure each axis instead of sphere radius
  
  ' Used for ArcTanPt()
  ATN_UseCtrPt = 1 ' Allows an offset value
  ATN_ScreenY = 2  ' Inverts the Y value
  ATN_AllowNeg = 4 ' Allows negative result to be returned.
                   
  
  ' For the SetCamPoint() function
  ' The Toolbox Camera Toolbar button indexes MUST match these!
  '
  CamPt_None = 0
  CamPt_Point = 1
  CamPt_Poly = 2
  
  ' for formMain.DoTeamColorsMenu(-1)
  '
  DTC_UseColors = 1
  DTC_MultiColors = 2
  DTC_UseSubs = 16
  DTC_MultiSubs = 32

  ' Used for pt_CenterOf() function
  '
  PTC_AllSel = 1
  PTC_CurPy = 2
  PTC_TexPy = 4
  PTC_GetMask = PTC_AllSel Or PTC_CurPy Or PTC_TexPy
  
  PTC_IntMath = P3D_IntMath ' &H10
  
  ' Used for the getBlankLevel() function
  '
  BlankLev_Points = 1
  BlankLev_StartPt = 3 ' * Includes Points flag
  BlankLev_Polys = 4
  BlankLev_Ccs = 8
  
  BlankLev_New0 = BlankLev_Points Or BlankLev_Polys Or BlankLev_Ccs
  BlankLev_New1 = BlankLev_New0 Or BlankLev_StartPt
  
  ' Used for determining usable format of a data file
  '
  Format_Unable = -1 ' Unable to determine file type
  Format_Invalid = 0 ' Not a Pie Slicer file
  Format_PieFile = 1 ' Pie format
  Format_AniFile = 2 ' ANI format (may include PIE data)
  Format_PckFile = 3 ' PCK format (may include PIE data)
  
  ' "Cak" format available only in very old versions of Pie Slicer.
  ' Not otherwise supported anymore.
  Format_CakFile = 4
  
  ' To optimize various data-related operations:
  '
  Sync_NoUpdates = 0&
    Sync_DoRedraw = 1& ' performs redraw on formMain
    Sync_UpdateLists = 2& ' updates listboxes in formLists
    Sync_SetCurrent = 4& ' Sets current point/poly
  Sync_UpdateAll = -1 ' all updates=yes
End Enum

' Enum for all errors that can occur whilst loading a file from disk.
' Note they are all negative values (very important).
'
' The uppermost byte represents the function that caused it, the rest
' represents the actual error.
Public Enum FileErrors
  FE_Okay = 0 ' No errors
  FE_Generic = &HFFFFFFFF ' Generic error code, no description
  
  ' Function spec, identifies where the error occured
  FE_LoadCak = &HF1000000    ' file_LoadCak()
  FE_LoadPie = &HF2000000    ' file_LoadPIE()
  FE_LoadAni = &HF3000000    ' LoadAniFile()
  FE_LoadPck = &HF4000000    ' LoadPckFile()
  FE_LoadPckPie = &HF5000000 ' file_LoadPCK_PieFile()
  FE_Filter = &HF6000000     ' file_GetType()
  FE_LoadFile = &HF8000000   ' file_LoadData()
  
  ' Error codes caused by bad model data.
  FE_FileNotFound = &HF0000100
  FE_BadFormat = &HF0000200
  FE_NoAniSrcFile = &HF0000300
  FE_BadAniSrcFile = &HF0000400
  
  FE_BadFileHeader = &HF0000500
  FE_NewerVersion = &HF0000600
  
  FE_BadPckSrcFile = &HF0000800 ' Also used in file_LoadPCK_PieFile()
  
  FE_NoData = &HF0008000
  FE_UserCancel = &HF000F000 ' User-selected cancel
  
  FE_Internal = &HF000FFFF ' Internal (miscellaneous) VB error
  
  ' Bitmask names.
  FE_ErrMask = FE_Internal
  FE_SubMask = &HFFFF0000
End Enum

' Data used for copy/paste functions.
' Note that only one level (submodel) can be
' copied at any one time--that is IMPORTANT!
Public PasteBuffr As LevelType

' Used when displaying a polygon's texture.
Public Tex As D3DVECTOR

' Constants for v_GetAxis() and camera,
' named after the numpad keys that represent them.
'
' Note that the axes are contiguous integers (2-7) - IMPORTANT!
'
Public Enum ScreenAxes
  ax_InvertMask = 1
  ax_AxisMask = 6
  
  axRight = 2
  axLeft = axRight Or ax_InvertMask
  axUp = 4
  axDown = axUp Or ax_InvertMask
  axPlus = 6
  axMinus = axPlus Or ax_InvertMask

  axFirst = axRight
  axLast = axMinus
End Enum

' Color related constants.
Public Enum ColorConstants
  
  ' Channel constants.
  ' These are typically multiplied by a 0-255 value
  ' to produce the RGB color value needed.
  '
  ' For example, RGB(128,0,0) would become 128 * col_R
  col_R = &H1&
  col_G = &H100&
  col_B = &H10000
  col_Yel = col_R Or col_G
  col_Cyn = col_G Or col_B
  col_Mag = col_R Or col_B
  col_RGB = col_R Or col_G Or col_B
  
  ' Channel masks - simple enough
  col_RMask = &HFF&
  col_GMask = &HFF00&
  col_BMask = &HFF0000
  col_YelMask = &HFFFF&
  col_CynMask = &HFFFF00
  col_MagMask = &HFF00FF
  Col_RGBMask = &HFFFFFF
End Enum

' Math related constants.  No need to compute them during run-time!
'
' pi related constants...
'
Public Const m_pi As Single = 3.14152965358979
Public Const m_Deg2Rad As Single = m_pi / 180
Public Const m_90Deg As Single = m_pi / 2
Public Const m_2pi As Single = 2 * m_pi

' log related constants...
Public Const m_e As Single = 2.71828182845905     ' =Exp(1)
Public Const m_Log2 As Single = 0.693147180559945 ' =Log(2)
Public Const m_Log10 As Single = 2.30258509299405 ' =Log(10)

' ---------------------------
' Trick borrowed from VBNet Code Library, should it ever be needed
' ( http://www.mvps.org/vbnet/code/ )

  Public Declare Function OpenProcess Lib "kernel32" (ByVal dwDesiredAccess As Long, ByVal bInheritHandle As Long, ByVal dwProcessId As Long) As Long
  Public Declare Function GetExitCodeProcess Lib "kernel32" (ByVal hProcess As Long, lpExitCode As Long) As Long
  Public Declare Function CloseHandle Lib "kernel32" (ByVal hObject As Long) As Long
  Public Const ProcInfo = &H400&
  Public Const StatPend = &H103&
'----------------------------

Public Function enumMaxLevels(ByRef Lev As LevelType) As Integer
'
' Enumerates the # of submodels in a PIE model.
'
' Faster than enumSubmodels() because it doesn't
' compile the list of submodel data (i.e. points, polys)
'
Syntax:
'
' enumMaxLevels(Lev)
'   Lev - Pie data to evaluate.
'
' Returns the maximum submodel found in the Pie data.

Dim A As Long
For A = 1 To UBound(Lev.Polys)
With Lev.Polys(A)
  
  If .Submodel > enumMaxLevels Then enumMaxLevels = .Submodel
  If enumMaxLevels > 9 Then Exit Function

End With
Next A

End Function

Public Function getViewAngles(ByRef HorzAngle As Integer, ByRef VertAngle As Integer, ByRef Zoom As Single) As Integer
'
' Returns the current view settings -- viewing angles and zoom scalar
'

Select Case ViewIndex
Case VIEW_BACK
  HorzAngle = 0
  VertAngle = 0

Case VIEW_SIDE
  HorzAngle = 270
  VertAngle = 0

Case VIEW_TOP
  HorzAngle = 0
  VertAngle = 90

Case VIEW_3D, VIEW_RENDERED
  HorzAngle = Sys.HorzAngle
  VertAngle = Sys.VertAngle
Case Else
  Debug.Assert False
End Select

Sys.ZoomScalar = 2 * 10 ^ (Sys.ZoomStep / 20)
Zoom = Sys.ZoomScalar

End Function


Public Function GetP3DFlags(DrawPie As PieType, ThisPie As PieType, CamPt As D3DVECTOR) As Long
'
' Fetches camera information of a PIE file for calling Project3D() with.
'
Syntax:
'
' GetP3DFlags(DrawPie, ThisPie, CamPt)
'   DrawPie - PIE file being drawn
'   ThisPie - PIE file being edited
'   CamPt   - (output) Camera offset for DrawPie coordinates
'
' Function sets CamPt and returns an appropriate Flags value
' to call Project3D() with.

CamPt.X = CamCentre.X - (DrawPie.PieOrigin.X - ThisPie.PieOrigin.X)
CamPt.Y = CamCentre.Y - (DrawPie.PieOrigin.Y - ThisPie.PieOrigin.Y)
CamPt.Z = CamCentre.Z - (DrawPie.PieOrigin.Z - ThisPie.PieOrigin.Z)

GetP3DFlags = P3D_Typ Or DrawPie.P3DFlags

End Function

Public Function Project3D(ByVal Flags As Long, Point As D3DVECTOR, CamPt As D3DVECTOR, ByVal Angl As Integer, ByVal Angl2 As Integer) As D3DVECTOR
'
' Projects a point from PIE file coordinates into 3D (view) space.
'
' * This function is time-critical
' * And it's used just about EVERYWHERE
'
Syntax:
'
' Project3D(Flags, Point, CamPt, FilePt, Angl, Angl2)
'
'   Flags  -- Directs the function's behavior:
'               P3D_MyAngles - Use Angl and Angl2
'                                * otherwise, current view angle will be used
'               P3D_MakeInts - Limit point coords to WZ precision
'               P3D_UseCamPt - Rotate around camera point.
'               P3D_InvertY  - Return screen-compatible coordinates.
'               P3D_IsVTOL   - Invert X/Y coordinates i.e. for aircraft data
'
'   Point  -- The point coordinates to transform.
'
'   CamPt  -- Location of camera in current file's model space.
'               * Used only when P3D_UseCamPt flag set;
'                 can be anything otherwise
'               * Must also include the difference in PIE file locations
'                 when rendering a point in a different PIE component
'
'   Angl   -- Horizontal viewing angle (y axis rotation)
'   Angl2  -- Vertical viewing angle (x axis rotation)
'               * Used only when P3D_MyAngles is set.
'
' Returns coordinates for the transformed point.
'

' Firstly, assume default viewing angles if needed...
If (Flags And P3D_MyAngles) = 0 Then
  getViewAngles Angl, Angl2, 0
End If

' ...and cache the rotational scalars
Dim Rad As Single
Dim C1 As Single, C2 As Single
Dim S1 As Single, S2 As Single
Rad = Angl * m_Deg2Rad
C1 = Cos(Rad): S1 = Sin(Rad)

Rad = Angl2 * m_Deg2Rad
C2 = Cos(Rad): S2 = Sin(Rad)

' Also adjust for integer coordinates if requested
Dim P As D3DVECTOR, P2 As D3DVECTOR
If Flags And P3D_MakeInts Then
  P.X = Int(Point.X + 0.5)
  P.Y = Int(Point.Y + 0.5)
  P.Z = Int(Point.Z + 0.5)
Else
  P = Point
End If

' Next, allow for VTOL type coordinates
If Flags And P3D_IsVTOL Then
  P.X = -P.X
  P.Y = -P.Y
End If

' Then, adjust for the camera placement.
'
' This must also include and offsets performed by file (0,0,0) placement.
'
' CamPt = (camera) - (PIE offset)
' where the PIE offset is = (point's PIE offset) - (current PIE offset)
'
If Flags And P3D_UseCamPt Then
  
  If Flags And P3D_MakeInts Then
    P2.X = Int(CamPt.X + 0.5)
    P2.Y = Int(CamPt.Y + 0.5)
    P2.Z = Int(CamPt.Z + 0.5)
  Else
    P2 = CamPt
  End If
  
  P.X = P.X - P2.X
  P.Y = P.Y - P2.Y
  P.Z = P.Z - P2.Z
End If

' After all that, now perform the rotation!

' Older formula
'V.X = P.X * C1 - P.z * S1
'V.Y = (P.z * C1 + P.X * S1) * S2 + P.Y * C2
'V.z = (P.z * C1 + P.X * S1) * C2 - P.Y * S2

' Horizontal rotation
P2.X = P.X * C1 - P.Z * S1
P2.Y = P.Y
P2.Z = P.X * S1 + P.Z * C1

' Vertical rotation
Project3D.X = P2.X
Project3D.Y = P2.Z * S2 + P2.Y * C2
Project3D.Z = P2.Z * C2 - P2.Y * S2

' Finally, allow for screen-compatible Y coordinate...
If Flags And P3D_InvertY Then
  Project3D.Y = -Project3D.Y
End If

' ...and the current zoom level
If Flags And P3D_AddZoom Then
  Project3D.X = Project3D.X * Sys.ZoomScalar
  Project3D.Y = Project3D.Y * Sys.ZoomScalar
End If

End Function

Public Function file_LoadANI(ByVal FileName As String, ByRef Data As PieModel, ByRef Dialog As CommonDialog) As Long

Syntax:
' file_LoadANI(FileName, Data, PieFile)
'
'   FileName -- The name of a Warzone ANI file.
'   Data     -- Model to store the ANI & associated PIE data in.
'   PieFile  -- (return value) Name of PIE file this ANI file requires.


' Prepare the filename in case errors occur.
Dim FileTitle As String
FileTitle = Mid$(FileName, InStrRev(FileName, "\") + 1)

Dim AnimData As AnimData
AnimData.FileName = FileName

If FileName = "" Or Dir(FileName) = "" Then
  file_LoadANI = FE_LoadAni Or FE_FileNotFound
  Exit Function
End If

Dim FileNum As Integer, FileData As String
FileNum = FreeFile
Open FileName For Binary Access Read Lock Write As #FileNum
  FileData = String$(LOF(FileNum), 32)
  Get #FileNum, 1, FileData
Close #FileNum

' Eradicate comments from the ANI file.
Dim NO As Long, NS As Long, N2 As Long
Do While InStr(FileData, "/*") > 0
  NO = InStr(FileData, "/*")
  NS = InStr(FileData, "//")
  If NS > 0 Then ' Eliminate C-style single-line comment
    N2 = InStr(NS, FileData, vbCr)
    Debug.Assert N2
    FileData = Left$(FileData, NS - 1) & Mid$(FileData, N2)
  
  ElseIf NO > 0 Then ' Eliminate C-style block comment
    N2 = InStr(NO + 2, FileData, "*/")
    Debug.Assert N2
    FileData = Left$(FileData, NO - 1) & Mid$(FileData, N2 + 2)
  
  Else ' No comments to delete?  Then exit loop.
    Exit Do
  End If
Loop

Dim TempName As String
TempName = App.Path & "\_tmp.ani"

FileNum = FreeFile
Open TempName For Output As #FileNum: Close #FileNum
Open TempName For Binary Access Write Lock Write As #FileNum
  Put #FileNum, 1, FileData
Close #FileNum

Open TempName For Input As #FileNum

  Dim SrcPieFile As String
   
  ' Get a line from the ANI file, ignoring blank lines
  Dim LineData As String
  Do
    Line Input #FileNum, LineData
  Loop While LineData = ""
  
  Dim QLef As Integer, QRig As Integer
  Dim MinSeg As Integer, MaxSeg As Integer, ThisSeg As Integer
  Select Case Left$(LineData, 12)
  
  Case "ANIM3DFRAMES"
    AnimData.AnimType = ANIM3DFRAMES
    
    ' Extract filename
    QLef = InStr(LineData, """")
    QRig = InStr(QLef, LineData, ".pie""")
    SrcPieFile = Mid$(LineData, QLef + 1, QRig - QLef + 3)
    Debug.Assert Right$(SrcPieFile, 4) = ".pie"
    
    ' Extract # of frames
    QLef = QRig + 5
    QRig = InStr(QLef + 1, LineData, " ")
    Debug.Assert IsNumeric(Mid(LineData, QLef + 1, QRig - QLef - 1))
    AnimData.MaxFrames = Mid(LineData, QLef + 1, QRig - QLef - 1)
    
    ' Extract # frames/sec
    QLef = QRig
    QRig = InStr(QLef + 1, LineData & " ", " ")
    Debug.Assert IsNumeric(Mid(LineData, QLef + 1, QRig - QLef - 1))
    AnimData.FrameRate = Mid(LineData, QLef + 1, QRig - QLef - 1)
    
    ' Declare one object to use & its # of frames.
    ReDim AnimData.Segments(1)
    ReDim AnimData.Segments(1).Frames(AnimData.MaxFrames)
    MinSeg = 1: MaxSeg = 1
  
    ThisSeg = 1
    
  Case "ANIM3DTRANS "
    AnimData.AnimType = ANIM3DTRANS
    
    ' Extract filename
    QLef = InStr(LineData, """")
    QRig = InStr(QLef, LineData, ".pie""")
    SrcPieFile = Mid$(LineData, QLef + 1, QRig - QLef + 3)
    Debug.Assert Right$(SrcPieFile, 4) = ".pie"
    
    ' Extract # of frames
    QLef = QRig + 5
    QRig = InStr(QLef + 1, LineData, " ")
    Debug.Assert IsNumeric(Mid(LineData, QLef + 1, QRig - QLef - 1))
    AnimData.MaxFrames = Mid(LineData, QLef + 1, QRig - QLef - 1)
    
    ' Extract # frames/sec
    QLef = QRig
    QRig = InStr(QLef + 1, LineData, " ")
    Debug.Assert IsNumeric(Mid(LineData, QLef + 1, QRig - QLef - 1))
    AnimData.FrameRate = Mid(LineData, QLef + 1, QRig - QLef - 1)
    
    ' Extract # of objects to animate
    QLef = QRig
    QRig = InStr(QLef + 1, LineData & " ", " ")
    Debug.Assert IsNumeric(Mid(LineData, QLef + 1, QRig - QLef - 1))
    MaxSeg = Mid(LineData, QLef + 1, QRig - QLef - 1)
    ReDim AnimData.Segments(MaxSeg)
    For ThisSeg = 1 To MaxSeg
      ReDim AnimData.Segments(MaxSeg).Frames(AnimData.MaxFrames)
    Next ThisSeg
    
    MinSeg = 1
  
    ThisSeg = 0
  Case Else
    Debug.Assert False ' Invalid file header for WZ ANI file
    file_LoadANI = FE_Generic
    Exit Function
  End Select
  
  
  ' Continue loading data from file
  Do
    ' Debugging check
    Dim Posit As Long
    Posit = Seek(FileNum)
    Seek #FileNum, Posit
    
    Do
      Line Input #FileNum, LineData
      file_LoadPIE_TrimAll LineData, True
    Loop While LineData = ""
    
    Dim N As Integer
    N = InStr(LineData & " ", " ")
    Select Case Left$(LineData, N - 1)
    
    Case "//" ' Comments.  Usually ignored, but may include embedded PIE data.
      If InStr(LineData, "FileData=") = 4 Then
        Dim Ver As Integer
        
        On Error Resume Next
        Ver = 0
        Ver = Mid(LineData, 13)
        On Error GoTo 0
        
        Dim Temp As PieModel
        
        ReDim Temp.Files(1)
        If Ver Then file_LoadPCK_FileData FileNum, Ver, Temp.Files(1)
      
      End If
    Case "{" ' Opening brace -- ignore
    
    Case "}" ' Closing brace -- ignore
    
    Case "ANIMOBJECT" ' Anim3Dtrans only: specify object submodel.
      Debug.Assert AnimData.AnimType = ANIM3DTRANS
      
      ' Specify object submodel being used.
      QLef = N: QRig = InStr(QLef + 1, LineData, " """)
      ThisSeg = Mid(LineData, QLef + 1, QRig - QLef - 1) + 1 '
      Debug.Assert ThisSeg > 0
      Debug.Assert ThisSeg <= UBound(AnimData.Segments)
      
      ' Grab name of the submodel
      QLef = QRig + 1
      QRig = InStr(QLef + 1, LineData, """")
      AnimData.Segments(ThisSeg).Name = Mid$(LineData, QLef + 1, QRig - QLef - 1)
      ReDim AnimData.Segments(ThisSeg).Frames(AnimData.MaxFrames)
      
    Case Else ' ANI numeric data (offsets, rotation, etc.)
      Debug.Assert ThisSeg > 0
      Debug.Assert ThisSeg <= UBound(AnimData.Segments)
      
      Dim Terms() As String
      Terms() = Split(LineData, " ")
      Debug.Assert UBound(Terms) > 5
      
      Dim ThisFrame As Integer
      ThisFrame = Terms(0) + 1
      Debug.Assert ThisFrame > 0
      Debug.Assert ThisFrame <= AnimData.MaxFrames
      
      Dim Offset As D3DVECTOR, Rota As D3DVECTOR
      
      ' Load offset data.  Note conversion from WZ to PieSlicer axes.
      Offset.X = -Terms(1) / 1000
      Offset.Y = Terms(3) / 1000
      Offset.Z = -Terms(2) / 1000
      
      Rota.X = Terms(4) / 1000
      Rota.Y = Terms(6) / 1000
      Rota.Z = Terms(5) / 1000
      
      With AnimData.Segments(ThisSeg).Frames(ThisFrame)
        .Offset = Offset
        .Rotation = Rota
      End With
      
    End Select
    
  Loop Until EOF(FileNum)
Close #FileNum

If SrcPieFile = "" Then
  Debug.Assert False
  file_LoadANI = FE_LoadAni Or FE_NoAniSrcFile
  Exit Function
End If

' If we got this far, the ANI data was OK,
' now to search for its requisite PIE data

' Look in the same directory for the needed PIE file.  If not there, prompt for it.
'
Dim PieFileName As String
PieFileName = Left$(FileName, InStrRev(FileName, "\")) & SrcPieFile

If Dir(PieFileName) = "" Then
  Dialog.Flags = cdl_OpenFlags
  Dialog.Filter = SrcPieFile & " (line filter)|" & SrcPieFile & "|" & SrcPieFile & " (word filter)" '|" & SrcPieFile & "|All Pie Files (line filter)|*.pie|All Pie Files (word filter)|*.pie"
  Dialog.FilterIndex = 2
  Dialog.InitDir = FileName
  Dialog.FileName = SrcPieFile
  Dialog.DialogTitle = "Select PIE File to animate"
  
  On Error Resume Next
    Dialog.ShowOpen
    If Err.Number Then
      Debug.Assert Err.Number = cdlCancel
      file_LoadANI = FE_LoadAni Or FE_UserCancel
      Exit Function
    End If
  On Error GoTo 0
  
  PieFileName = Dialog.FileName
  Dialog.InitDir = Dialog.FileName
  Sys.UseLineFilter = (Dialog.FilterIndex And 1)
Else
  Sys.UseLineFilter = 0
End If


If file_GetType(PieFileName) <> Format_PieFile Then
  file_LoadANI = FE_LoadAni Or FE_BadAniSrcFile
  Exit Function
End If


Dim Code As Integer
ReDim Temp.Files(1)
'ReDim Temp.TexFiles(0)

Code = file_LoadPIE(PieFileName, Temp.Files(1))
If Code < 1 Then
  file_LoadANI = IIf(Code < 0, Code, FE_LoadAni Or FE_NoData)
  Exit Function
End If

Temp.Files(1).Anim = AnimData
ReDim Temp.Tex_Files(0)

Data = Temp

' Success!  Whew!
file_LoadANI = Code

End Function


Public Function file_SaveANI(ByRef Data As PieType) As Integer

Syntax:
' file_SaveANI(Data)
'
'   Data     -- PIE file to save attached ANI data for.
'

With Data.Anim
If .FileName = "" Or .AnimType = 0 Then
  Exit Function
End If

If Dir(.FileName) <> "" Then
  If GetAttr(.FileName) And vbReadOnly Then Exit Function
End If

Dim TempFile As String
file_CreateTempName TempFile

Dim MaxLevs As Integer
MaxLevs = enumMaxLevels(Data.Levs)

Dim File As Integer
File = FreeFile
Open TempFile For Output As #File

' Eradicate comments from the ANI file.

  Select Case .AnimType
  Case ANIM3DFRAMES
    'Debug.Assert .MaxFrames = UBound(Data.Levs)
    Print #File, "ANIM3DFRAMES " & .MaxFrames & " " & .FrameRate
    Print #File, "{"
    
    Dim F As Integer
    For F = 1 To .MaxFrames
    With .Segments(1).Frames(F)
      Print #File, "    " & (F - 1);
      Print #File, "   " & -.Offset.X * 1000 & " " & -.Offset.Z * 1000 & " " & .Offset.Y * 1000;
      Print #File, "   " & .Rotation.X * 1000 & " " & .Rotation.Z * 1000 & " " & .Rotation.Y * 1000;
      Print #File, "   1000 1000 1000"
    End With
    Next F
    
    Print #File, "}"
    
  Case ANIM3DTRANS
    
    Debug.Assert UBound(.Segments) = MaxLevs
    Print #File, "ANIM3DTRANS " & .MaxFrames & .FrameRate & UBound(.Segments)
    Print #File, "{"
    
    Dim S As Integer
    For S = 1 To UBound(.Segments)
    With .Segments(S)
      Print #File, "  ANIMOBJECT " & (S - 1) & """" & IIf(.Name = "", "Segment" & Format$(S, "00"), .Name) & """"
      Print #File, "  {"
      
      For F = 1 To MaxLevs
      With .Frames(F)
        Print #File, "    " & (F - 1);
        Print #File, "   " & -.Offset.X * 1000 & " " & -.Offset.Z * 1000 & " " & .Offset.Y * 1000;
        Print #File, "   " & .Rotation.X * 1000 & " " & .Rotation.Z * 1000 & " " & .Rotation.Y * 1000;
        Print #File, "   1000 1000 1000"
      End With
      Next F
      
      Print #File, "  }"
    End With
    Next S
    
  Case Else
  
  End Select
  
  file_SavePCK_FileData File, Data, True
  
Close #File
End With

FileCopy TempFile, Data.Anim.FileName
Kill TempFile

' Remove this statement when procedure is operational
'Debug.Assert False

End Function

Public Function v_DotP(V1 As D3DVECTOR, V2 As D3DVECTOR) As Single
'
' 3-dimensional dot product of 2 vectors.
'
' It's pretty simple, so if speed is important, use inline code instead.
'
Syntax:
' v_DotP(V1, V2)
'   V1, V2 -- The vectors to multiply.
'

v_DotP = V1.X * V2.X + V1.Y * V2.Y + V1.Z * V2.Z

End Function

Public Function v_CrossP(V1 As D3DVECTOR, V2 As D3DVECTOR, Optional ByVal Normalizer As Integer = 0) As D3DVECTOR
Dim Vout As D3DVECTOR
'
' Cross product function with optional magnitude adjustment.
'
' Used primarily to determine a polygon's normal vector.
'
Syntax:
'
' v_CrossP(V1, V2, Normalizer)
'   V1, V2      - Vectors to cross
'   Normalizer  - Length to normalize (scale) the result to.
'                   * 0 = don't bother

' Calculate cross product.
'
Dx_7.VectorCrossProduct v_CrossP, V1, V2

' Scale the result if necessary, and return it.
'
If Normalizer > 0 Then v_CrossP = v_Normalize(v_CrossP, Normalizer)

End Function

Public Function v_Project(Ray As D3DVECTOR, Axis As D3DVECTOR, ByRef Scalar As Single) As D3DVECTOR
'
' Projects one direction vector onto another.
'
Syntax:
'
' v_Project(Ray, Axis, Scalar)
'   Ray    - First vector.
'   Axis   - Second vector.
'   Scalar   (output) Magnitude scalar for the result.
'
' Returns the projected vector and sets Scalar to reflect its magnitude.
'

Scalar = v_DotP(Ray, Axis) / (Axis.X ^ 2 + Axis.Y ^ 2 + Axis.Z ^ 2)

With v_Project
  .X = Scalar * Axis.X
  .Y = Scalar * Axis.Y
  .Z = Scalar * Axis.Z
End With

End Function

Public Function v_Normalize(Vector As D3DVECTOR, Optional ByVal Length As Single = 1) As D3DVECTOR
'
' Scales a vector to unit or any other specified magnitude.
'
Syntax:
'
' v_Normalize(Vector, [Length] )
'   Vector -- Vector to normalize.
'   Length -- Desired length of the resultant vector.
'               * default=1 (unit magnitude)
'
' Returns a scaled version of Vector.
'

Dim OldLen As Single, Scalar As Single

' Ensure proper calling syntax
If Length = 0 Then Exit Function

' No need to evaluate zero vectors.
OldLen = v_Magna(Vector)
If OldLen = 0 Then Exit Function

Scalar = Length / OldLen

' Scale vector to unit length.
v_Normalize.X = Vector.X * Scalar
v_Normalize.Y = Vector.Y * Scalar
v_Normalize.Z = Vector.Z * Scalar

End Function

Public Function py_Flip(Polygon As PolygonData) As Long
Dim PolyTemp As PolygonData
Dim NP As Integer, Pt As Integer
'
' Flips which side a polygon is facing by reversing its vertex order.
'
' May be called on ANY polygon in any pie file data.
'
Syntax:
'
' py_Flip(Polygon)
'   Polygon -- The polygon to flip.
'
' No return value.


' Make a reference copy of the polygon
PolyTemp = Polygon

' Cache this value...
NP = UBound(Polygon.PtData)

' Constant specifying which vertex to start at.
' Only two values can be used:
'   1 - Flips all vertexes
'   2 - Flips all but the base vertex (preserves trianglefan shape)
Const SV As Integer = 2

' Reverse the vertex order.
For Pt = SV To NP
  Polygon.PtData(Pt) = PolyTemp.PtData(NP + SV - Pt)
Next Pt

' Debug check to make sure the first vertex is preserved when SV=2
Debug.Assert (SV = 2) Imp (Polygon.PtData(1).Index = PolyTemp.PtData(1).Index)

End Function

Public Function file_SavePCK(ByVal FileName As String) As Boolean
'
' Saves a pie model to disk as a PCK file.
'
Syntax:
'
' file_SavePCK(FileName)
'   FileName -- Full path/name of PCK file to save to.
'
' No return value.

Dim File As Integer, F As Integer, TempFile As String
Dim N As Integer, N2 As Integer, A As Integer
Dim PathName As String

' Record the pathname for the output file
N = InStrRev(FileName, "\")
PathName = Left$(FileName, N)

' Begin writing to a temp file.
file_CreateTempName TempFile
File = FreeFile
Open TempFile For Output As #File

Print #File, "PieDX " & (VerPCK \ 100) & "." & ((VerPCK \ 10) Mod 10) & "." & (VerPCK Mod 10)

Dim Dialog As CommonDialog
Set Dialog = formMaster.Dialog

For F = 1 To UBound(Model.Files)
With Model.Files(F)
  
  If .FileName = "" Then
    Dialog.Flags = cdl_SaveFlags Or IIf(GetAttr(.FileName), cdl_IsReadOnly, 0)
    Dialog.DialogTitle = "Save Pie File #" & F & " As"
    Dialog.Filter = "Warzone PIE File (*.pie)|*.pie"
    Dialog.FilterIndex = 1
  
    On Error Resume Next
      Err.Clear
      Dialog.ShowSave
      If Err.Number <> cdlCancel Then .FileName = Dialog.FileName
    On Error GoTo 0
  End If
  
  N = InStrRev(.FileName, PathName)
  Select Case .PieType
  Case TYPE_SINGLE: Print #File, "File=";
  Case TYPE_BODY: Print #File, "Body=";
  Case TYPE_PROP: Print #File, "Prop=";
  Case TYPE_GNDTUR: Print #File, "Tur0=";
  Case TYPE_AIRTUR: Print #File, "Tur1=";
  Case TYPE_MUZL: Print #File, "Tur2=";
  Case TYPE_MUZLCYB: Print #File, "Tur3=";
  Case TYPE_FLASH: Print #File, "Flas=";
  Case TYPE_PROJ: Print #File, "Proj=";
  Case TYPE_HIDE: Print #File, "Hide=";
  End Select
  
  If .FileName = "" Then ' Untitled PIE file
    Print #File, ""
  
  Else ' Named PIE file - save it to disk for use in WZ
    Print #File, Mid$(.FileName, N + Len(PathName))
    file_SavePIE .FileName, Model.Files(F), False
  
  End If
  
  ' Include a copy of the PIE file's data
  file_SavePCK_FileData File, Model.Files(F)
  
End With
Next F

' Only output vidmem textures that are NEEDED to render the PIE file
'
Dim T As Integer
For T = 1 To UBound(Sys.VidMem)
  
  For F = 1 To UBound(Model.Files)
  With Model.Files(F)
      
    ' If a texpage is needed, write it to file
    '
    ' The naming convention identifies whether
    ' it's a default or custom texture.
    '
    If .TexFile = T Then
      With Sys.VidMem(T)
        N2 = InStrRev(.FileName, "\")
        Print #File, "Texture=" & Mid$(.FileName, N2 + 1)
        F = UBound(Model.Files) + 1 ' Declare this loop complete
      End With
      
      Exit For
    End If
  
  End With
  Next F

Next T

Close #File

' Flag that the model, and all files, have been saved.
ClearAllChanges 1

FileCopy TempFile, FileName
Kill TempFile
End Function

Public Function file_SavePIE(OutFile As String, ByRef Data As PieType, ByVal OverPrompt As Integer) As Long
'
' Saves a single PIE file to disk.
'
Syntax:
'
' file_SavePIE(OutFile, Data, OverPrompt)
'   OutFile    - Full path/name of the PIE file to save to.
'   Data       - The PIE file's data to save.
'   OverPrompt - Whether Pie Slicer can overwrite the file
'                if it has the read-only attribute set.
'                  1 = Ask user for permission
'                        * ONLY allowed when using "Save Pie File" command
'                          on formMaster.  NOT allowed while saving
'                          a PCK file's model data, or when auto-saving.
'                  0 = Do not overwrite a readonly file.
'
' No return value.
'
' * This does NOT update the PIE file's timestamps
'   (Data.Changed and Data.Loaded).  That must be
'   performed as a separate step after successfully saving.
'
Dim File As Integer, FileStats As Long
Dim Lv As Integer, Pt As Integer
Dim Po As Integer, NumPts As Integer
Dim Cn As Integer, BS As Integer

If OutFile = "" Then Exit Function

' Read-only check
FileStats = file_GetAttrs(OutFile)
If FileStats And vbReadOnly Then
  If OverPrompt = 0 Then Exit Function
    
  If MsgBox("Caution:  " & Mid$(OutFile, InStrRev(OutFile, "\") + 1) & " is marked as read-only.  Are you sure you wish to save over it?", vbYesNo + vbDefaultButton2 + vbCritical) = vbNo Then Exit Function
End If

' Create a temp-file
Dim TempFile As String
file_CreateTempName TempFile

Dim St As Single
St = Timer

Dim PtList() As Integer, PyList As Integer, MaxLevs As Integer

' Count the submodels.
Dim Ct As Integer, PieFlags As Long
MaxLevs = enumSubmodels(Data.Levs, PtList(), PyList, PieFlags)
'Debug.Assert PieFlags
If PieFlags = 0 Then PieFlags = &H200
If MaxLevs < 0 Then MaxLevs = 1

File = FreeFile
Open TempFile For Output As #File
  
  ' Write PIE header data.
  Print #File, "PIE 2"
  Print #File, "TYPE " & Hex$(PieFlags)
  
  ' Write texture page & size of it.
  '
  If GetTexPageNum(Data.TexPage, "") = 0 Then ' Support NOTEXTURE PIE files.
    Print #File, "NOTEXTURE 0 NOTEXTURE 0 0"
  Else
  
    Print #File, "TEXTURE 0 " & Data.TexPage & " 256 256"
  End If
  
  ' Begin writing submodel data...
  Dim MaxPt As Integer, MaxPy As Integer
  Debug.Assert MaxLevs
  Print #File, "LEVELS " & MaxLevs
  
  Dim SubF As Integer
  SubF = stat_Sub1
  
  For Lv% = 1 To MaxLevs
  With Data.Levs
    Print #File, "LEVEL " & Lv%
  
    ' Compile point data for the submodel.
    ' All points required for the submodel's polygons will be written to file.
    ' Submodel 1 also contains all points not attached to any polygons
    '
    Dim PtMatrix() As Integer
    ReDim PtMatrix(UBound(.Points))
    
    Dim Stri As String
    Stri = "": MaxPt = 0
    
    For Pt% = 1 To UBound(.Points)
    With .Points(Pt%)
    If (PtList(Pt) And SubF) Or (PtList(Pt) = 0 And Lv = 1) Then
        
      ' Record th PIE file index of this point.
      ' We'll use it later when writing polygon data.
      MaxPt = MaxPt + 1
      PtMatrix(Pt) = MaxPt
      
      Stri = Stri & vbTab & -Int(.Loc.X + 0.5) & " " & Int(.Loc.Y + 0.5) & " " & -Int(.Loc.Z + 0.5) & vbCrLf
        
    End If
    End With
    Next Pt%
  
    ' Output the points data section - header & data
    Print #File, "POINTS " & MaxPt
    Print #File, Stri
    
    ' Compile all polygon data.
    ' Only polygons included in the current submodel will be written.
    '
    Stri = "": MaxPy = 0
    
    For Po = 1 To UBound(.Polys)
    With .Polys(Po%)
    If .Submodel = Lv Then
      
      ' Record the # of polygons written.
      MaxPy = MaxPy + 1
      
      ' Bugfix -- strip team-colors flag unless there's >1 frame
      If (.Flags And &H4000&) And (.TeamC.Animated < 2) Then
        .Flags = .Flags And Not &H4000&
      End If
      
      ' Begin compiling polygon data, starting with flags & num pts
      NumPts = UBound(.PtData)
      
      Stri = Stri & vbTab & Hex$(.Flags) & " " & CStr(NumPts) & " "
  
      For Pt = 1 To NumPts
      With .PtData(Pt)
        
        Debug.Assert PtMatrix(.Index) > 0
        Debug.Assert PtList(.Index) And SubF
        
        Stri = Stri & Format$(PtMatrix(.Index) - 1, "###0") & " "
      
      End With
      Next Pt
      
      ' Animated polygons only:  Write animation data.
      '
      ' Note the bugfix for time, width & height values
      '
      If .Flags And &H4000 Then
        If .TeamC.AnimTime = 0 Then .TeamC.AnimTime = 1
        
        If .TeamC.AnimW > 256 Then .TeamC.AnimW = .TeamC.AnimW Mod 256
        If .TeamC.AnimW < 1 Then .TeamC.AnimW = 256
        
        If .TeamC.AnimH < 1 Then .TeamC.AnimH = 1
        Stri = Stri & .TeamC.Animated & " " & _
                      .TeamC.AnimTime & " " & _
                      .TeamC.AnimW & " " & _
                      .TeamC.AnimH & " "
      End If
  
      ' Format texture alignment data for type +200 and +8000 polys
      If .Flags And &H8200& Then
        For Pt% = 1 To NumPts
          Stri$ = Stri$ & Int(.PtData(Pt%).Loc.X + 0.5) & " " & _
                          Int(.PtData(Pt%).Loc.Y + 0.5) & " "
        Next Pt%
      End If
      
      Stri = Stri & vbCrLf
      
    ' Next polygon.
    End If
    End With
    Next Po
  
    ' Write the polygon data to file now
    '
    Print #File, "POLYGONS " & MaxPy
    Print #File, Stri
  
    ' Write connectors data (if any).
    For Cn% = 1 To UBound(.Ccs)
      If Cn% = 1 Then Print #File, "CONNECTORS " & UBound(.Ccs)
      With .Ccs(Cn%)
        
        Print #File, vbTab & -Int(.Loc.X + 0.5) & " " & _
                              -Int(.Loc.Z + 0.5) & " " & _
                               Int(.Loc.Y + 0.5)
      End With
    Next Cn%
  
    SubF = SubF * 2
  End With
  Next Lv%

Close File

file_BuildPath OutFile

' Allow overwriting of a read-only file
If FileStats And vbReadOnly Then SetAttr OutFile, (FileStats And Not vbReadOnly)

' Save to the file on disk
FileCopy TempFile, OutFile
Kill TempFile

' Restore the file's attributes
SetAttr OutFile, FileStats

Debug.Print "SaveFile: " & (Timer - St)
' Save associated ANI data, if applicable
If Data.Anim.AnimType > 0 Then file_SaveANI Data

End Function

Public Function py_getNormalV(ByRef Lev As LevelType, ByVal Index As Integer, Optional ByVal Start As Integer = 0, Optional ByVal Scalar As Integer = 25) As D3DVECTOR
'
' Computes the normal vector for a polygon or one of its trianglefan components.
'
' * This routine is time-critical
'
Syntax:
'
' py_getNormalV(Level, Index, [Start], [Scalar])
'
'   Level    - Submodel containing the polygon of interest.
'   Index    - Polygon index ##.
'   [Start]  - 0 = Process all trianglefans in the polygon
'              n = Process the trianglefan starting at vertex n
'   [Scalar] - Desired magnitude of the result.
'               * 0 = don't rescale
'               * default = 25 (for blitting purposes)

' Returns a vector representing the polygon's normal direction.

Dim Vector1 As D3DVECTOR, Vector2 As D3DVECTOR
Dim DrawP1 As D3DVECTOR, SumVec As D3DVECTOR
Dim MP As Integer, A As Integer
Dim Polygon As PolygonData
Dim PtB As D3DVECTOR, Ptr As D3DVECTOR
Dim Indx1 As Integer

' Cache the polygon
Polygon = Lev.Polys(Index)

' Grab # of points in current polygon for reference.
MP = UBound(Polygon.PtData)
If Start = MP Then Start = MP - 1

' Cache the base vertex of this poly
Indx1 = Polygon.PtData(1).Index
PtB = Lev.Points(Indx1).Loc

' Begin computing
Dim Min As Long, Max As Long
If Start < 2 Then
  Min = 2
  Max = MP - 1
Else
  Min = Start
  Max = Start
End If

For A% = Min To Max
  
  ' Create first vector (base to vertex A).
  Ptr = Lev.Points(Polygon.PtData(A).Index).Loc
  Vector1.X = Ptr.X - PtB.X
  Vector1.Y = Ptr.Y - PtB.Y
  Vector1.Z = Ptr.Z - PtB.Z

  ' Create second vector (base to vertex A+1)
  If A = MP Then
    Ptr = Lev.Points(Polygon.PtData(1).Index).Loc
  Else
    Ptr = Lev.Points(Polygon.PtData(A + 1).Index).Loc
  End If
  Vector2.X = Ptr.X - PtB.X
  Vector2.Y = Ptr.Y - PtB.Y
  Vector2.Z = Ptr.Z - PtB.Z

  ' Calculate the normal.
  DrawP1 = v_Normalize(v_CrossP(Vector1, Vector2), 1)

  ' Add it to the sum.
  SumVec.X = SumVec.X + DrawP1.X
  SumVec.Y = SumVec.Y + DrawP1.Y
  SumVec.Z = SumVec.Z + DrawP1.Z

Next A%

If Scalar > 0 Then
  py_getNormalV = v_Normalize(SumVec, Scalar)
Else
  py_getNormalV = SumVec
End If

End Function


Public Function pt_Rotate3D(Point1 As D3DVECTOR, CenterP As D3DVECTOR, Axis As Integer, Optional ByVal Rate = 15) As D3DVECTOR
'
' Rotates a single point around another and returns the new location.
'
Syntax:
'
' pt_Rotate3D(Point1, CenterP, Axis, [Rate])
'   Point1  - The point to rotate.
'   CenterP - The point to rotate around.
'   Axis    - One of the following:
'               1 = X-axis rotation
'               2 = Y-axis rotation
'               3 = Z-axis rotation
'   Rate    - Number of degrees to rotate by.
'               * Default = 15
'

Dim TempP As D3DVECTOR, ResP As D3DVECTOR
Dim Rad As Single
Rad = Rate * m_Deg2Rad

' Convert points to relative coordinates.
ResP = Point1
TempP.X = Point1.X - CenterP.X
TempP.Y = Point1.Y - CenterP.Y
TempP.Z = Point1.Z - CenterP.Z

Dim CS As Single, SN As Single
CS = Cos(Rad): SN = Sin(Rad)

Select Case Axis
Case 1 ' X axis rotation.
  ResP.X = TempP.X + CenterP.X
  ResP.Y = TempP.Y * CS + TempP.Z * SN + CenterP.Y
  ResP.Z = TempP.Z * CS - TempP.Y * SN + CenterP.Z
Case 2 ' Y axis rotation.
  ResP.X = TempP.X * CS - TempP.Z * SN + CenterP.X
  ResP.Y = TempP.Y + CenterP.Y
  ResP.Z = TempP.Z * CS + TempP.X * SN + CenterP.Z
Case 3 ' Z axis rotation.
  ResP.X = TempP.X * CS - TempP.Y * SN + CenterP.X
  ResP.Y = TempP.Y * CS + TempP.X * SN + CenterP.Y
  ResP.Z = TempP.Z + CenterP.Z
End Select

pt_Rotate3D = ResP

End Function
'
Public Function pt_Scale3D(ScaleP As D3DVECTOR, Center As D3DVECTOR, Optional ByVal ScalX As Single = 1, Optional ByVal ScalY As Single = 1, Optional ByVal ScalZ As Single = 1) As D3DVECTOR
Dim TempP As D3DVECTOR
' Scales / flips a single point in 3-D space.

Syntax:
' pt_Scale3D(point, center, [x-fac], [y-fac], [z-fac])
'  point  - The point to scale, x-y-z format.
'  center - The center of scaling, x-y-z format.
'  x-fac  - Scaling factor about the x-axis.
'  y-fac  - Scaling factor about the y-axis.
'  z-fac  - Scaling factor about the z-axis.


' Calculate the results
TempP = ScaleP
TempP.X = (ScaleP.X - Center.X) * ScalX + Center.X
TempP.Y = (ScaleP.Y - Center.Y) * ScalY + Center.Y
TempP.Z = (ScaleP.Z - Center.Z) * ScalZ + Center.Z

' Assign the results
pt_Scale3D = TempP

End Function

Public Function enumSubmodels(ByRef Lev As LevelType, ByRef PtFlags() As Integer, ByRef pyFlags As Integer, ByRef Flags As Long) As Integer
'
' Enumerates all points required by the polygons in a submodel.
'
Syntax:
'
' enumSubmodels(Level, ptFlags(), pyFlags, Flags)
'   Level     - PIE model data of interest.
'   ptFlags() - Array to hold the results in.
'   pyFlags   - Return value indicating which submodels have polygons.
'   Flags     - Additional return value indicating PIE type:
'                 200 - Normal
'                 220 - Includes no-culled polygons
'
' Returns the maximum submodel found (=1 or higher), and:
'   - pyList    returns flags saying which submodels have polygon data.
'   - ptList(n) returns flags saying which submodels require point #n
'

' Create a quick-reference list of bitflags.
' Not really an easy way to do this in VB...
Dim P1 As Integer, Vx As Integer
Dim getSub(1 To 10) As Integer
'getSub(0) = stat_Sub1
getSub(1) = stat_Sub1
For P1 = 2 To 10
  getSub(P1) = getSub(P1 - 1) * 2
Next P1

enumSubmodels = 1

' Initialize the array to match the model data
ReDim PtFlags(UBound(Lev.Points))

' Begin looping through the polygons in the model
For P1 = 1 To UBound(Lev.Polys)
With Lev.Polys(P1)

  ' Enumerate flags used by the polygons.
  If .Flags And &H200 Then Flags = Flags Or &H200
  If .Flags And &H2000 Then Flags = Flags Or &H20
  
  ' Record maximum # of submodels found
  If .Submodel < 1 Then .Submodel = 1
  If .Submodel > enumSubmodels Then enumSubmodels = .Submodel
  
  ' Flag index zero to record that polygons exist in this submodel
  Dim SubF As Integer
  SubF = getSub(.Submodel)
  
  pyFlags = pyFlags Or SubF

  ' Iterate through the vertex data
  For Vx = 1 To UBound(.PtData)
  With .PtData(Vx)
  
    ' Flag a point as belonging in this submodel.
    Debug.Assert .Index > 0
    PtFlags(.Index) = PtFlags(.Index) Or SubF
    
  End With
  Next Vx

End With
Next P1

End Function

Public Function enumSelection(Lev As LevelType, ByVal Query As Long, ByVal Index As Long, Results() As Integer) As Integer
'
' Enumerates various data of interest in PIE data.
'
Syntax:
'
' enumSelection(Lev, Query, Index, Results)
'   Lev     - PIE data to evaluate.
'
'   Query   - What to search for:
'               * One of:
'                   Get_Pts  - Look for points
'                   Get_Pys  - Look for polygons
'               * Plus (OR) one of:
'                   Get_CurOnly - Point/polygon #Index
'                   Get_AllVisi - All visible points/polys
'                   Get_AllSel  - All visible+selected points/polys
'                   Get_WithCur - Points/polys attached to item #Index
'                   Get_WithSel - Points/polys attached to other selection
'                   Get_InSub   - Points/polys in submodel #Index
'
'   Index   - Additional data.
'               when +Get_InSub,  Index = submodel to search
'               when +Get_CurOnly Index = point/polygon index of interest
'               when +Get_WithCur Index = point/polygon index to review
'               otherwise, no effect.
'
'   Results - Array of values to return.
'               * Results(0) dictates how to interpret the return values:
'                   0       - Null data
'                   Get_Pts - Results(#) are point indexes
'                   Get_Pys - Results(#) are polygon indexes
'
' When successful:
'  - Sets Results() to the data found by the query, if any
'  - Returns the # of that data which is already selected
'
' If no data was found by the query, sets Results(0) = 0 and returns zero.
'
' If the query is invalid, returns -1 and sets Results(0) = 0.
'

' General count vars.
Dim P1 As Integer, P2 As Integer, P3 As Integer

' Index for the Results() array
Dim RsIndx As Integer

ReDim Results(0) As Integer

' Temp array.
Dim Temp() As Integer

Select Case Query And Get_HowMask
Case Get_CurOnly
  ' Evaluate current point or polygon only.
  ' If time is critical, use inline code instead
  '
  ' * Index = point/polygon index
  '
  If Index < 1 Then enumSelection = -1: Exit Function
  ReDim Results(1)
  
  Select Case Query And Get_QueryMask
  Case Get_Pys
    If (Lev.Polys(Index).StatFlags And stat_SelMask) = stat_Sel Then enumSelection = 1
    Results(0) = Get_Pys
    Results(1) = ThisPoly
    
  Case Get_Pts
    If (Lev.Points(Index).StatFlags And stat_SelMask) = stat_Sel Then enumSelection = 1
    Results(0) = Get_Pts
    Results(1) = ThisPoint
  
  Case Else
    enumSelection = -1
    Exit Function
  End Select
  
Case Get_WithCur
  ' Evaluate data attached to a given point or polygon.
  '
  ' * Index = point/polygon index
  '
  If Index < 1 Then enumSelection = -1: Exit Function
  
  Select Case Query And Get_QueryMask
  Case Get_Pys
    ' Examine polygons attached to the current point.
    '
    
    ' Results() will store polygon indexes
    ReDim Results(UBound(Lev.Polys))
    Results(0) = Get_Pys
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
      
      For P2 = 1 To UBound(.PtData)
      P3 = .PtData(P2).Index
      If P3 = Index Then
        
        RsIndx = RsIndx + 1
        Results(RsIndx) = P1
        
        ' Count selected polygons
        If (.StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
        
        Exit For
      End If
      Next P2
    
    End With
    Next P1
    
    ' Crop the results array to fit.
    ReDim Preserve Results(RsIndx)
    
  Case Get_Pts
    ' Examine points used by the current polygon.
    '
    
    enumSelection = 0
    
    With Lev.Polys(ThisPoly)
      
      ' Results() will store vertex (point) indexes
      P1 = UBound(.PtData)
      ReDim Results(P1)
      
      Results(0) = Get_Pts
    
      ' Loop through & record the polygon vertexes
      For P2 = 1 To P1
        P3 = .PtData(P2).Index
        Results(P2) = P3
        
        ' Count the # of selected vertexes.
        If (Lev.Points(P3).StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
      Next P2
    End With
  
  Case Else
    enumSelection = -1
    Exit Function
  
  End Select

Case Get_WithSel
  ' Evaluate data attached to other selected data.
  '
  ' * Requires that some data is already selected
  '
  Select Case Query And Get_QueryMask
  Case Get_Pts
    ' Note points used by all selected polygons.
    '
    ReDim Temp(UBound(Lev.Points))
    ReDim Results(UBound(Lev.Points))
    Results(0) = Get_Pts
    RsIndx = 0
  
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
    If (.StatFlags And stat_SelMask) = stat_Sel Then
        
      For P2 = 1 To UBound(.PtData)
        P3 = .PtData(P2).Index
        
        ' Record new vertexes only.
        If Temp(P3) = 0 Then
          RsIndx = RsIndx + 1
          Results(RsIndx) = P3
          Temp(P3) = 1
          
          ' Count selected vertices
          If (Lev.Points(P3).StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
        End If
      Next P2
    
    End If
    End With
    Next P1
    
    ReDim Preserve Results(RsIndx)
  
  Case Get_Pys
    ' Note all polygons attached to selected points
    '
    ReDim Results(UBound(Lev.Polys))
    Results(0) = Get_Pys
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
        
      For P2 = 1 To UBound(.PtData)
        P3 = .PtData(P2).Index
        
        ' If a point (vertex) is selected, record the polygon.
        If (Lev.Points(P3).StatFlags And stat_SelMask) = stat_Sel Then
          RsIndx = RsIndx + 1
          Results(RsIndx) = P1
          
          ' Count if the polygon is selected
          If (.StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
          Exit For ' Continue to next polygon
        End If
      Next P2
      
    End With
    Next P1
    
    ReDim Preserve Results(RsIndx)
  
  Case Else
    enumSelection = -1
    Exit Function
  
  End Select
  
Case Get_InSub
  ' Evaluate data contained in a given submodel.
  '
  ' Since only polygons have a submodel stat,
  ' this will always loop through the polygons array.
  '
  ' * Index = submodel ##
  '
  If Index < 1 Then
    enumSelection = -1
    Exit Function
  End If
  
  Select Case Query And Get_QueryMask
  Case Get_Pys
    ' Record polygons in the submodel
    '
    ReDim Results(UBound(Lev.Polys))
    Results(0) = Get_Pys
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
    If .Submodel = Index Then
      
      ' Record this polygon.
      RsIndx = RsIndx + 1
      Results(RsIndx) = P1
        
      ' Count selected polygons
      If (.StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
    
    End If
    End With
    Next P1
    
  Case Get_Pts
    ' Record points attached to polygons in the submodel
    '
    ReDim Temp(UBound(Lev.Points))
    ReDim Results(UBound(Lev.Points))
    Results(0) = Get_Pts
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
    If .Submodel = Index Then
      
      ' Record the vertexes
      For P2 = 1 To UBound(.PtData)
        P3 = .PtData(P2).Index
        If Temp(P3) = 0 Then
          RsIndx = RsIndx + 1
          Results(RsIndx) = P3
          
          Temp(P3) = 1
        End If
      Next P2
      
      ' Count selected points
      If (Lev.Points(P3).StatFlags And stat_SelMask) = stat_Sel Then enumSelection = enumSelection + 1
    
    End If
    End With
    Next P1
  
  Case Else
    enumSelection = -1
    Exit Function
  
  End Select
  
  
Case Get_AllVisi
  ' Evaluate all visible data in a PIE model.
  '
  Select Case Query And Get_QueryMask
  Case Get_Pys
    ' Record visible polygons
    '
    ReDim Results(UBound(Lev.Polys))
    Results(0) = Get_Pys
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
    If (.StatFlags And stat_Hide) = 0 Then
      
      ' Record this polygon.
      RsIndx = RsIndx + 1
      Results(RsIndx) = P1
      
      ' Count selected polygons.
      If .StatFlags And stat_SelMask Then enumSelection = enumSelection + 1
      
    End If
    End With
    Next P1
    
    ' Crop the array
    ReDim Preserve Results(RsIndx)
    
  Case Get_Pts
    ' Record visible points
    '
    ReDim Results(UBound(Lev.Points))
    Results(0) = Get_Pts
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Points)
    With Lev.Points(P1)
    If (.StatFlags And stat_Hide) = 0 Then
      
      ' Note this point.
      RsIndx = RsIndx + 1
      Results(RsIndx) = P1
        
      ' Count selected points.
      If .StatFlags And stat_SelMask Then enumSelection = enumSelection + 1
      
    End If
    End With
    Next P1
  
    ' Crop the array
    ReDim Preserve Results(RsIndx)
    
  Case Else
    enumSelection = -1
    Exit Function
  End Select

Case Get_AllSel
  ' Evaluate all selected+visible data in a PIE model.
  '
  Select Case Query And Get_QueryMask
  Case Get_Pys
    '
    ' Find selected polys
    '
    ReDim Results(UBound(Lev.Polys))
    Results(0) = Get_Pys
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Polys)
    With Lev.Polys(P1)
    If (.StatFlags And stat_Hide) = 0 Then
      
      ' Count selected polygons.
      If .StatFlags And stat_SelMask Then
        RsIndx = RsIndx + 1
        Results(RsIndx) = P1
        enumSelection = enumSelection + 1
      End If
      
    End If
    End With
    Next P1
    
    ' Crop the array
    ReDim Preserve Results(RsIndx)
    
  Case Get_Pts
    '
    ' Find selected points
    '
    ReDim Results(UBound(Lev.Points))
    Results(0) = Get_Pts
    RsIndx = 0
    
    For P1 = 1 To UBound(Lev.Points)
    With Lev.Points(P1)
    If (.StatFlags And stat_Hide) = 0 Then
      
      ' Count selected pts
      If .StatFlags And stat_SelMask Then
        RsIndx = RsIndx + 1
        Results(RsIndx) = P1
        enumSelection = enumSelection + 1
      End If
      
    End If
    End With
    Next P1
  
    ' Crop the array
    ReDim Preserve Results(RsIndx)
    
  Case Else
    enumSelection = -1
    Exit Function
  End Select

Case Else
  Debug.Assert False
  enumSelection = -1
  Exit Function
  
End Select

End Function

Public Function file_LoadCak(NameOfCake As String, ByRef Data As PieType) As Long
Dim TexX As Integer, TexY As Integer
Dim MaxLevs As Integer
Dim Lv As Integer, Pt As Integer, Po As Integer, Cn As Integer
Dim LoPts As Integer, LoPys As Integer, MaxConn As Integer
Dim NumPts As Integer, Poy As Integer
Dim FrameT As Integer, Index As Integer
Dim FileNam As String
Dim Fil As Integer
Dim HeaderL As String, LineN As Integer

' Loads CAK data from disk and returns # of levels.

Fil = FreeFile

Open NameOfCake For Input As #Fil
On Error GoTo BadFile
' Get CAK header data.
Line Input #Fil, HeaderL
LineN = 1
If Len(HeaderL) <> 10 Or Left(HeaderL, 4) <> "CAKE" Then
  file_LoadCak = FE_Generic
  Exit Function
End If

' Get texture page & its dimensions.
'ReDim Data.Levs(1)

Input #Fil, Data.TexPage, TexX%, TexY%
LineN = 2

Data.FileName = FileNam

' Get # of submodels and prepare data array.
Input #Fil, MaxLevs%
LineN = 3
'ReDim Data.Levs(0)
'Data.Levs(0) = getBlankLevel(BlankLev_New0)

Data.Levs = getBlankLevel(BlankLev_New0)

For Lv = 1 To MaxLevs%
'  ReDim Preserve Data.Levs(Lv)
  
  Pt = 0: LoPts% = 0
  Po = 0: LoPys% = 0
  Cn = 0: MaxConn% = 0

  ' Get # of points and prepare data arrays.
  Input #Fil, LoPts%
  LineN = LineN + 1
  ReDim Preserve Data.Levs.Points(LoPts%)
  ' Get point data.
  For Pt = 1 To LoPts%
    LineN = LineN + 1
    Dim Vec As D3DVECTOR
    Input #Fil, Vec.X, Vec.Y, Vec.Z
    With Data.Levs.Points(Pt)
      .Loc.X = -Vec.X
      .Loc.Y = Vec.Y
      .Loc.Z = -Vec.Z
    End With
  Next Pt

  ' Get # of polygons, prepare data arrays.
  LineN = LineN + 1
  Input #Fil, LoPys%
  ReDim Preserve Data.Levs.Polys(LoPys%)

  ' Get polygon data.
  For Po = 1 To LoPys%
    ' Input polygon type.
    LineN = LineN + 1
    With Data.Levs.Polys(Po)
      Input #Fil, .Flags
  
      ' Input polygon's # of points
      ' and prepare data arrays.
      
      Input #Fil, NumPts
      ReDim .PtData(NumPts)
  
      On Error Resume Next
      ' Input polygon vertex and texture data.
      For Poy = 1 To NumPts
        Input #Fil, Index%
        .PtData(Poy).Index = Index% + 1
        Input #Fil, .PtData(Poy).Loc.X
        Input #Fil, .PtData(Poy).Loc.Y
      Next Poy
  
      ' Input animation data for animated polygons only
      
      If .Flags And &H4000 Then
        Input #Fil, .TeamC.Animated, .TeamC.AnimTime, .TeamC.AnimW, .TeamC.AnimH
      End If
      
      If Err.Number Then
        MsgBox "Could not load the texture data for polygon #" & Po & ".  It will need re-skinning."
      End If
      On Error GoTo BadFile
    End With
  ' Next polygon.
  Next Po

  ' Get # of connectors and prepare data array.
  LineN = LineN + 1
  Input #Fil, MaxConn%
  ReDim Preserve Data.Levs.Ccs(MaxConn%)
  For Cn = 1 To MaxConn%
    LineN = LineN + 1
    Input #Fil, Vec.X, Vec.Y, Vec.Z
    With Data.Levs.Ccs(Cn)
      .Loc.X = -Vec.X
      .Loc.Y = Vec.Y
      .Loc.Z = -Vec.Z
    End With
  Next Cn

  ' No BSP data to load from CAK files.
Next Lv
Close Fil

file_LoadCak = MaxLevs%

Exit Function

BadFile:
' Give an informative error message.
FileNam = NameOfCake
Do Until InStr(FileNam, "\") = 0
  FileNam = Mid(FileNam, InStr(FileNam, "\") + 1)
Loop

' Unknown error; give an informative error message.  PIE files have some flags still unknown....
formMaster.SetStatus 0, "5|Errors detected in """ & FileNam & """.  See ""PIESlicer.log"" for details.", FormMain.statBar
Beep
If Dir(App.Path & "\PIESlicer.log") <> "" Then _
  Kill App.Path & "\PIESlicer.log"
Open App.Path & "\PIESlicer.log" For Output As #10
Print #10, "Errors encountered while loading """ & FileNam & """ : "
Print #10,
Print #10, "Line:  " & LineN
Print #10, "Error: " & Err.Description
Print #10, ""
Print #10, " If you would like technical assistance with this error, contact Strata of NEWST (stratadrake@hotmail.com) with the PIE file and this message."
Close #10


Close #1

' If an error code occurs, invalidate the loaded
' data (so it won't corrupt the current data)
ReDim Data.Levs.Points(0)
ReDim Data.Levs.Polys(0)
ReDim Data.Levs.Ccs(0)

End Function


Public Function CopySelected(FromHere As LevelType, ToHere As LevelType, ByVal DoPolys As Boolean, ByVal EndValue As Integer) As Long
' Copies *selected* data from one submodel to another.

Syntax:
' CopySelected(FromHere, ToHere, DoPolys, EndValue)
'   FromHere - A submodel to copy data from.
'   ToHere   - A submodel to place the data in.
'   DoPolys  - 0=copy points only
'              1=copy all data
'   EndValue - New select status for the copied data
'                * 0 or stat_Sel

' Create reference vars
Dim PtMax As Integer, PtOffset As Integer
Dim PyMax As Integer, PyOffset As Integer
PtMax = UBound(FromHere.Points) ' # of points to copy
PyMax = UBound(FromHere.Polys) ' # of polygons to copy

PtOffset = UBound(ToHere.Points) ' offset value for pasting
PyOffset = UBound(ToHere.Polys) ' offest value for pasting

' Create an array of offset-values to hold the point data.
' For each point we copy, this will determine where to store it.
ReDim Transform(PtMax) As Integer
  
Dim Po As Integer, Pt As Integer
Dim A As Integer, PtIndx As Integer

' Record of points required by selected polygons
ReDim Selec(PtMax) As Boolean

' Note any points attached to selected polygons
'
If DoPolys Then
  For Po = 1 To PyMax
    With FromHere.Polys(Po)
      If (.StatFlags And stat_SelMask) = stat_Sel Then
        For A = 1 To UBound(.PtData)
          Selec(.PtData(A).Index) = True
        Next A
      End If
    End With
  Next Po
End If

' Copy all selected & required points.
PtIndx = 0
For Pt = 1 To PtMax
  If Selec(Pt) Or (FromHere.Points(Pt).StatFlags And stat_SelMask) = stat_Sel Then
    PtIndx = PtIndx + 1

    ' Record which new point corresponds to this selected point
    Transform(Pt) = PtOffset + PtIndx

    ' Copy the point into the destination array
    ReDim Preserve ToHere.Points(PtOffset + PtIndx)
    ToHere.Points(PtOffset + PtIndx) = FromHere.Points(Pt)
    ToHere.Points(PtOffset + PtIndx).StatFlags = EndValue
  End If
Next Pt

' Copy the selected polygons.
If DoPolys Then
  Dim PyIndx As Integer, Py As Integer, Polydx As Integer
  PyIndx = 0
  For Py = 1 To PyMax
    If (FromHere.Polys(Py).StatFlags And stat_SelMask) = stat_Sel Then
      ' Create some index values.
      PyIndx = PyIndx + 1
      Polydx = (PyOffset + PyIndx)

      ' Copy the selected polygon
      ReDim Preserve ToHere.Polys(Polydx)
      ToHere.Polys(Polydx) = FromHere.Polys(Py)
      ToHere.Polys(Polydx).StatFlags = EndValue

      ' Change its point index, using the
      ' "Transform" array as a database.
      Dim PtIndx1 As Integer
      For Pt = 1 To UBound(FromHere.Polys(Py).PtData)
        PtIndx1 = FromHere.Polys(Py).PtData(Pt).Index
        ToHere.Polys(Polydx).PtData(Pt).Index = Transform(PtIndx1)
      Next Pt
    End If
  Next Py
End If

' Return the # of copied data as two packed integers
CopySelected = PyIndx * &H10000 Or PtIndx

End Function

Public Function v_Subtract(Valu As D3DVECTOR, FromValue As D3DVECTOR) As D3DVECTOR
' Creates a directional vector by subtracting one point from another.
'
' When speed is important, use inline code instead.
'
Dim Result As D3DVECTOR
Result.X = FromValue.X - Valu.X
Result.Y = FromValue.Y - Valu.Y
Result.Z = FromValue.Z - Valu.Z
v_Subtract = Result

End Function

Public Function v_Magna(Vector As D3DVECTOR, Optional Pwr As Integer = 1) As Single
'
' Returns the magnitude (length) of a vector, or the magnitude squared.
'
' * This function is time-critical
'
Syntax:
'
' v_Magna(Vector, [Pwr])
'   Vector - Vector to evaluate
'   Pwr    - 2=return magnitude^2 instead of magnitude
'              (slightly faster)

With Vector
  v_Magna = (.X * .X) + (.Y * .Y) + (.Z * .Z)
End With
  
If Pwr <> 2 Then v_Magna = Sqr(v_Magna)

End Function

Public Function pt_CenterOf(ByVal Flags As Long, Lev As LevelType, ByVal Index As Integer, ByRef Code As Long) As D3DVECTOR
Dim TempP As D3DVECTOR
'
' Computes the arithmatic mean for various points in a PIE file.
'
Syntax:
'
' pt_CenterOf(Lev, Flags, Index, Code)
'
'  Flags - How to evaluate the submodel:
'            PTC_AllSel  - Evaluate all selected points
'            PTC_CurPy   - Evaluate all vertexes in polygon #Index
'            PTC_TexPy   - Evaluate texture data in polygon #Index
'            PTC_IntMath - Use WZ PIE precision
'
'  Lev   - Submodel to refer to.
'  Index - Polygon index to use.
'            * Valid only with PTC_CurPy or PTC_TexPy
'
'  Code  - (output) For debugging purposes only.
'            * -1 for bad calling syntax
'
' Returns the arithmatic mean of all points evaluated.
'   * If PTC_TexPy is set, .z member is not used
'
Dim PtList() As Integer

Select Case Flags And PTC_GetMask
Case PTC_AllSel
  ' Use all selected points in model
  
  Code = enumSelection(Lev, Get_Pts Or Get_AllSel, 0, PtList())
  If Code < 1 Then Exit Function

Case PTC_CurPy, PTC_TexPy
  ' Use current polygon only
  
  If Index < 1 Then Code = -1: Exit Function
  
  With Lev.Polys(Index)
    Code = UBound(.PtData)
    
    If Flags And PTC_TexPy Then
      ' Sum polygon's texture data for averaging later
      '
      Dim Pt As Integer
      For Pt = 1 To Code
      With .PtData(Pt)
        
        TempP.X = TempP.X + .Loc.X
        TempP.Y = TempP.Y + .Loc.Y
        TempP.Z = 0
      
      End With
      Next Pt
    
    Else
      ' Enumerate polygon's vertex data for averaging later
      '
      ReDim PtList(Code)
      For Pt = 1 To Code
        PtList(Pt) = .PtData(Pt).Index
      Next Pt
    
    End If
  End With

End Select

If Code < 1 Then Exit Function

If (Flags And PTC_TexPy) = 0 Then
  ' Calculate the average of PtList().
  ' (skip if PTC_TexPy flag is set)
  '
  For Pt = 1 To Code
  With Lev.Points(PtList(Pt))
    
    TempP.X = TempP.X + .Loc.X
    TempP.Y = TempP.Y + .Loc.Y
    TempP.Z = TempP.Z + .Loc.Z
    
  End With
  Next Pt
End If

' Assign the results.
If Flags And PTC_IntMath Then
  pt_CenterOf.X = TempP.X \ Code
  pt_CenterOf.Y = TempP.Y \ Code
  pt_CenterOf.Z = TempP.Z \ Code
Else
  pt_CenterOf.X = TempP.X / Code
  pt_CenterOf.Y = TempP.Y / Code
  pt_CenterOf.Z = TempP.Z / Code
End If

End Function

Public Function v_MidPt(Point1 As D3DVECTOR, Point2 As D3DVECTOR, Optional ByVal Step As Single = 0.5) As D3DVECTOR
'
' Calculates a point somewhere along a line defined by two point vectors.
'
Syntax:
'
' MidPt(Point1, Point2, [Step])
'   Point1 - Start point.
'   Point2 - Finish point.
'   Step   - Percentage distance from start point to compute.
'              * Interpolation: (0 < Step < 1)
'                               0.5 (halfway) is default
'              * Extrapolation: Specify Step<0 or Step>1
'

With Point1
  v_MidPt.X = .X + (Point2.X - .X) * Step
  v_MidPt.Y = .Y + (Point2.Y - .Y) * Step
  v_MidPt.Z = .Z + (Point2.Z - .Z) * Step
End With

End Function

Public Function py_NearestL(Mouse As D3DVECTOR, Polygon As PolygonData) As Integer
Dim DrawP As D3DVECTOR
Dim Pt As Integer, NumPts As Integer

' Searches through the supplied polygon and
' returns whichever of its edges is closest to the
' cursor.
'
Syntax:
'
' py_NearestL(Mouse, Polygon)
'   Mouse   -- Mouse coordinates
'   Polygon -- The polygon data to search.
'
NumPts = UBound(Polygon.PtData)

' Get the p3d flags info
Dim CamC As D3DVECTOR, PFlags As Long
PFlags = GetP3DFlags(Model.Files(ThisFile), Model.Files(ThisFile), CamC)

Dim TestP1 As D3DVECTOR, TestP2 As D3DVECTOR
Dim Dist As Single, ThisDist As Single

' Initialize the loop vars.
py_NearestL = 0
Dist = &HFFFFFFF

For Pt = 1 To NumPts
  
  With Model.Files(ThisFile).Levs
    TestP1 = .Points(Polygon.PtData(Pt).Index).Loc
    TestP2 = .Points(Polygon.PtData(Pt Mod NumPts + 1).Index).Loc
  End With
    
  DrawP = Project3D(PFlags, v_MidPt(TestP1, TestP2), CamC, 0, 0)
  ThisDist = Distance(Mouse, DrawP, , 2, 2)

  If ThisDist <= Dist Then
    py_NearestL = Pt
    Dist = ThisDist
  End If
Next Pt

End Function

Public Function txt_Verify(ByRef Source As TextBox, ByRef Code As Integer, Optional ByVal Forbid As String = "", Optional Blank As Boolean = False) As Integer
'
' Used in various textbox "KeyPress" events
' to disable non-numeric keys.
'
' In theory, saves having to perform an IsNumeric() check on it later...
'
Syntax:
'
' txt_Verify(Source, Code, [Forbid], [Blank])
'   Source - Textbox to evaluate.
'   Code   - Key code pressed.
'   Forbid - Optional list of prohibited keycodes (sign, decimal point, etc.)
'   Blank  - Whether to allow a blank textbox when user hits Backspace key
'
' Return value is void.

Select Case Code
Case 8 ' Backspace - usually enabled
  If Len(Source.Text) <= 1 And Not Blank Then Code = 0

Case 48 To 57
  ' Numeric keys cannot precede a negative sign-
  ' UNLESS we've highlighted the sign & are
  ' going to replace it
  If Source.SelStart + Source.SelLength < InStr(Source.Text, "-") Then
    Code = 0
  End If

Case 46 'Asc(".")
  ' Only allow one period
  If InStr(Source.Text, ".") Then Code = 0

  ' We can prohibit the period, if desired
  If InStr(Forbid, ".") Then Code = 0

Case 45 'Asc("-")
  ' Negative signs are valid only at the start
  If InStr(Source.Text, "-") Or Source.SelStart > 0 Then Code = 0

  ' We can prohibit this too...
  If InStr(Forbid, "-") Then Code = 0

Case Else ' Anything else (non-numeric)

  Code = 0

End Select

End Function

Public Sub Center(ByRef CenterF As Form, Optional ByVal OffsetY As Integer = 40, Optional ByVal OffsetX As Integer = 50)
' Centers any form on the screen.

' Syntax:
' Center (formname), [horz-off], [vert-off]
'
' formname  - Name of form to center.  "Me" works just fine.
' horz-off  - Horizontal location of form (optional):
'               Zero=left, 100=right.
' vert-off  - Vertical location of form (optional):
'               Zero=top, 100=bottom.

CenterF.Left = (Screen.Width - CenterF.Width) * OffsetX / 100
CenterF.Top = (Screen.Height - CenterF.Height) * OffsetY / 100

End Sub

Private Function file_LoadPIE_TrimAll(ByRef StrData As String, Optional TrimEdges As Boolean = False) As Integer
'
' Handles whitespace inside a string when loading PIE files.
'

' Convert tabs to spaces
Dim N As Integer
Do
  N = InStr(StrData, vbTab)
  If N = 0 Then Exit Do
  
  StrData = Left$(StrData, N - 1) & " " & Mid$(StrData, N + 1)
Loop

' Consolidate redundant spaces.
Do
  N = InStr(StrData, "  ")
  If N = 0 Then Exit Do
  
  StrData = Left$(StrData, N - 1) & " " & Mid$(StrData, N + 2)
Loop

' If desired, remove leading/trailing spaces too
'
If TrimEdges Then StrData = Trim$(StrData)

End Function

Public Function PickAreaSel(ByVal What As Integer, Pt1 As D3DVECTOR, Pt2 As D3DVECTOR) As Integer
'
' Selects all points or polygons contained within an area defined by two vectors.
'
Syntax:
'
' PickAreaSel(What, Pt1, Pt2)
'   What    - +1= points
'             +2= polygons
'   Pt1,Pt2 - Corners of a rectangular selection area.
'
' Returns the # of items selected.
'


'
' Compile a selection rectangle from the two coords.
'
Dim R As RECT
If Pt1.Y < Pt2.Y Then
  R.Top = Pt1.Y: R.Bottom = Pt2.Y
Else
  R.Top = Pt2.Y: R.Bottom = Pt1.Y
End If

If Pt1.X < Pt2.X Then
  R.Left = Pt1.X: R.Right = Pt2.X
Else
  R.Left = Pt2.X: R.Right = Pt1.X
End If

Dim DrawP As D3DVECTOR
Dim A As Integer, Num As Integer

' Cache the pie file & camera offset coords
Dim PFlags As Long, CamPt As D3DVECTOR
PFlags = GetP3DFlags(Model.Files(ThisFile), Model.Files(ThisFile), CamPt)

With Model.Files(ThisFile)

If What And 1 Then
  
  ' Select all visible points in the specified area.
  '
  
  For A = 1 To UBound(.Levs.Points)
  With .Levs.Points(A)
  If (.StatFlags And stat_Hide) = 0 Then
    
    DrawP = Project3D(PFlags, .Loc, CamPt, 0, 0)
    
    ' Is our point within the selection area?
    If DrawP.X >= R.Left And DrawP.X <= R.Right And _
       DrawP.Y >= R.Top And DrawP.Y <= R.Bottom Then
        
        .StatFlags = (.StatFlags And stat_NotSelMask) Or stat_Sel
        
        SetItemStat formLists.lstView(0).ListItems(A), stat_Sel
        Num = Num + 1
    End If
  
  End If
  End With
  Next A
End If

If What And 2 Then

  ' Select all polygons centered about the specified area.

  ' Loop through all visible polygons
  For A = 1 To UBound(.Levs.Polys)
  If (.Levs.Polys(A).StatFlags And stat_Hide) = 0 Then
      
    DrawP = Project3D(P3D_Typ, pt_CenterOf(PTC_CurPy, .Levs, A, 0), CamPt, 0, 0)
  
    ' Is our point within the selection area?
    If DrawP.X >= R.Left And DrawP.X <= R.Right And _
       DrawP.Y >= R.Top And DrawP.Y <= R.Bottom Then
      .Levs.Polys(A).StatFlags = .Levs.Polys(A).StatFlags Or stat_Sel
      Num = Num + 1
    End If
  
  End If
  Next A
  
End If

End With

PickAreaSel = Num

End Function


Public Function file_LoadPIE(ByVal FileName As String, ByRef Data As PieType) As Long
'
' Parses a PIE file and returns the number of submodels found.
'
Syntax:
'
' file_LoadPIE(FileName, Data)
'   FileName -- Full path & name of the file to load.
'   Data     -- Place to store the data if the function is successful.
'
' Additionally, Sys.UseLineFilter variable affects
' exactly how the PIE file data is loaded:
'   0 -- word by word
'   1 -- line by line
'
' Returns < 0 if an error occurs -- examine
' the result to determine the reason why.

Dim StartP As Integer, EndP As Integer
Dim TypeL As String, TexPage As String
Dim NumLevs As String
Dim PtLine As String, PyLine As String
Dim ConLine As String, DataLine As String

' Loads a PIE file and returns the # of levels contained whithin.

If FileName = "" Or Dir(FileName) = "" Then
  file_LoadPIE = FE_LoadPie Or FE_FileNotFound
  Exit Function
End If

' Prepare the file-title in case errors occur.
Dim FileTitle As String
Dim N As Integer
N = InStrRev(FileName, "\")
FileTitle = Mid$(FileName, N + 1)

' Suppress error messages
On Error GoTo BadFile

Dim FileNum As Integer
FileNum = FreeFile
Open FileName For Input As #FileNum
  
  ' This here for debug purposes
  Seek #FileNum, 1
  
  ' First two lines aren't needed...yet.
  
  Dim HeaderL As String
  Dim LineN As Long
  
  LineN = 1: Line Input #FileNum, HeaderL
  If Left$(HeaderL, 5) <> "PIE 2" Then
    Debug.Assert False
    MsgBox "Unable to load " & FileTitle & ": it is not a valid PIE file.", vbCritical
    
    file_LoadPIE = FE_LoadPie Or FE_BadFileHeader
    Exit Function
  End If
  
  LineN = 2: Line Input #FileNum, TypeL$
  file_LoadPIE_TrimAll TypeL$
  
  If Left$(TypeL, 4) <> "TYPE" Then
    Debug.Assert False
    MsgBox "Unable to load " & FileTitle & ": the PIE format is invalid.", vbCritical
    
    file_LoadPIE = FE_LoadPie Or FE_BadFileHeader
    Exit Function
  End If
  
  ' PCX file for textures
  LineN = 3: Line Input #FileNum, TexPage
  file_LoadPIE_TrimAll TexPage
  
  Dim S1 As Integer, Sx1 As Integer, Sx2 As Integer
  If InStr(UCase$(TexPage), "NOTEXTURE") Then
    Data.TexPage = ""
  Else
    S1 = InStr(11, TexPage, ".pcx ")
    If S1 = 0 Then
      MsgBox FileTitle & " does not contain a PCX texture.", vbInformation
      Data.TexPage = ""
    Else
      ' Get texture sizes.
      Sx1 = InStr(S1 + 4, TexPage, " ")
      Sx2 = InStr(Sx1 + 1, TexPage & " ", " ")
      'SX3 = InStr(SX2 + 1, TexPage & " ", " ")
      'SY2 = Len(TexPage)
      Data.TexPage = Mid(TexPage, 11, S1 - 7)
    End If
  End If
  
  Dim StartPtr As Long
'  StartPtr = Seek(FileNum)
  
  ' Number of submodels in file
  LineN = 4: Line Input #FileNum, NumLevs
  file_LoadPIE_TrimAll NumLevs
  
  With Data.Levs '(LevelID)
    ReDim .Points(0)
    ReDim .Polys(0)
    ReDim .Ccs(0)
  End With
  
  
  ' Initialize main data array.
  Dim Numlev As Integer
  
  Numlev% = Mid(NumLevs, InStr(NumLevs, " "))
  'ReDim Data.Levs
  
  StartPtr = Seek(FileNum)
  
  Select Case Sys.UseLineFilter
  Case False
    
   
    ' Re-open the file in BINARY mode, and
    ' grab the remaining data in one *BIG* chunk.
    Dim FileData As String
    Close #FileNum
    Open FileName For Binary Access Read Lock Write As #FileNum
    
    ' Calculate how much buffer space is needed
    FileData$ = String$((LOF(FileNum) - StartPtr) + 1, " ")
    Seek #FileNum, StartPtr
    Get #FileNum, , FileData$
    
    ' Process whitespace for easier parsing.
    Dim NotDone As Boolean, N1 As Integer
    Do
      NotDone = False
      If InStr(FileData$, vbTab) > 0 Then
        NotDone = True
        N1 = InStr(FileData$, vbTab)
        FileData$ = Left$(FileData$, N1 - 1) & " " & Mid$(FileData$, N1 + 1)
      End If
      If InStr(FileData$, vbLf) > 0 Then
        NotDone = True
        N1 = InStr(FileData$, vbLf)
        FileData$ = Left$(FileData$, N1 - 1) & " " & Mid$(FileData$, N1 + 1)
      End If
      If InStr(FileData$, vbCr) > 0 Then
        NotDone = True
        N1 = InStr(FileData$, vbCr)
        FileData$ = Left$(FileData$, N1 - 1) & " " & Mid$(FileData$, N1 + 1)
      End If
    Loop While NotDone
    
    ' Initialize our variables.
    EndP = 0
    
    Dim numLoadedPts As Integer, numLoadedPys As Integer, numLoadedCcs As Integer
    numLoadedPts = 0
    numLoadedPys = 0
    numLoadedCcs = 0
    
    Dim NumPts As Integer: NumPts = 0
    Dim NumPolys As Integer: NumPolys = 0
    Dim NumVertex As Integer: NumVertex = 0
    Dim NumCcs As Integer: NumCcs = 0
    Dim NumBSP As Integer: NumBSP = 0
    
    Dim LevelID As Integer: LevelID = 0
    Dim BSPID As Integer: BSPID = 0
    Dim PointID As Integer: PointID = 0
    Dim CcID As Integer: CcID = 0
    Dim PolyID As Integer: PolyID = 0
    Dim VertexID As Integer: VertexID = 0
    
    
    ' Constants used when parsing word-by-word.
    Const Lev_Max As Long = 8, Lev_ID As Long = 9
    
    Const Pt_Max As Long = 16
    Const Pt_X As Long = 17, Pt_Y As Long = 18, Pt_Z As Long = 19
    
    Const Py_Max As Long = 32
    Const Py_Type As Long = 33
    Const Py_Shape As Long = 34 ' # of vertexes
    Const Py_Vertex As Long = 35 ' vertex ID
    Const Py_TC_Frames As Long = 36
    Const Py_TC_Timing As Long = 37
    Const Py_TC_Width As Long = 38, Py_TC_Height As Long = 39
    Const Py_TexX As Long = 40, Py_TexY As Long = 41
  
    Const Cc_Max As Long = 64
    Const Cc_X As Long = 65, Cc_Y As Long = 66, Cc_Z As Long = 67
    
    Const BSP_Max As Long = 127
    
    Dim Expect As Long: Expect = 0
    
    
    ' Begin parsing the data!
    Do
      
      Dim Word As String
      GoSub GetNextWord
    
      If EndP > Len(FileData$) And Expect <> CStr(Py_TexY) Then Exit Do
    
      Select Case Word$
      ' Major keywords tell us what data to load next
      Case ""
      Case "LEVELS":     Expect = Lev_Max
      Case "LEVEL":      Expect = Lev_ID
      Case "POINTS":     Expect = Pt_Max
      Case "CONNECTORS": Expect = Cc_Max
      Case "POLYGONS":   Expect = Py_Max
      Case "BSP":        Expect = BSP_Max
    
      ' Minor keywords (numbers) will actually load the data
      Case Else
        Dim NumLevels
        Select Case Expect
        Case Lev_Max
        ' Reference only:  how many submodels?
          NumLevels = Word$
          'ReDim Data.Levs(NumLevels)
          
          Expect = ""
        Case Lev_ID
        ' Assign the current level.
          LevelID = Word$
          
          numLoadedPts = UBound(Data.Levs.Points)
          numLoadedPys = UBound(Data.Levs.Polys)
          numLoadedCcs = UBound(Data.Levs.Ccs)
          
          ' Crop empty space from the submodel before loading more data
          If PointID < numLoadedPts Then ReDim Preserve Data.Levs.Points(PointID)
          If PolyID < numLoadedPys Then ReDim Preserve Data.Levs.Polys(PolyID)
          If CcID < numLoadedCcs Then ReDim Preserve Data.Levs.Ccs(CcID)
          
        Case Pt_Max
        ' Reference only:  how many points do we have?
          NumPts = Word$
          N = UBound(Data.Levs.Points)
          If N = 0 Then
            ReDim Data.Levs.Points(NumPts)
          Else
            ReDim Preserve Data.Levs.Points(N + NumPts)
          End If
          
          Expect = Pt_X
        
        Case Pt_X
        ' Start of line:  x coordinate
          PointID = PointID + 1
          With Data.Levs '(LevelID)
            If PointID > UBound(.Points) Then
              ReDim Preserve .Points(PointID)
            End If
            .Points(PointID).Loc.X = -Word$
          End With
          Expect = Pt_Y
          
        Case Pt_Y
        ' Y coordinate of a point.
          Data.Levs.Points(PointID).Loc.Y = Word$
          Expect = Pt_Z
        
        Case Pt_Z
        ' Z coordinate of a point.
          Data.Levs.Points(PointID).Loc.Z = -Word$
          Expect = Pt_X
    
        Case Py_Max
        ' Reference only:  how many polygons?
          NumPolys = Word$
          N = UBound(Data.Levs.Polys)
          
'          Debug.Assert LevelID > 0
          If N = 0 Then
            ReDim Data.Levs.Polys(NumPolys)
          Else
            ReDim Preserve Data.Levs.Polys(N + NumPolys)
          End If
          
          Expect = Py_Type
          
          If NumPolys >= 1 And UBound(Data.Levs.Points) < 1 Then
            Word$ = "Polygon data block"
            Err.Raise 1, , "Points must be defined before polygons"
          End If
        
        Case Py_Type
        ' Start of line:  polygon type number.
          PolyID = PolyID + 1
          
          With Data.Levs
            If PolyID > UBound(.Polys) Then
              ReDim Preserve .Polys(PolyID)
            End If
    '        ReDim Preserve .Polys(PolyID)
            
            .Polys(PolyID).Flags = CLng("&H" & Trim(Word$)) And &HFF00&
            .Polys(PolyID).Submodel = LevelID
            Expect = Py_Shape
          End With
          
        Case Py_Shape
        ' Number of vertices in the polygon
          NumVertex = Word$
          VertexID = 0
          Debug.Assert NumVertex > 2 And NumVertex <= WZ_MaxPolySize
          ReDim Data.Levs.Polys(PolyID).PtData(NumVertex)
          Expect = Py_Vertex
    
        Case Py_Vertex
        ' Read the vertex data
          VertexID = VertexID + 1
          Data.Levs.Polys(PolyID).PtData(VertexID).Index = CInt(Word$) + 1 + numLoadedPts
          If CSng(Word$) > UBound(Data.Levs.Points) Then
            Err.Raise 1, , "Specified point does not exist in submodel"
          End If
          If VertexID >= NumVertex Then
            VertexID = 0
            If Data.Levs.Polys(PolyID).Flags And &H4000 Then
              Expect = Py_TC_Frames
            ElseIf Data.Levs.Polys(PolyID).Flags And &H8200& Then
              Expect = Py_TexX
            Else
              Expect = Py_Type
            End If
          End If
    
        ' Read the animation data
        Case Py_TC_Frames
          
          ' Assign team-colors data. Note that if zero frames are specified,
          ' team-colors flag will be removed.
          With Data.Levs.Polys(PolyID)
            .TeamC.Animated = Word$
            If .TeamC.Animated = 0 Then .Flags = .Flags And &HFBF00
          End With
          
          Expect = Py_TC_Timing
        
        Case Py_TC_Timing
          Data.Levs.Polys(PolyID).TeamC.AnimTime = Word$
          Expect = Py_TC_Width
        
        Case Py_TC_Width
          ' Also, if frame width = 0, team-colors flag will be removed.
          With Data.Levs.Polys(PolyID)
            .TeamC.AnimW = Word$
            If .TeamC.AnimW = 0 Then .Flags = .Flags And &HFBF00
          End With
          
          Expect = Py_TC_Height
        
        Case Py_TC_Height
          Data.Levs.Polys(PolyID).TeamC.AnimH = Word$
          If Data.Levs.Polys(PolyID).Flags And &H8200& Then
            Expect = Py_TexX
          Else
            Expect = Py_Type
          End If
        
        ' Read the vertex data
        Case Py_TexX
          VertexID = VertexID + 1
          Data.Levs.Polys(PolyID).PtData(VertexID).Loc.X = Word$
          Expect = Py_TexY
        
        Case Py_TexY
          Data.Levs.Polys(PolyID).PtData(VertexID).Loc.Y = Word$
          If VertexID < NumVertex Then
            Expect = Py_TexX
          Else
            Expect = Py_Type
          End If
    
        Case Cc_Max
          NumCcs = Word$
          ReDim Data.Levs.Ccs(NumCcs)
          Expect = Cc_X
        
        Case Cc_X
        ' Start of line:  x coordinate
          CcID = CcID + 1
          If CcID > NumCcs Then ReDim Preserve Data.Levs.Ccs(CcID)
          Data.Levs.Ccs(CcID).Loc.X = -Word$
          Expect = Cc_Y
        
        Case Cc_Y
        ' Z coordinate is second, not third
          Data.Levs.Ccs(CcID).Loc.Z = -Word$
          Expect = Cc_Z
        
        Case Cc_Z
        ' Y coordinate
          Data.Levs.Ccs(CcID).Loc.Y = Word$

          Dim BadCc As Boolean
          With Data.Levs.Ccs(CcID).Loc
            If (Abs(.X) > &H8000&) Or (Abs(.Y) > &H8000&) Or (Abs(.Z) > &H8000&) Then
              BadCc = True
            End If
          End With
          
          If BadCc Then
            BadCc = False
            MsgBox "Connector " & CcID & " is out of bounds.  Discarding..."
            ReDim Preserve Data.Levs.Ccs(CcID - 1)
          End If
          
          Expect = Cc_X
    
        Case BSP_Max ' Do nothing with the BSP data
        End Select
      End Select
    
      If EndP > Len(FileData$) Then Exit Do
    Loop
    
  Case True
    ' Load remainin PIE data using Line Input.
    
    ' Initialize main Data array.
    Dim Done As Boolean
    
    Dim PtIndx As Single
    Dim PolyType As Long, S2 As Integer, NumPs As Integer
    Dim Pt As Integer, TexDataX As Integer, TexDataY As Integer
    Dim A As Integer
    Dim BadPys As Integer, Mesg2 As String, Mesg3 As String
    
    Numlev% = Mid(NumLevs, InStr(NumLevs, " "))
    
    Data.Levs = getBlankLevel(BlankLev_New0)
    
    Done = False
    'formMain.lstSlices.AddItem "(All Levels)"
    Do
      Line Input #FileNum, DataLine
      file_LoadPIE_TrimAll DataLine
    
      LineN = LineN + 1
      If Left$(DataLine, 1) = vbTab Then DataLine = Mid(DataLine, 2)
      Select Case Left$(DataLine & " ", InStr(DataLine, " ") - 1)
      Case "LEVEL"
        'LevelID = UBound(Data.Levs) + 1
        'ReDim Preserve Data.Levs
        ReDim Data.Levs.Ccs(0)
        
        ' Prepare the first submodel.
        Data.Levs = getBlankLevel(False)
        
        PointID = -1: PolyID = -1: VertexID = -1
        BSPID = 0
    
        LevelID = Mid(DataLine, InStr(DataLine, " "))
        
      Case "POINTS" ' Point Data goes here
        Dim Vx As Integer, vy As Integer, P As D3DVECTOR
        
        ' Input number of points - prepare buffer for it
        '
        NumPts = Mid(DataLine, InStr(DataLine, " "))
        ReDim Preserve Data.Levs.Points(NumPts)
        
        ' Parse point Data.
        For PointID = 1 To NumPts
          Line Input #FileNum, PtLine
          file_LoadPIE_TrimAll PtLine
    
          LineN = LineN + 1
          If Left(PtLine, 1) = vbTab Then PtLine = Mid(PtLine, 2)
          PtLine = Trim(PtLine)
          Vx = InStr(PtLine, " ")
          vy = InStr(Vx + 1, PtLine, " ")
          P.X = Left(PtLine, Vx)
          P.Y = Mid(PtLine, Vx, vy - Vx)
          P.Z = Mid(PtLine, vy)
          
          ' NOTE THE INVERTED X/Z COORDS
          With Data.Levs.Points(PointID)
            .Loc.X = -P.X
            .Loc.Y = P.Y
            .Loc.Z = -P.Z
          End With
          
        Next PointID
      
      Case "POLYGONS" ' Polygon data goes here
        
        ' Input numer of polygons.
        NumPolys = Mid(DataLine, InStr(DataLine, " "))
        ReDim Preserve Data.Levs.Polys(NumPolys)
    
        ' Begin inputting polygon data.
        For PolyID = 1 To NumPolys
        With Data.Levs.Polys(PolyID)
          
          Line Input #FileNum, PyLine
          file_LoadPIE_TrimAll PyLine
    
          LineN = LineN + 1
          If Left$(PyLine, 1) = vbTab Then PyLine = Mid(PyLine, 2)
          PyLine = Trim(PyLine) & " "
          ' Get polygon's type.
          S1 = InStr(PyLine, " ")
    
          PolyType = CLng("&H" & Left(PyLine, S1 - 1)) And &HFF00&
          .Flags = PolyType
    
          ' Define # of polygon points -- NumPs%
          S2 = InStr(S1 + 1, PyLine, " ")
          NumPs% = Mid(PyLine, S1, S2 - S1)
          ReDim Preserve .PtData(NumPs%)
    
          ' Input point Data.
          
          For VertexID = 1 To NumPs%
            S1 = S2: S2 = InStr(S1 + 1, PyLine, " ")
            ' Get and assign point index value.
            PtIndx = Mid(PyLine, S1, S2 - S1)
            
            ' Adjust for the new 1-based arrays
            .PtData(VertexID).Index = PtIndx + 1
          Next VertexID
          
          On Error Resume Next
            ' Animated polygons have 4 extra terms, so load those terms if needed.
            If PolyType And &H4000& Then
              For A = 1 To 4
                S1 = S2: S2 = InStr(S1 + 1, PyLine & " ", " ")
                Select Case A
                Case 1: .TeamC.Animated = Mid(PyLine, S1, S2 - S1)
                Case 2: .TeamC.AnimTime = Mid(PyLine, S1, S2 - S1)
                Case 3: .TeamC.AnimW = Mid(PyLine, S1, S2 - S1)
                Case 4: .TeamC.AnimH = Mid(PyLine, S1, S2 - S1)
                End Select
              Next A
              
              ' Disable team-colors flag for invalid team-color options
              If .TeamC.Animated = 0 Or .TeamC.AnimW = 0 Then .Flags = .Flags And &HFBF00
            End If
            
            ' Textured polys only
            If .Flags And &H200 Then
              For Pt = 1 To NumPs%
                S1 = S2: S2 = InStr(S1 + 1, PyLine & " ", " ")
                TexDataX = Mid(PyLine, S1, S2 - S1)
                S1 = S2: S2 = InStr(S1 + 1, PyLine & " ", " ")
                TexDataY = Mid(PyLine, S1, S2 - S1)
                
                With .PtData(Pt)
                  .Loc.X = TexDataX
                  .Loc.Y = TexDataY
                End With
              Next Pt
            End If
            
            If Err.Number > 0 And Data.TexPage <> "" Then
              BadPys% = BadPys% + 1
              If BadPys% <= 5 Then
                Mesg2$ = Mesg2$ & vbCrLf & "     Polygon #" & PolyID
              Else
                Mesg3$ = vbCrLf & "        (and " & BadPys% - 5 & " others)"
              End If
            End If
          On Error GoTo BadFile
    
        End With
        Next PolyID
        
      Case "CONNECTORS" ' Connectors data.
        NumCcs = Mid(DataLine, InStr(DataLine, " "))
        ReDim Preserve Data.Levs.Ccs(NumCcs)
        CcID = 0
        ' Get connectors data
        ' Must be structured as a Do...Loop and not For...Next
        Do
    
          CcID = CcID + 1
          Line Input #FileNum, ConLine
          file_LoadPIE_TrimAll ConLine
    
          LineN = LineN + 1
          If Left$(ConLine, 1) = vbTab Then ConLine = Mid(ConLine, 2)
          ConLine = Trim(ConLine)
          S2 = 1
          For A% = 1 To 3
    
            S1 = S2: S2 = InStr(S1 + 1, ConLine & " ", " ")
    
            ' Get connector data.
            PtIndx = Mid(ConLine, S1, S2 - S1)
            If Abs(PtIndx) > &H10000 Then
              MsgBox "Connector " & CcID & " of " & NumCcs & " is out of bounds.  Discarding..."
              ReDim Preserve Data.Levs.Ccs(NumCcs - 1)
              CcID = CcID - 1
              NumCcs = NumCcs - 1
              Exit For
            End If
    
            ' Store connector data.  Note the swapped y & z axes!
            ' Invertex x & z axes are normal for Pie Slicer translation.
            Select Case A%
            Case 1: Data.Levs.Ccs(CcID).Loc.X = -PtIndx
            Case 2: Data.Levs.Ccs(CcID).Loc.Z = -PtIndx
            Case 3: Data.Levs.Ccs(CcID).Loc.Y = PtIndx
            End Select
          Next A%
        Loop Until (CcID >= NumCcs) Or (NumCcs = 0)
      
      Case "BSP"
        NumBSP = 0  'Mid(Dataline, InStr(Dataline, " "))
        ReDim Preserve Data.Levs.BSP(NumBSP)
'        ' Increment through the file
'        For BSPID = 1 To Numb%
'          Line Input #FileNum, Data.Levs.BSP(BSPID)
'          LineN = LineN + 1
'        Next BSPID
      
      Case Else
        Done = True
      
      End Select
      If EOF(FileNum) Then Done = True
    Loop Until Done
  
    ' Any texture problems?
    If Mesg2$ <> "" Then
      MsgBox "Could not load the texture data for the following polygons:" & vbCrLf & Mesg2$ & Mesg3$ & vbCrLf & vbCrLf & "They will need to be re-textured.", vbInformation, FileTitle
    End If
  
  End Select
Close #FileNum

Data.Loaded = Timer
Data.Changed = Data.Loaded
Data.FileName = FileName

file_LoadPIE = Numlev
Exit Function


BadFile:
Debug.Assert False

Resume Next
MsgBox "Internal error #" & Err.Number & " occured while loading " & FileTitle & " .  See ""PieSlicer.log"" for details."

' Unknown error; give an informative error message.  PIE files have some flags still unknown....
formMaster.SetStatus 0, "5|Errors detected in """ & FileTitle & """.  See ""PIESlicer.log"" for details."

If Dir(App.Path & "\PIESlicer.log") <> "" Then Kill App.Path & "\PIESlicer.log"

Dim LogNum As Integer
LogNum = FreeFile
Open App.Path & "\PIESlicer.log" For Output As #LogNum
  Print #LogNum, "Errors encountered while translating """ & FileName & """ : "
  Print #LogNum,
  Print #LogNum, "Submodel:  " & LevelID
  
  If Sys.UseLineFilter Then
    Print #LogNum, "Line:  " & LineN
  Else
    Print #LogNum, "Data Code:  " & Expect
    Print #LogNum, "Data Read:  " & Word$
  End If
  
  Print #LogNum, "Point:  " & PointID
  Print #LogNum, "Polygon: " & PolyID
  Print #LogNum, "Vertex: " & VertexID
  Print #LogNum, "Cc:  " & CcID
  
  Print #LogNum, "Error #" & Err.Number & ": " & Err.Description
  
  Print #LogNum, ""
  Print #LogNum, "If you would like technical assistance with this error,"
  Print #LogNum, "contact Strata of RTS.Net (strata_ranger@hotmail.com) with the PIE file and this message."
Close #LogNum

file_LoadPIE = FE_LoadPie Or FE_Internal

Close #FileNum

Exit Function

' This is for debugging purposes
On Error GoTo 0: Resume

'
' Inline subfunctions begin here
'

'
' GetNextWord: Parses FileData$ and returns the
' next token (whitespace delimited) in var Word$.
'
' * Set EndP before calling.
'   May be called several times until EndP=0, indicating end of string.
'
GetNextWord:
  Do
    StartP = EndP
    EndP = InStr(StartP + 1, FileData$ & " ", " ")
    Word$ = Mid(FileData$, StartP + 1, EndP - StartP - 1)
  Loop Until Word$ <> "" Or EndP > Len(FileData$)
Return

End Function

Public Function Distance(Point1 As D3DVECTOR, Point2 As D3DVECTOR, Optional ByVal Intger As Boolean = True, Optional Dims As Integer = 3, Optional Pwr As Integer = 1) As Single
'
' Returns the linear distance between two points in the same space.
'
' * This function is time-critical
'
Syntax:
'
' Distance(Point1, Point2, [Intger], [Dims], [Pwr])
'   Point1 - First point.
'   Point2 - Second point.
'   Intger - Round to integers (WZ precision) when TRUE.
'   Dims   - Whether to return a 2D or 3D result.
'              Dims=2 - Calculate 2D style (ignoring .z values)
'   Pwr    - 2=return distance^2, 1=return distance

Dim D As D3DVECTOR
If Intger Then
  D.X = Int(Point2.X + 0.5) - Int(Point1.X + 0.5)
  D.Y = Int(Point2.Y + 0.5) - Int(Point1.Y + 0.5)
  D.Z = Int(Point2.Z + 0.5) - Int(Point1.Z + 0.5)
Else
  D.X = Point2.X - Point1.X
  D.Y = Point2.Y - Point1.Y
  D.Z = Point2.Z - Point1.Z
End If

If Dims = 2 Then
  Distance = (D.X * D.X) + (D.Y * D.Y)
Else
  Distance = (D.X * D.X) + (D.Y * D.Y) + (D.Z * D.Z)
End If

If Pwr <> 2 Then Distance = Sqr(Distance)

End Function

Public Function pt_AddNew(Lev As LevelType) As Integer
'
' Adds a new point to the PIE model.
'
' * Use inline code instead during time-critical functions
'
Syntax:
'
' pt_AddNew(Lev)
'   Lev - Submodel to add the point to.
'
' Function returns the index to the new point.
'
pt_AddNew = UBound(Lev.Points) + 1
ReDim Preserve Lev.Points(pt_AddNew)

With Lev.Points(pt_AddNew)
  .Loc.X = 0
  .Loc.Y = 0
  .Loc.Z = 0
End With

End Function

Public Function ProjectInv3D(ByRef Point As D3DVECTOR, Optional ByVal Angl As Integer = -400, Optional ByVal Angl2 As Integer = -400, Optional ByVal Intger As Boolean = True) As D3DVECTOR
Dim M As D3DVECTOR, V As D3DVECTOR, P As D3DVECTOR
' Inverse of Project3D function, converting
' display coordinates into world coordinates.
'
'
' Uses basically the same syntax as Project3D().


' These are the default settings for those optional arguments.
If Angl = -400 Then getViewAngles Angl, 0, 0
If Angl2 = -400 Then getViewAngles 0, Angl2, 0

' Calculate sine/cosine scalars
Dim Rad As Single
Dim CV As Single, CH As Single
Dim SV As Single, SH As Single

Rad = Angl * m_Deg2Rad
CH = Cos(Rad): SH = Sin(Rad)

Rad = Angl2 * m_Deg2Rad
CV = Cos(Rad): SV = Sin(Rad)

' Rotation from Display -> World

'' Compute visual X coordinate.
'vx = Px * Cos(Rad) - Pz * Sin(Rad)
'
'' Compute visual Y coordinate.
'vy = (Pz * Cos(Rad) + Px * Sin(Rad)) * -Sin(Rad2) - Py * Cos(Rad2)
'
'' Compute visual Z (depth) coordinate.
'' This is used for color fading and perspective.
'vz = (Pz * Cos(Rad) + Px * Sin(Rad)) * Cos(Rad2) - Py * Sin(Rad2)

V = Point

' Undo the vertical rotation.
M.X = V.X
M.Y = -V.Y * CV - V.Z * SV
M.Z = V.Z * CV - V.Y * SV

' Undo the horizontal rotation.
P.X = M.X * CH + M.Z * SH
P.Y = M.Y
P.Z = M.Z * CH - M.X * SH


If Intger Then P.X = Int(P.X + 0.5)
If Intger Then P.Y = Int(P.Y + 0.5)
If Intger Then P.Z = Int(P.Z + 0.5)

ProjectInv3D.X = P.X
ProjectInv3D.Y = P.Y
ProjectInv3D.Z = P.Z

End Function

Public Function GetArcTan(Pt As D3DVECTOR, Flags As Long, CtrP As D3DVECTOR, Optional ByVal ScalY As Single = 1) As Single
'
' Performs an arc-tangent calculation of a point.
'
Syntax:
'
' GetArcTan(Pt, Flags, CtrP, [ScalY])
'   Pt    = Point to calculate
'   Flags = Used to modify behavior:
'             ATN_ScreenY  - Invert Y coordinate
'             ATN_UseCtrPt - Allows use of CtrP argument
'             ATN_AllowNeg - If set, return value will be from -ScalY/2 to ScalY/2
'                            Otherwise, return value will be from 0 to ScalY
'   CtrP  = Location that Pt is relative to.
'             * Used only when ATN_UseCtrPt is set
'   ScalY = Angle system to portray the results in.
'             * Degrees = 360
'             * Radians = m_2pi
'             * Default value is in % of full circle
'
' Returns an angle calibrated to the desired measurement system.
'
Dim PtX As Single, PtY As Single

If Flags And ATN_UseCtrPt Then
  PtX = Pt.X - CtrP.X
  PtY = Pt.Y - CtrP.Y
Else
  PtX = Pt.X
  PtY = Pt.Y
End If

If Flags And ATN_ScreenY Then Pt.Y = -Pt.Y

On Error Resume Next
Dim Angle As Single
Angle = Atn(PtY / PtX) / m_2pi ' Angle will be in range (-0.25, 0.25)

If Err.Number = 6 Then ' Overflow condition i.e. 0/0
  GetArcTan = 0
  Exit Function

ElseIf Err.Number = 11 Then ' Divide-by-zero condition i.e. PtX=0
  GetArcTan = IIf(PtY > 0, 0.25, 0.75) * ScalY
  Exit Function
  
End If
  
If PtX < 0 Then Angle = Angle + 0.5
If Angle < 0 And (Flags And ATN_AllowNeg) = 0 Then Angle = Angle + 1

GetArcTan = Angle * ScalY

End Function

Public Function LoadSetupFile()
' Load system configuration data.
On Error GoTo 0
'On Error GoTo NoFile

User = DefaultUser

If Dir(App.Path & "\config.dtb") = "" Then
NoFile:
  If Err.Number > 0 Then MsgBox "Unable to load program configuration.  Resetting to default...", vbCritical
  Exit Function
End If

Open App.Path & "\config.dtb" For Input As #1
Dim TempV As Integer

Do Until EOF(1)
  Dim DataType As String, RevNum As Single
  Line Input #1, DataType
  Select Case LCase$(DataType)
    
  Case "[options]"
    ' Revised Pie Slicer options/config format.
    '
    Dim VerNum As Integer
    Input #1, VerNum
    
    If VerNum <= VerCFG_Options Then ' Only load from supported versions
      Input #1, User.AutoSave
      Input #1, User.TexFormat
      Input #1, User.AlwaysOnTop
      Input #1, User.UseColorKey
      Input #1, User.AlphaRef
      
      Dim N As Long
      Input #1, N
      If N Then User.Settings = User.Settings Or Set_2xRes
      
      'User.Settings = User.Settings Or Set_FullPCX
      
      Dim Cmd As String
      Do
        Input #1, Cmd
        Select Case Cmd
        Case "Grid"
          With User.GridSpacing
            Input #1, .X, .Y, .Z
          End With
          
        Case "Insert"
          With User.NewOffset
            Input #1, .X, .Y, .Z
          End With
        
        Case "Icons"
          Input #1, User.IconSet
        
        End Select
      Loop Until Cmd = ""
    End If
    
  Case "[general]"
    ' Older Pie Slicer options/config format
    '
    If VerNum = 0 Then
      Input #1, RevNum
      Input #1, TempV ' Set_KeyFaceUp flag - no longer used
      'User.Settings = User.Settings And Not Set_KeyFaceUp Or IIf(TempV, Set_KeyFaceUp, 0)
      
      If RevNum > 1 Then Input #1, User.AutoSave
      If RevNum > 1.1 Then Input #1, User.TexFormat
      If RevNum > 1.2 Then Input #1, User.AlwaysOnTop
      
      If RevNum > 1.3 Then
        Input #1, User.UseColorKey
        Input #1, User.AlphaRef
      End If
    End If
    
  Case "[grid]"
    ' Input grid configuration.
    Input #1, RevNum
    
    With User.GridSpacing
      Input #1, N: .X = N
      Input #1, N: .Y = N
      Input #1, N: .Z = N
    End With
    
    If RevNum > 1 And RevNum < 1.3 Then
      Input #1, TempV
      If TempV Then
        User.Settings = User.Settings Or Set_MoveBinary
      Else
        User.Settings = User.Settings And Not Set_MoveBinary
      End If
    End If
    
    If RevNum = 1.1 Then
      With User.NewOffset
        Input #1, .X, .Y, .Z
      End With
    End If
  
  Case "[points]" ' Moved to the [editing] block
    Input #1, RevNum
    With User.NewOffset
      Input #1, N: .X = 10 * N
      Input #1, N: .Y = 10 * N
      Input #1, N: .Z = 10 * N
    End With
    
  Case "[editing]"
    Input #1, RevNum
    With User.NewOffset
      Input #1, .X, .Y, .Z
    End With
    Input #1, TempV
    If TempV Then
      User.Settings = User.Settings Or Set_MoveBinary
    Else
      User.Settings = User.Settings And Not Set_MoveBinary
    End If
    
    Input #1, TempV
    'User.Settings = User.Settings And Not Set_ShiftIns Or IIf(TempV, Set_ShiftIns, 0)
  
  Case "[files]"
    Dim F As Integer
    Input #1, User.MaxFiles
    
    N = Seek(1)
    F% = 0
    
    ReDim Sys.RecentFiles(User.MaxFiles)
    Do Until EOF(1)
      F% = F% + 1
      Dim FileN As String
      Line Input #1, FileN$
      If Left$(FileN$, 1) = "[" And Right$(FileN$, 1) = "]" Then Exit Do
      If F% <= User.MaxFiles Then
        If F > UBound(Sys.RecentFiles) Then
          ReDim Preserve Sys.RecentFiles(F)
        End If
        Sys.RecentFiles(F%) = FileN$
      End If
      N& = Seek(1)
    Loop
    Seek #1, N&
  
  Case "[windowpos]"
    Input #1, RevNum
    
    Dim DataLine As String, Terms() As String
    Do
      Line Input #1, DataLine
      If DataLine = "" Then Exit Do
      N = InStr(DataLine, "=")
      If N > 0 Then
        Terms() = Split(Mid$(DataLine, N + 1), ",")
      
        Dim A As Integer
        Select Case LCase$(Left$(DataLine, N - 1))
          Case "master": A = win_Master
          Case "main": A = win_Model
          Case "list": A = win_Lists
          Case "tool": A = win_Tools
          Case "tex": A = win_Texture
          Case "desktop": A = 0
          Case Else: Debug.Assert False
        End Select
        
        With Sys.WindowPos(A)
          If A Then
            .X1 = Terms(0)
            .Y1 = Terms(1)
            .X2 = Terms(2)
            .Y2 = Terms(3)
          Else ' "Desktop" item has no x/y values
            .X1 = 0
            .Y1 = 0
            .X2 = Terms(0)
            .Y2 = Terms(1)
          End If
        End With
      End If
   Loop Until N = 0
  
  Case Else
    Debug.Print "LoadSetupFile() -- unknown section " & LCase$(DataType)
  
  End Select
Loop

Close #1
End Function

Public Function SaveSetupFile()
'
' Saves Pie Slicer user configuration and recent file listing to disk.
'
Dim OutFile As String
OutFile = App.Path & "\config.dtb"

'If Dir(OutFile) <> "" Then
'  Kill OutFile
'End If

Open OutFile For Output As #1

Print #1, "[Options]"
  Print #1, "100,";
  Print #1, CStr(User.AutoSave) & ",";
  Print #1, CStr(User.TexFormat) & ",";
  Print #1, IIf(User.AlwaysOnTop, "1,", "0,");
  Print #1, CStr(User.UseColorKey) & ",";
  Print #1, CStr(User.AlphaRef) & ",";
  
  If User.Settings And Set_2xRes Then
    Print #1, "1,";
  Else
    Print #1, "0,";
  End If
  
  Print #1, "Grid,";
  With User.GridSpacing
    Print #1, CStr(.X) & ","; CStr(.Y) & ","; CStr(.Z)
  End With
  
  Print #1, "Insert,";
  With User.NewOffset
    Print #1, CStr(.X) & ","; CStr(.Y) & ","; CStr(.Z)
  End With
  
  Print #1, "Icons,"; User.IconSet
  
  Print #1, ""
  
' For compatibility with older Pie Slicer versions
'Print #1, "[General]"
'  Print #1, "1.5"
'  Print #1, "1" ' ; IIf(User.Settings And Set_KeyFaceUp, 1, 0) & ",";
'  Print #1, User.AutoSave
'  Print #1, User.TexFormat
'  Print #1, IIf(User.AlwaysOnTop, 1, 0)
'  Print #1, User.UseColorKey
'  Print #1, User.AlphaRef
'  Print #1, IIf(User.Settings And Set_NoPCX, 1, 0)
'
'Print #1, "[Grid]"
'  Print #1, "1.3"
'
'  With User.GridSpacing
'    Write #1, .X; .Y; .Z
'  End With
'
'Print #1, "[Editing]"
'  Print #1, "1.1"
'
'  With User.NewOffset
'    Write #1, .X; .Y; .Z; IIf(User.Settings And Set_MoveBinary, 1, 0)
'  End With
'
'  Write #1, "1" ' Shift+Ins = Draw polygons with mouse
  
  
Print #1, "[WindowPos]"
  Print #1, "1.0"
  
  Dim A As Integer
  For A = 0 To win_Last
  With Sys.WindowPos(A)
    Select Case A
      Case 0
        Print #1, "Desktop="; Screen.Width & ","; Screen.Height
      Case win_Master: Print #1, "Master=";
      Case win_Model: Print #1, "Main=";
      Case win_Lists: Print #1, "List=";
      Case win_Tools: Print #1, "Tool=";
      Case win_Texture: Print #1, "Tex=";
    End Select
    If A Then Print #1, .X1 & "," & .Y1 & "," & .X2 & "," & .Y2
  End With
  Next A
  Print #1, ""
  
Print #1, "[Files]"
  Write #1, User.MaxFiles
  
  For A% = 1 To UBound(Sys.RecentFiles)
    Print #1, Sys.RecentFiles(A)
  Next A%
Close #1

Exit Function


End Function


Public Function FormBorder(ByRef What As Form) As D3DVECTOR
' Measures the border width of a window.
' Only works if the form is scaled in twips or pixels.

Dim Result As D3DVECTOR
Select Case What.ScaleMode
Case 3 ' Pixels
  Result.X = What.Width - What.ScaleWidth * Screen.TwipsPerPixelX
  Result.Y = What.Height - What.ScaleHeight * Screen.TwipsPerPixelY
Case 1 ' Twips
  Result.X = What.Width - What.ScaleWidth
  Result.Y = What.Height - What.ScaleHeight
Case Else
  Debug.Assert False
End Select

FormBorder = Result

End Function

Public Function getPieSizeBox(ByRef Lev As LevelType, Mins As D3DVECTOR, Maxs As D3DVECTOR) As Integer
'
' Evaluates the minimums/maximums of a PIE file.
'
' In other words, returns the dimensions of a 3D rectangle that the PIE file fits in.
'
Syntax:
'
' getPieSizeBox(Lev, Mins, Maxs)
'   Lev  - PIE data to evaluate.
'   Mins - (output) Low corner of the rectangular area
'   Maxs - (output) High corner of the rectangular area
'
' Function returns -1 if no points are available to check.

If UBound(Lev.Points) < 1 Then getPieSizeBox = -1: Exit Function

Maxs = Lev.Points(1).Loc
Mins = Lev.Points(1).Loc

Dim A As Integer
For A = 2 To UBound(Lev.Points)
With Lev.Points(A).Loc
  
  If .X > Maxs.X Then Maxs.X = .X
  If .Y > Maxs.Y Then Maxs.Y = .Y
  If .Z > Maxs.Z Then Maxs.Z = .Z
  If .X < Mins.X Then Mins.X = .X
  If .Y < Mins.Y Then Mins.Y = .Y
  If .Z < Mins.Z Then Mins.Z = .Z
  
End With
Next A

End Function

Public Function getPieSize(ByRef Lev As LevelType, Optional Flags As MiscFlags) As Single
' Returns the maximum size of a submodel.

Dim Limit As Single, A As Integer, Radius As Single

Limit = 1
For A = 1 To UBound(Lev.Points)
With Lev.Points(A).Loc
  
  If Flags And GetSize_AxialComp Then ' Axes measurement
    If Abs(.X) > Abs(.Y) And Abs(.X) > Abs(.Z) Then
      Radius = Abs(.X)
    ElseIf Abs(.Y) > Abs(.Z) Then
      Radius = Abs(.Y)
    Else
      Radius = Abs(.Z)
    End If
  Else ' Spherical measurement
    Radius = Sqr(.X ^ 2 + .Y ^ 2 + .Z ^ 2)
  End If
  
  If Radius > Limit Then Limit = Radius

End With
Next A

If Flags And GetSize_ZoomLevel Then
  ' Return a zoom factor scaled to the model's extremes.
  ' If the model measures 1 map-square, zoom factor will be zero.
  
' Sys.ZoomScalar = 2 * 10 ^ (Sys.ZoomStep / 20)
  
  Const ZScal As Single = -20 / m_Log10
  getPieSize = CInt(Log(Limit / 128) * ZScal)
  
  If getPieSize > 15 Then getPieSize = 15
  If getPieSize < -25 Then getPieSize = -25
Else
  getPieSize = Limit!
End If

End Function

Public Function py_getNormalA(ByRef Level As LevelType, ByVal Index As Integer) As D3DVECTOR
'
' Returns viewing angles that correspond to a polygon's normal vector.
' Mostly used for the polygon auto-texture function.
'
' * Not to be called during time-critical routines
'
Syntax:
'
' py_GetNormalA(Level, Index)
'   Level - PIE submodel to check.
'   Index - Polygon index ##.
'
' Function returns a D3DVECTOR representing the view angles.
'   .x member returns horizontal angle
'   .y member returns vertical angle
'   .z member is not used


' Snag the polygon's normal.
Dim Temp As D3DVECTOR, XY As D3DVECTOR
Temp = py_getNormalV(Level, Index)

' Calculate horizontal angle.
Dim AngleH As Single, AngleV As Single
If Temp.X = 0 And Temp.Z = 0 Then ' Verticals check
  py_getNormalA.X = 90
  py_getNormalA.Y = 90
  Exit Function
Else
  XY.Y = Temp.Z
  XY.X = Temp.X
  AngleH = GetArcTan(XY, 0, XY, 360)
End If

' Calculate vertical angle.
If Temp.Y = 0 Then
  AngleV = 0
Else
  XY.Y = Temp.Y
  XY.X = Sqr(Temp.X ^ 2 + Temp.Z ^ 2)
  AngleV = GetArcTan(XY, ATN_AllowNeg, XY, 360)
End If

py_getNormalA.X = AngleH
py_getNormalA.Y = AngleV

End Function

Public Function v_GetAxis(ByVal KeyCode As Integer) As D3DVECTOR
Dim Axes(1 To 3) As D3DVECTOR
Dim DoInvert As Integer
Dim Result As D3DVECTOR
'
' Returns an axis vector corresponding to an onscreen direction.
'
Syntax:
'
' NearAxis(key)
'
' key - The onscreen direction.  Use the constants:
'        axRight, axLeft
'        axUp, axDown
'        axPlus, axMinus
'
' Return value will be an x-, y-, or z- axis vector.


' Create the axes vectors.
Axes(1).X = 1: Axes(2).Y = 1: Axes(3).Z = 1

' Cache the viewing angles (in radians)
Dim Horz As Integer, Vert As Integer
Dim H2 As Single, V2 As Single
getViewAngles Horz, Vert, 0
H2 = Horz * m_Deg2Rad: V2 = Vert * m_Deg2Rad

' Determine which axes we want.
Select Case KeyCode

'
' "Left" and "Right" directions
'
Case axRight, axLeft
  
  ' Orthogonal views
  Select Case ViewIndex
  Case VIEW_SIDE
    Result = Axes(3)
  
  Case Is < VIEW_3D
    Result = Axes(1)
  
  ' 3D views
  Case Else
    ' X axes pointing up/down: return z
    If Abs(Sin(H2)) > Abs(Cos(H2)) Then
      Result = Axes(3)
      If Sin(H2) > 0 Then DoInvert = DoInvert + 1
    ' Z axes pointing up/down: return x
    Else
      Result = Axes(1)
      If Cos(H2) < 0 Then DoInvert = DoInvert + 1
    End If
  End Select

'
' "Up" and "Down" directions
'
Case axUp, axDown
  
  ' Orthogonal views
  Select Case ViewIndex
  Case VIEW_TOP: Result = Axes(3)
  
  Case Is < VIEW_3D
    Result = Axes(2)
  
  ' 3D views
  Case Else
    If Abs(Vert) < 45 Then ' All side views:  Return 'y' axis.
      Result = Axes(2)
    
    Else ' Top/bottom views:  Return 'x' or 'z' axis.
      If Abs(Sin(H2)) > Abs(Cos(H2)) Then
        Result = Axes(1)
        If Sgn(V2) * Sin(H2) < 0 Then DoInvert = DoInvert + 1
      Else
        Result = Axes(3)
        If Sgn(V2) * Cos(H2) < 0 Then DoInvert = DoInvert + 1
      End If
    End If
  End Select

'
' "Plus" and "Minus" directions
'
Case axPlus, axMinus
  
  ' Orthogonal views
  Select Case ViewIndex
  Case VIEW_BACK: Result = Axes(3): DoInvert = 1
  Case VIEW_TOP:  Result = Axes(2)
  Case VIEW_SIDE: Result = Axes(1)
  
  ' 3D views
  Case Else
    If Abs(Vert) < 45 Then
      If Abs(Sin(H2)) > Abs(Cos(H2)) Then
        Result = Axes(1)
        If Sin(H2) > 0 Then DoInvert = DoInvert + 1
      Else
        Result = Axes(3)
        If Cos(H2) > 0 Then DoInvert = DoInvert + 1
      End If
    Else
      Result = v_Invert(Axes(2))
      If Vert > 0 Then DoInvert = DoInvert + 1
    End If
  End Select
  
End Select

If KeyCode And ax_InvertMask Then DoInvert = DoInvert + 1

If DoInvert And 1 Then
  v_GetAxis = v_Invert(Result)
Else
  v_GetAxis = Result
End If

End Function

Public Function v_Invert(ByRef V1 As D3DVECTOR) As D3DVECTOR
'
' Inverts a vector.
'
' So simple, it really should be written via inline code.
'
Syntax:
'
' v_Invert(V1)
'   V1 - Vector to invert.
'

With v_Invert
  .X = -V1.X
  .Y = -V1.Y
  .Z = -V1.Z
End With

End Function

Public Function pt_MoveSel(Vector As D3DVECTOR, ByVal Incr As Integer) As Integer
Dim MoveP As D3DVECTOR
Dim Pt As Integer, CC As Integer
'
' Arbitrarily moves a submodel's selected points + connectors.
'
Syntax:
'
' MovePoints(vector, distance)
'   vector   - Unit vector in X, Y, and/or Z axes directions.
'   distance - Distance to move, in PIE units.
'

Vector.X = Sgn(Vector.X) * Incr
Vector.Y = Sgn(Vector.Y) * Incr
Vector.Z = Sgn(Vector.Z) * Incr

Dim Temp As PieModel
Temp = Model

With Model.Files(ThisFile).Levs '(ThisLev)
  ' Loop through all points.
  For Pt = 1 To UBound(.Points)
    ' If the point was selected...
    With .Points(Pt)
      If (.StatFlags And stat_SelMask) = stat_Sel Then
        pt_MoveSel = pt_MoveSel + 1
        
        ' Move the point.
        MoveP.X = .Loc.X + Vector.X
        MoveP.Y = .Loc.Y + Vector.Y
        MoveP.Z = .Loc.Z + Vector.Z
        
        ' Assign point data in its entirety
        .Loc = MoveP
        
        ' Update the points databox.
        With formLists.lstView(0).ListItems(Pt)
          .SubItems(1) = txt_Round1(MoveP.X)
          .SubItems(2) = txt_Round1(MoveP.Y)
          .SubItems(3) = txt_Round1(MoveP.Z)
        End With
      End If
    End With
  Next Pt
  
   ' Loop through all connectors
  For CC = 1 To UBound(.Ccs)
  With .Ccs(CC)
  
    ' If selected...
    If (.StatFlags And stat_SelMask) = stat_Sel Then
      pt_MoveSel = pt_MoveSel + 1
      
      ' Move the connector.
      MoveP.X = .Loc.X + Vector.X
      MoveP.Y = .Loc.Y + Vector.Y
      MoveP.Z = .Loc.Z + Vector.Z
      
      ' Assign point data in its entirety
      .Loc = MoveP
      
      ' Update the points databox.
      With formLists.lstView(5).ListItems(CC)
        .SubItems(1) = txt_Round1(MoveP.X)
        .SubItems(2) = txt_Round1(MoveP.Y)
        .SubItems(3) = txt_Round1(MoveP.Z)
      End With
    End If
  
  End With
  Next CC

End With

SetUndoStep Temp, Undo_MovePts

End Function

Public Function txt_Round1(ByVal Numbr As Single) As String
'
' Formats a WZ coordinate into a Pie Slicer coordinate.
'
' Used mainly for GUI-related routines.
'
txt_Round1 = Format(Int(Numbr + 0.5) / 10, "0.0")

End Function


Public Function DefaultUser() As UserType
' PIE Slicer's default settings.

With DefaultUser
  .GridSpacing.X = 1.6
  .GridSpacing.Y = 1.6
  .GridSpacing.Z = 1.6
  .Settings = Set_Default

  .NewOffset.X = 0
  .NewOffset.Y = 0
  .NewOffset.Z = 1

  .AutoSave = 5 ' minutes
  .MaxFiles = 5
  
  .AlphaRef = 64
  .UseColorKey = 1
End With

End Function

Public Function py_DrawNew(ByRef Lev As LevelType, DrawData() As Integer, NumPts As Integer) As Integer
'
' Create a polygon from an array of point indexes.
'
Syntax:
'
' py_DrawNew(Lev, DrawData(), [NumPts])
'
'   Lev        - Submodel of interest.
'   DrawData() - Array of point indexes dictating the polygon's shape
'   NumPts     - # of vertexes in DrawData() to use.
'
' Returns the new polygon's index if successful.
'
Dim MxPts As Integer, MxPys As Integer
Dim Poly As Integer, Pt As Integer, NewPt As Integer
MxPts% = UBound(Lev.Points)
MxPys% = UBound(Lev.Polys)

' Debug check - compensate if first & last points are the same
If DrawData(1) = DrawData(NumPts) Then NumPts = NumPts - 1

' Enforce proper syntax.
If NumPts < 3 Or NumPts > WZ_MaxPolySize Then py_DrawNew = -1: Exit Function

' Create the new polygon
'
Poly = UBound(Lev.Polys) + 1
ReDim Preserve Lev.Polys(Poly)
With Lev.Polys(Poly)
  
  ReDim .PtData(NumPts)
  For Pt = 1 To NumPts
    .PtData(Pt).Index = DrawData(Pt)
  Next Pt

End With

' Assign its initial texture alignment.
py_ResetTexmap Lev, Poly

' Specify it as textured
Lev.Polys(Poly).Flags = &H200

' Return its data index
py_DrawNew = Poly

End Function



Public Function GetTexPageNum(ByVal TexPage As String, ByRef TexName As String) As Integer
'
' Returns the page number associated to a texture page filename.
'
Syntax:
'
' GetTexPageNum(TexPage, TexName)
'   TexPage - Texpage filename to evaluate.
'             Typically the same as specified by the PIE file...
'   TexName - (output) Returns the name portion of the texpage
'             that occurs between the page ## and .pcx extension
'
' Function returns the page ## as extracted from TexPage, and sets
' TexName according to the remainder of the TexPage after the page ##.
'
' May return 0 if no page number is available,
' or -1 for invalid calling syntax.
'

Dim N As Integer, Search As String
If TexPage = "" Then
  TexName = ""
  GetTexPageNum = 0
  Exit Function
End If

' Strip any directory entries from the texture page
TexPage = Mid$(TexPage, InStrRev("\" & TexPage, "\"))

' Extract the page number.  page(##) and page-##- formats are both supported.
Select Case LCase$(Left$(TexPage, 5))
Case "page-", "page("
  N% = 6
  Do Until InStr("01234567890", Mid$(TexPage, N%, 1)) = 0
    N% = N% + 1
  Loop
  If N = 6 Then
    GetTexPageNum = -1
    TexName = ""
    Exit Function
  Else
    Search = LCase$(Left$(TexPage, N% - 1))
    GetTexPageNum = Mid(Search, 6)
  End If
  
  ' Then extract the rest...
  Dim N2 As Integer
  N2 = InStrRev(TexPage, ".")
  If N2 = 0 Then N2 = Len(TexPage)
  If N2 = N Then TexName = "" Else TexName = Mid$(TexPage, N, N2 - N)

Case Else
  TexName = ""
  GetTexPageNum = -1
  Exit Function

End Select

End Function


Public Function getBlankLevel(Optional Flags As Long = BlankLev_New1) As LevelType
Dim TempLev As LevelType
' Returns a blank level structure.

If Flags And BlankLev_Points Then
  If Flags And (BlankLev_StartPt And Not BlankLev_Points) Then
    ReDim TempLev.Points(1) ' point #1 at 0,0,0
  Else
    ReDim TempLev.Points(0)
  End If
End If

If Flags And BlankLev_Polys Then
  ReDim TempLev.Polys(0)
  ReDim TempLev.Polys(0).PtData(0)
End If

If Flags And BlankLev_Ccs Then
  ReDim TempLev.Ccs(0)
End If

getBlankLevel = TempLev

End Function

Public Sub py_getTeamOffset(ByRef Polygon As PolygonData, ByVal Frame As Integer, ByRef ReturnV As D3DVECTOR)
'
' Computes and returns offset coordinates used in team-colors animation.
'
Syntax:
'
' py_GetTeamOffset(Polygon, Frame, ReturnV)
'
' Polygon -- Source polygon to evaluate.
'              * Polygon.TeamC.Type1 must have &H4000 set
'              * Polygon.TeamC.Animated must be > 1
'
' Frame   -- Frame # of animation to compute offsets for.
'              * Starts at 1, unlike WZ
'              * Animations loop.  Frame ##'s greater than
'                Polygon.TeamC.Animated will be wrapped into that range.
'
' Function returns a D3DVector indicating the offset
' from current texture u/v coordinates.
'
' The .z member of the return value is not used and = 0.
'

ReturnV.X = 0
ReturnV.Y = 0
ReturnV.Z = 0

Dim TeamC As TeamSettings
TeamC = Polygon.TeamC

' Must have at least 2 frames of animation
If TeamC.Animated < 2 Then Exit Sub

' Must have a valid width parameter
If TeamC.AnimW < 1 Then Exit Sub

' Wrap large widths to fit.
If TeamC.AnimW > 256 Then TeamC.AnimW = (TeamC.AnimW - 1) Mod 256 + 1

' Wrap frame values to the # of frames specified.
If Frame > TeamC.Animated Then Frame = (Frame - 1) Mod TeamC.Animated + 1

' Frame #1 needs no offset value.
If Frame < 2 Then Exit Sub

' Warzone source team-colors algorithm - very simple and efficient
'
' * NOTE!  Frame=0 represents the first frame
'
' // HACK - fix this!
' framesPerLine = 256 / poly->pTexAnim->textureWidth;
' //should be   framesPerLine = iV_TEXTEX(texPage)->width / poly->pTexAnim->textureWidth;
'        vFrame = 0;
'        While (Frame >= framesPerLine)
'        {
'          frame -= framesPerLine;
'          vFrame += poly->pTexAnim->textureHeight;
'        }
'        uFrame = frame * poly->pTexAnim->textureWidth;
'
'        for (j=0; j<poly->nVrts; j++)
'        {
'          poly->pVrts[j].tu += uFrame;
'          poly->pVrts[j].tv += vFrame;
'        }
'
' Evaluation and implementation:
'
' 1 - Compute maximum # of frames per line, rounding down
Dim LineFrames As Integer
LineFrames = 256 \ Polygon.TeamC.AnimW

' 2 - Compute vertical offset (once for every time Frame>LineFrames)
ReturnV.Y = Polygon.TeamC.AnimH * ((Frame - 1) \ LineFrames)

' 3 - Computer horizontal offset (remainder after the above)
ReturnV.X = Polygon.TeamC.AnimW * ((Frame - 1) Mod LineFrames)



' PIE Slicer's custom algorithm.
'
' Produced same results as WZ 99% of the time, if a lot more complex
'
'Dim StartX As Integer, B As Integer, A As Integer, TmpX As Integer
'Dim EndX As Integer
'
'' Find left-most and right-most corners of texture.
'
'StartX = 256: EndX = 0
'For B% = 1 To UBound(Polygon.PtData)
'  TmpX% = Polygon.PtData(B%).Loc.X
'  If TmpX% < StartX Then StartX = TmpX%
'  If TmpX% > EndX Then EndX = TmpX%
'Next B%
'
'
'' Assume that each frame starts at a multiple of Polygon.TeamC.AnimW,
'' so calculate an initial offset value for frame 0/1
'
'Dim Offset As D3DVECTOR
'Dim OffsetL As Integer
'If Polygon.TeamC.AnimW = 0 Then Exit Sub
'OffsetL = (StartX Mod Polygon.TeamC.AnimW)
'Offset.X = StartX - OffsetL
'Offset.Y = 0
'
'' Compute the frames progressively.
'
'For A% = 1 To Frame
'
'  ' Value to return
'  ReturnV = Offset
'
'  ' Calculate next iteration
'  Offset.X = Offset.X + Polygon.TeamC.AnimW
'
'  ' Test for edge wrapping
'  If Offset.X + (Polygon.TeamC.AnimW - 1) > 256 Then
'
'    Offset.X = 0
'    Offset.Y = Offset.Y + Polygon.TeamC.AnimH
'
'  End If
'Next A%
'
'ReturnV.X = ReturnV.X - StartX + OffsetL

End Sub

Public Sub ListPoint(ByRef Lev As LevelType, ByVal Pt As Integer)
' Updates a point entry in the list box.
Dim Item As ListItem
Set Item = formLists.lstView(0).ListItems(Pt)
With Lev.Points(Pt)
  Item.SubItems(1) = txt_Round1(.Loc.X)
  Item.SubItems(2) = txt_Round1(.Loc.Y)
  Item.SubItems(3) = txt_Round1(.Loc.Z)
End With

End Sub

Public Sub ListCc(ByRef Lev As LevelType, ByVal CC As Integer)
' Updates a connector entry in the list box.
Dim Item As ListItem
Set Item = formLists.lstView(5).ListItems(CC)

With Lev.Ccs(CC)
  Item.SubItems(1) = txt_Round1(.Loc.X)
  Item.SubItems(2) = txt_Round1(.Loc.Y)
  Item.SubItems(3) = txt_Round1(.Loc.Z)
End With

End Sub


Public Function py_Join(ByRef Lev As LevelType, ByVal PolyL As Integer, ByVal PolyH As Integer) As Integer
'
' Fuses two polygons together along one common edge.
'
Syntax:
'
' py_Join(Lev, PolyL, PolyH)
'   Lev   - Submodel of interest.
'   PolyL - Polygon #1's index.
'   PolyH - Polygon #2's index.
'
' When successful, function returns an index to the new polygon.
'
' Remember to update the list data and GUI afterwards!
'
Dim NumPts1 As Integer, NumPts2 As Integer
Dim Share As Integer, BestShare As Integer
Dim Pt As Integer, Pt2 As Integer
Dim TPt As Integer, TPt2 As Integer
Dim StartPt1 As Integer, StartPt2 As Integer
Dim Polydx As Integer, NewNumPts As Integer
Dim A As Integer
Dim NewPt As Integer, PtIndx As Integer

' Ensure that the arguments are supplied in proper order.
If PolyL% = PolyH% Then
  formMaster.SetStatus 0, "4|Cannot join the current polygon to itself", FormMain.statBar
  Exit Function
End If

' Ensure that PolyL variable is the lower-index polygon
If PolyL > PolyH Then
  A = PolyH
  PolyH = PolyL
  PolyL = A
End If

' Cache the polygons being used.
Dim LoPoly As PolygonData, HiPoly As PolygonData
LoPoly = Lev.Polys(PolyL)
HiPoly = Lev.Polys(PolyH)

' Both polygons must belong in the same submodel
If LoPoly.Submodel <> HiPoly.Submodel Then
  formMaster.SetStatus 0, "4|Cannot join polygons from different submodels", FormMain.statBar
  py_Join = -1
  Exit Function
End If

NumPts1% = UBound(LoPoly.PtData)
NumPts2% = UBound(HiPoly.PtData)

' This procedure was difficult to code--stay sharp!

BestShare% = 0

' Verify that these polygons share at least
' two points.....in the SAME ORDER!
For Pt% = 1 To NumPts1%
  Dim TestIndex As Integer
  TestIndex = LoPoly.PtData(Pt).Index
  
  For Pt2% = 1 To NumPts2%
    If TestIndex = HiPoly.PtData(Pt2%).Index Then
      ' If our polygons share one point...
      TPt% = Pt%: TPt2% = Pt2%
      Share% = 1
      Do
        If TPt2 = 1 Then TPt2 = NumPts2 Else TPt2 = TPt2 - 1
        If TPt = NumPts1 Then TPt = 1 Else TPt = TPt + 1
        ' Is the next point shared?
        If LoPoly.PtData(TPt%).Index = HiPoly.PtData(TPt2%).Index Then
          Share% = Share% + 1
        Else
          ' If not, stop our search.
          Exit Do
        End If
        ' Keep looping until we finish the polygon
      Loop Until TPt% = Pt%

      ' Did we find a common edge?
      If Share% > BestShare% Then
        BestShare% = Share%
        ' Index where the common edge starts.
        StartPt1% = Pt%
        StartPt2% = Pt2%
      End If
    End If
  Next Pt2%
Next Pt%

' Double-check to see if we can do this
If BestShare% = NumPts1% And BestShare% = NumPts2% Then
  formMaster.SetStatus 0, "5|Cannot join these two polygons", FormMain.statBar
  Exit Function
ElseIf BestShare% < 2 Then
  formMaster.SetStatus 0, "4|These polygons must share at least two points", FormMain.statBar
  Exit Function
ElseIf MsgBox("Join these two polygons?  (The resulting polygon may need re-skinning)", vbYesNo) = vbNo Then
  Exit Function
End If

SetUndoStep Model, Undo_JoinPolys

' Create the new polygon.
'
Dim NewPoly As PolygonData
NewPoly.Submodel = LoPoly.Submodel
NewPoly.Flags = LoPoly.Flags
NewPoly.TeamC = LoPoly.TeamC

' How many points do we need?
' = sum of all points, minus # of shared points that can be eliminated
'   from each polygon (2 * (shared-1))
'
NewNumPts% = NumPts1% + NumPts2% - 2 * BestShare% + 2
If NewNumPts > WZ_MaxPolySize Then
  MsgBox "Cannot join these two polygons; the resultant polygon would be too complex.", vbCritical
  Exit Function
End If

' Initialize the polygon vertices
ReDim NewPoly.PtData(NewNumPts%)

' Add the point data from our second polygon
For NewPt% = 1 To NumPts2% - BestShare% + 1
  PtIndx% = StartPt2% + NewPt%
  Do Until PtIndx% <= NumPts2%
    PtIndx% = PtIndx% - NumPts2%
  Loop

  NewPoly.PtData(NewPt%) = HiPoly.PtData(PtIndx%)
Next NewPt%

' Then add the point data from the first polygon
For NewPt% = 1 To NumPts1% - BestShare% + 1
  PtIndx% = StartPt1% - NewPt% + 1
  If PtIndx% < 1 Then PtIndx% = PtIndx% + NumPts1%

  NewPoly.PtData(NewNumPts% + 1 - NewPt%) = LoPoly.PtData(PtIndx%)
Next NewPt%

' Verify that the algorithm worked.
For A% = 1 To NewNumPts%
  If NewPoly.PtData(A%).Index < 1 Then
    Debug.Assert False
    Beep
    formMaster.SetStatus 0, "5|Internal failure -- could not join these polygons.", FormMain.statBar
    Exit Function
  End If
Next A%

' Apply the new polygon to the model
Polydx% = UBound(Lev.Polys) + 1
ReDim Preserve Lev.Polys(Polydx%)
Lev.Polys(Polydx%) = NewPoly
  
' Remove the higher polygon first, followed by the lower
py_Remove Lev, PolyH%, 0
py_Remove Lev, PolyL%, 0

' Return the new polygon index.
py_Join = UBound(Lev.Polys)

End Function

Public Function py_Link(ByRef Lev As LevelType, ByVal PolyL As Integer, ByVal PolyH As Integer) As Integer
'
' Links two polygons together.
'
Syntax:
'
' py_Join(Lev, PolyL, PolyH)
'   Lev   - Submodel of interest.
'   PolyL - Polygon #1's index.
'   PolyH - Polygon #2's index.
'
' When successful, function returns an index to the new polygon.
'
' Remember to update the list data and GUI afterwards!
'

' Ensure that the arguments are supplied in proper order.
If PolyL = PolyH Then
  formMaster.SetStatus 0, "4|Cannot link the current polygon to itself", FormMain.statBar
  Exit Function
End If

' Cache the polygons being used.
Dim LoPoly As PolygonData, HiPoly As PolygonData
LoPoly = Lev.Polys(PolyL)
HiPoly = Lev.Polys(PolyH)

Dim NumPts As Integer
NumPts = UBound(LoPoly.PtData)
If NumPts <> UBound(HiPoly.PtData) Then
  formMaster.SetStatus 1, "4|Cannot link polygons of different shapes", FormMain.statBar
  py_Link = -1
  Exit Function
End If

' Both polygons must belong in the same submodel
If LoPoly.Submodel <> HiPoly.Submodel Then
  formMaster.SetStatus 0, "4|Cannot join polygons from different submodels", FormMain.statBar
  py_Link = -1
  Exit Function
End If

' Begin creating the polygons.
Dim PyIndx As Integer

PyIndx = UBound(Lev.Polys) + 1
ReDim Preserve Lev.Polys(UBound(Lev.Polys) + NumPts)

Const Offset As Integer = 0
Dim Vx1 As Integer, Vx2 As Integer
Vx1 = 1
Vx2 = Vx1 Mod NumPts + Offset

Dim Ax As Integer
For Ax = 1 To NumPts
With Lev.Polys(PyIndx)
  ReDim .PtData(4)
  
  .Flags = LoPoly.Flags
  .TeamC = LoPoly.TeamC
  
  .PtData(1) = LoPoly.PtData(Vx1)
  .PtData(4) = HiPoly.PtData(Vx2)
  
  Vx1 = IIf(Vx1 = NumPts, 1, Vx1 + 1)
  Vx2 = IIf(Vx2 = 1, NumPts, Vx2 - 1)
  
  .PtData(2) = LoPoly.PtData(Vx1)
  .PtData(3) = HiPoly.PtData(Vx2)
  
End With
PyIndx = PyIndx + 1
Next Ax

SetUndoStep Model, Undo_JoinPolys

py_Link = PyIndx

End Function



Public Sub py_ResetTexmap(ByRef Level As LevelType, ByVal Polydex As Integer)
'
' Re-initializes a polygon's texture alignment.
'
With py_getNormalA(Level, Polydex)
  py_AutoTexMap Level, Polydex, Int(.X), Int(.Y)
End With

End Sub


Public Sub py_AutoTexMap(ByRef Lev As LevelType, ByVal Polydx As Integer, Optional ByVal HzAngle As Integer = -1, Optional ByVal VtAngle As Integer = -100, Optional ByVal Size As Integer = -1)
Dim DrawPts() As D3DVECTOR
Dim Mini As D3DVECTOR, Maxi As D3DVECTOR
Dim Ctr As D3DVECTOR, Range As D3DVECTOR
Dim Polygon As PolygonData
Dim MyPt As Integer, PolyPts As Integer


If HzAngle = -1 Then getViewAngles HzAngle, 0, 0
If VtAngle = -100 Then getViewAngles 0, VtAngle, 0

' Assigns a polygon's texture-alignment relative to any viewing angle.

' py_AutoTexMap Level, Index, [hz-angle], [vt-angle], [size]
'
' Level    - Reference to the submodel we are going to modify.
' Index    - The polygon index in that submodel.
' hz-angle - Optional; horizontal viewing angle.
' vt-angle - Optional; vertical viewing angle.
' size     - Optional; size of texture space to use.

' Camera angle will be used if angles are omitted.


' Back-up this polygon.  IMPORTANT
Polygon = Lev.Polys(Polydx)

' Get # of points in this polygon.
On Error Resume Next
PolyPts = UBound(Polygon.PtData)
If PolyPts < 0 Or Err.Number = 9 Then Exit Sub
On Error GoTo 0

ReDim DrawPts(PolyPts) As D3DVECTOR

' Compute basic point coordinates.


For MyPt = 1 To PolyPts
  Dim PtIndx As Integer
  PtIndx = Polygon.PtData(MyPt).Index
  DrawPts(MyPt) = Project3D(P3D_MyAngles, Lev.Points(PtIndx).Loc, CamCentre, HzAngle, VtAngle)

  With DrawPts(MyPt)
    If Mini.X > .X Then Mini.X = .X
    If Mini.Y > .Y Then Mini.Y = .Y
    If Maxi.X < .X Then Maxi.X = .X
    If Maxi.Y < .Y Then Maxi.Y = .Y
  End With

Next MyPt

' Where is the polygon located?
Ctr.X = (Mini.X + Maxi.X) / 2
Ctr.Y = (Mini.Y + Maxi.Y) / 2

' How large (visually) is the polygon?
Range.X = Abs(Maxi.X - Mini.X)
Range.Y = Abs(Maxi.Y - Mini.Y)

Dim Radius As Single
Radius = IIf(Range.X > Range.Y, Range.X, Range.Y)

If Radius = 0 Then Radius = 1
If Size > 0 Then Radius = Size

Dim TexC As D3DVECTOR: TexC = formTex.getTexCenterV

' Assign texture coordinates
For MyPt = 1 To PolyPts
  With Lev.Polys(Polydx).PtData(MyPt)
    .Loc.X = TexC.X + (DrawPts(MyPt).X - Ctr.X) * (Sys.ZoomScalar / 2)
    .Loc.Y = TexC.Y + (DrawPts(MyPt).Y - Ctr.Y) * (Sys.ZoomScalar / 2)
  End With
Next MyPt

End Sub



Public Sub py_AddVertex(ByRef Lev As LevelType, ByVal Index As Integer, ByVal Where As Integer, Optional HowMany As Integer = 1)
Dim TestP1 As PointData, TestP2 As PointData
Dim TestPt As PointData
Dim ThisPoly As PolygonData
Dim NumPts As Integer

' Adds a point to a polygon's edge.

With Lev.Polys(Index)
  NumPts = UBound(.PtData)

' Grab two points from our polygon.
  TestP1 = .PtData(Where)
  TestP2 = .PtData(Where Mod NumPts + 1)
  
  ReDim Preserve .PtData(NumPts + HowMany)
End With


' Up-date the polygon's point order.
ThisPoly = Lev.Polys(Index)

Dim A As Integer, B As Integer, Pt As Integer
Dim NewPt As Integer

For A% = 1 To NumPts
  Pt% = Pt% + 1
  Lev.Polys(Index).PtData(Pt%) = ThisPoly.PtData(A%)
  If A% = Where Then
    ' Are we adding one point or two?
    For B% = 1 To HowMany
      ' Add the new point.
      NewPt% = UBound(Lev.Points) + 1
      ReDim Preserve Lev.Points(NewPt%)
      ' Assign the point's coordinates.
      Lev.Points(NewPt%).Loc = v_MidPt(Lev.Points(TestP1.Index).Loc, Lev.Points(TestP2.Index).Loc, B / (HowMany + 1))
      ' Assign texture data
      TestPt.Loc = v_MidPt(TestP1.Loc, TestP2.Loc, B% / (HowMany + 1))
      ' Assign point index
      TestPt.Index = NewPt
      ' Assign the vertex
      Pt% = Pt% + 1
      Lev.Polys(Index).PtData(Pt%) = TestPt
    Next B%
  End If
Next A%

' Wrap up.
formLists.LoadPointData

FormMain.Redraw Everything

End Sub

Public Function py_MoveVertex(ByRef Lev As LevelType, ByVal Poly As Integer, ByVal Vertex As Integer, Optional Where As Integer = -1) As Integer
Dim Item As ListItem
Dim OldPt As Integer, NewPt As Integer
Dim Min As Integer, Max As Integer
' Detaches a polygon's vertex from its point and moves it elsewhere.
'
Syntax:
'
' py_MoveVertex(Lev, Poly, Vertex, [Where])
'   Lev    -- Submodel to use.
'   Poly   -- Index of the desired polygon in that submodel.
'   Vertex -- Vertex to detach.  Specify -1 to detach the entire polygon.
'               * To detach an entire polygon from its neighbors,
'                 set Vertex=-1 and Where=-1
'   Where  -- Where to move a singular vertex.
'               * Valid only if Vertex > 0
'
' Returns < 0 if it fails.
'

' If vertex & destination point are specified, move the vertex elsewhere
If Vertex > 0 And Where > 0 Then
  ' Move this vertex to an existing point
    
  ' Affect the polygon's vertex
  Lev.Polys(Poly).PtData(Vertex).Index = Where
  formMaster.SetStatus 0, "1|Moved this vertex to point #" & Where, FormMain.statBar
  
  Exit Function ' Done!
End If

' Otherwise, detach vertex(es) from other polygons
'
If Vertex = -1 Then
  ' Enforce proper syntax
  If Where > 0 Then
    py_MoveVertex = -1
    Debug.Assert False
    Exit Function
  End If
  
  Min = 1
  Max = UBound(Lev.Polys(Poly).PtData)
Else
  Min = Vertex
  Max = Vertex
End If

Dim Pt As Integer, Py As Integer, Vt As Integer, PtNum As Integer
' Iterate through the polygon's vertex(es) to see if they need detaching
For Pt = Min To Max
  PtNum = Lev.Polys(Poly).PtData(Pt).Index
  Where = -1
  
  For Py = 1 To UBound(Lev.Polys)
  If Py <> Poly Then
    
    With Lev.Polys(Py)
    
      For Vt = 1 To UBound(.PtData)
        ' If a vertex is shared by other polys, detach it
        '
        If .PtData(Vt).Index = PtNum Then
          ' Create a new point for the vertex
          '
          Where = UBound(Lev.Points) + 1
          ReDim Preserve Lev.Points(Where)
          Lev.Points(Where) = Lev.Points(PtNum)
          
          ' Update the vertex & go to the next lap
          Lev.Polys(Poly).PtData(Pt).Index = Where
          Exit For
        End If
      Next Vt
    
    End With
    If Where > 0 Then Exit For
  
  End If
  Next Py
  
Next Pt

If Where = -1 Then
  formMaster.SetStatus 0, "4|This " & IIf(Vertex = -1, "polygon", "vertex") & " is not attached to other polygons"
  py_MoveVertex = -1
  Exit Function
End If
 
If Vertex = -1 Then
  formLists.SetCurPoint Where
End If

' Update the Points listbox.
formLists.LoadPointData

formMaster.SetStatus 0, "1|Assigned this " & IIf(Vertex = -1, "polygon", "vertex") & " to separate points", FormMain.statBar

End Function

Public Sub SetItemStat(Item As ListItem, Status As Integer)
Debug.Assert Not Item Is Nothing

If (Status And stat_SelMask) = stat_Sel Then
  Item.Text = "*" & Item.Index
  Item.Bold = True
  Item.ForeColor = &HFF0000
Else
  Item.Text = "  " & Item.Index
  Item.Bold = 0
  Item.ForeColor = IIf(Status And stat_Hide, &HC0C0C0, 0)
End If

Dim SubItem As ListSubItem
For Each SubItem In Item.ListSubItems
  SubItem.ForeColor = Item.ForeColor
Next SubItem

End Sub

Public Sub py_DeleteVertex(ByRef Polygon As PolygonData, ByVal Vertex As Integer)
Dim ThisPoly As PolygonData
Dim NumPts As Integer, A As Integer
'
' Removes a point (vertex) from the supplied polygon.
'
Syntax:
'
' py_DeleteVertex(Polygon, Vertex)
'   Polygon  - A polygon reference to modify.
'   Vertex   - Which vertex to remove.

' Grab # of points.
NumPts% = UBound(Polygon.PtData)
If NumPts < 4 Then Exit Sub

' Back-up this polygon, just in case
ThisPoly = Polygon

' Delete vertex from this polygon.
For A% = Vertex To NumPts% - 1
  Polygon.PtData(A%) = Polygon.PtData(A% + 1)
Next A%

' Crop the array.
ReDim Preserve Polygon.PtData(NumPts% - 1)

End Sub

Public Function py_AddStd(NumPts As Integer, ByRef Locale As D3DVECTOR, ByRef Mouse As D3DVECTOR, View1 As Integer, View2 As Integer, ByRef Radi As Single, ByRef Angl As Single, Optional TwoSided As Boolean = False) As LevelType
'
' Creates submodel data for a standard (regular) polygon.
'
Syntax:
'
' py_AddStd(NumPts, Locale, Mouse, View1, View2, Radi, Aggl, [TwoSided = False] )
'   NumPts   - Shape of polygon.
'   Locale   - Center-point of polygon.
'   Mouse    - Mouse location relative to (0,0,0).
'              * Radius/angle will be determined from it.
'   View1    - Horizontal rotation.
'   View2    - Vertical rotation.
'   Radi     - (output) Radius of polygon created
'   Angl     - (output) Offset angle of polygon created
'   TwoSided - Whether to make it a two-sided polygon or not.
'              * default=False
'
' Function returns submodel data with 1 polygon when successful.

Dim VMouse As D3DVECTOR
Dim TempPt As D3DVECTOR
Dim PyOffset As D3DVECTOR
Dim DrawPt As D3DVECTOR

ReDim py_AddStd.Points(0)
ReDim py_AddStd.Polys(0)
ReDim py_AddStd.Ccs(0)
If NumPts < 3 Or NumPts > 8 Then Exit Function

' Compute rotation angle and radius.
'
DrawPt = Project3D(P3D_Math, Locale, CamCentre, 0, 0)
VMouse.Z = DrawPt.Z
VMouse.X = -Mouse.X / Sys.ZoomScalar
VMouse.Y = -Mouse.Y / Sys.ZoomScalar

Angl = GetArcTan(VMouse, ATN_UseCtrPt, DrawPt, NumPts)
Radi = Format(Distance(VMouse, DrawPt), "0.0")

' Create the submodel structure
With py_AddStd
  ReDim .Points(NumPts) As PointData
  ReDim .Polys(1) As PolygonData
  ReDim .Polys(1).PtData(NumPts) As PointData
End With

' Create the polygon's points.
Dim A As Integer
For A% = 1 To NumPts

  TempPt.X = -Radi * Cos((A% + Angl) / NumPts * m_2pi)
  TempPt.Z = 0
  TempPt.Y = Radi * Sin((A% + Angl) / NumPts * m_2pi)

  ' Figure out where the visual coordinates belong,
  ' in terms of world coordinates
  PyOffset = ProjectInv3D(TempPt, View1, View2)

  ' Assign a new point for the polygon
  With PyOffset
    .X = .X + Locale.X
    .Y = .Y + Locale.Y
    .Z = .Z + Locale.Z
  End With
  
  py_AddStd.Points(A%).Loc = PyOffset

  py_AddStd.Polys(1).PtData(A%).Index = (NumPts - A%) + 1
Next A%

' Assign initial polygon texture alignment.
py_ResetTexmap py_AddStd, 1

' Assign polygon attributes and select status
With py_AddStd.Polys(1)
  .Flags = IIf(TwoSided, &H2200, &H200)
  .StatFlags = stat_Sel
End With

' Return the data
py_AddStd = py_AddStd

End Function

Public Function py_Slice(ByRef Lev As LevelType, ByVal Record As Integer, Pt1 As Integer, Pt2 As Integer) As Integer
Dim This_Poly As PolygonData
' Slices a polygon along one of its diagonals.

Syntax:
' py_Slice polygon, Pt1, Pt2


' NOTE:  Pt1 must be less than Pt2.
Dim Pt3 As Integer
If Pt1 > Pt2 Then
  Pt3% = Pt1
  Pt1 = Pt2:  Pt2 = Pt3%
End If

' Create a temporary polygon to copy from....
This_Poly = Lev.Polys(Record)
Dim NumPts As Integer
NumPts% = UBound(This_Poly.PtData)

' Prepare two new polygons at the end of the list
Dim Num1 As Integer, Num2 As Integer
With Lev
  Num1% = UBound(.Polys) + 1
  ReDim Preserve .Polys(Num1%)
  .Polys(Num1).Submodel = .Polys(Record).Submodel
  
  Num2% = UBound(.Polys) + 1
  ReDim Preserve .Polys(Num2%)
  .Polys(Num2).Submodel = .Polys(Record).Submodel
End With

' Create one polygon.
Dim NumPts1 As Integer, NumPts2 As Integer
Dim D As Integer, A As Integer
D% = 0
For A% = Pt1 To Pt2
  D% = D% + 1
  ' Assign point data.
  ReDim Preserve Lev.Polys(Num1%).PtData(D%)
  Lev.Polys(Num1%).PtData(D%) = This_Poly.PtData(A%)
Next A%
NumPts1 = D

' Initialize the polygon's data
Lev.Polys(Num1%).Flags = This_Poly.Flags
If This_Poly.Flags And &H4000& Then
  With Lev.Polys(Num1)
    .TeamC = This_Poly.TeamC
  End With
End If

Dim Item As ListItem
Set Item = formLists.lstView(1).ListItems.Add(, , "  " & Num1% - 1)
Item.SubItems(1) = Hex$(This_Poly.Flags)
Item.SubItems(2) = D%

' Create the other polygon.
D% = 0
Dim PtNum As Integer
For A% = Pt2 To Pt1 + NumPts%
  D% = D% + 1
  PtNum% = IIf(A% > NumPts%, A% - NumPts%, A%)
  ' Assign point data.
  ReDim Preserve Lev.Polys(Num2%).PtData(D%)
  Lev.Polys(Num2%).PtData(D%) = This_Poly.PtData(PtNum%)
Next A%
NumPts2 = D

' Initialize the other polygon.
Lev.Polys(Num2%).Flags = This_Poly.Flags
If This_Poly.Flags And &H4000& Then
  With Lev.Polys(Num2)
    .TeamC = This_Poly.TeamC
  End With
End If

Set Item = formLists.lstView(1).ListItems.Add(, , "  " & Num2% - 1)
Item.SubItems(1) = Hex$(This_Poly.Flags)
Item.SubItems(2) = D

' We don't need the old polygon no more
py_Remove Lev, Record, Sync_UpdateLists

' Return the index for the larger of the 2 slices.
' ( #-1 because a polygon was removed from a lower index)
py_Slice = IIf(NumPts2 > NumPts1, Num2, Num1) - 1

formMaster.SetStatus 0, "1|Split polygon #" & Record & " into two pieces", FormMain.statBar

' Wrap up with GUI operations
formLists.SetCurPoly py_Slice
FormMain.Redraw Everything

End Function

Public Function pt_Remove(ByRef Lev As LevelType, ByVal Record As Integer, Optional DataOnly As Boolean = False) As Integer
Dim MaxPoints As Integer
'
' Removes a point from the model.
'
Syntax:
'
' pt_Remove(Lev, Record, [DataOnly])
'   Lev      - PIE data to remove a point from.
'   Record   - Point index to delete.
'   DataOnly - If TRUE, suppresses errors and doesn't update any GUI elements.


If Record < 1 Then
  If Not DataOnly Then formMaster.SetStatus Msg_NeedPts, FormMain.statBar
  pt_Remove = -1
  Exit Function
End If

' Quick test.
MaxPoints = UBound(Lev.Points)
If MaxPoints < 1 Then
  pt_Remove = -1
  Exit Function
End If

Dim FPolys As Integer
FPolys% = 0

' Is the point being used by any polygons?
Dim A As Integer, B As Integer
For A% = 1 To UBound(Lev.Polys)
  For B% = 1 To UBound(Lev.Polys(A%).PtData)
    If Record = Lev.Polys(A%).PtData(B%).Index Then
    ' If so, abort!!
      If Not DataOnly Then formMaster.SetStatus 0, "4|Point #" & Record & " has polygons attached to it", FormMain.statBar
      pt_Remove = -1
      Exit Function
    End If
  Next B%
Next A%

' User confirmation when deleting the last point.
If MaxPoints = 1 Then
  If MsgBox("Caution:  You are about to delete the last point in this model.  Proceed?", vbExclamation + vbOKCancel) = vbCancel Then Exit Function
End If

' Update all polygon references
For A% = 1 To UBound(Lev.Polys)
  For B% = 1 To UBound(Lev.Polys(A%).PtData)
    With Lev.Polys(A%).PtData(B%)
      If .Index > Record Then .Index = .Index - 1
    End With
  Next B%
Next A%

' Remove this point from the array...
Dim Pt As Integer, Nm As Integer
For Pt% = Record To UBound(Lev.Points) - 1
  Lev.Points(Pt%) = Lev.Points(Pt% + 1)
Next Pt%

ReDim Preserve Lev.Points(UBound(Lev.Points) - 1)

' ...and from the listbox, if desired
If Not DataOnly Then
  With formLists.lstView(0)
    .ListItems.Remove Record
    For Nm% = Record To UBound(Lev.Points)
      .ListItems(Nm%).Text = Nm%
    Next Nm%
  End With

  If Record < 0 Then Record = 0
  If Record >= MaxPoints Then Record = MaxPoints - 1
  
  formMaster.SetStatus 0, "1|Removed one point.", FormMain.statBar
End If

pt_Remove = Record

If ThisPoly > 0 Then formTools.py_LoadPointOrder

End Function

Public Function py_Shatter(ByRef Lev As LevelType, ByVal Record As Integer, Style As Integer) As Integer
Dim TempP As D3DVECTOR, Polygon As PolygonData, NumPts As Integer
' Fractures a polygon into smaller triangles.

Syntax:
' py_Shatter(Lev, Record, Style)

' Lev    -- Submodel of interest
'
' Record -- Polygon index ## to shatter
'
' Style  -- 0=shatter from a central point
'
'           1=shatter along trianglefan lines
'             * doesn't work on triangles
'
'
' Returns >0 when successful.
' Returns -1 if no polygon could be created.
'
'

If Record < 1 Then py_Shatter = -1: Exit Function

' Create common reference info
'
Polygon = Lev.Polys(Record)
NumPts = UBound(Polygon.PtData)

Select Case Style
Case 1
  ' Shatter along trianglefan lines
  '''''
  
  ' Make sure trianglefans are avaiable
  NumPts% = UBound(Polygon.PtData)
  If NumPts < 4 Then
    formMaster.SetStatus Msg_NoSplitTri, 0, FormMain.statBar
    py_Shatter = -1
    Exit Function
  End If
  
  Dim PyPt As Integer, PolyN As Integer
  For PyPt% = 1 To NumPts% - 2
    PolyN% = UBound(Lev.Polys) + 1
  
    ' Add a triangle.
    ReDim Preserve Lev.Polys(PolyN%)
    With Lev.Polys(PolyN)
      .Submodel = Polygon.Submodel
      .Flags = Polygon.Flags
    
      ReDim Preserve Lev.Polys(PolyN%).PtData(3)
      ' First point
      .PtData(1) = Polygon.PtData(1)
      ' Second point
      .PtData(2) = Polygon.PtData(1 + PyPt%)
      ' Center point
      .PtData(3) = Polygon.PtData(2 + PyPt%)
    End With
  
    ' Update the polygon list box
    Dim Item As ListItem
    Set Item = formLists.lstView(1).ListItems.Add(, , " " & PolyN%)
    ListPolyData Polygon, Item
  
  Next PyPt%
  
Case 0
  ' Shatter around a central point
  '''''
  
  ' Create the central point & its corresponding texmapping
  Dim newPtNum As Integer, TexP As D3DVECTOR
  newPtNum% = pt_AddNew(Lev)
  Lev.Points(newPtNum%).Loc = pt_CenterOf(PTC_CurPy, Lev, Record, 0)
  
  TexP = pt_CenterOf(PTC_TexPy Or PTC_IntMath, Lev, Record, 0)
  
  ' Create the triangles
  For PyPt% = 1 To NumPts%
    
    ' Add a triangle
    PolyN% = UBound(Lev.Polys) + 1
    ReDim Preserve Lev.Polys(PolyN%)
    With Lev.Polys(PolyN)
      .Submodel = Polygon.Submodel
      .Flags = Polygon.Flags
    
      ' Define the triangle
      ReDim Preserve .PtData(3)
      
      ' First point on triangle
      .PtData(1) = Polygon.PtData(PyPt%)
      ' Second point on triangle
      .PtData(2) = Polygon.PtData(PyPt% Mod NumPts% + 1)
      ' Center point
      .PtData(3).Loc = TexP
      .PtData(3).Index = newPtNum%
    End With
  
    ' Update the polygon list box
    Set Item = Nothing
    Set Item = formLists.lstView(1).ListItems.Add(, , " " & PolyN%)
    ListPolyData Polygon, Item
  
  Next PyPt%
  
End Select

' Cleanup
'''''

' Remove the old polygon & update the list data, but DO NOT
' change the current polygon
py_Remove Lev, Record, Sync_UpdateLists

' Return value. Minus one because of the removed polygon.
py_Shatter = (PolyN - 1)

End Function

Public Sub py_Remove(ByRef Lev As LevelType, ByVal Number As Integer, ByVal SyncFlags As Long)
'
' Removes a polygon from a submodel.
'
Syntax:
'
' py_Remove(Lev, Number, SyncFlags)
'   Lev       - Submodel of interest.
'   Number    - Polygon index to remove.
'   SyncFlags - Flags for auto cleanup:
'                Sync_SetCurrent - Set current polygon
'                Sync_DoRedraw   - Perform a redraw on formMain

Dim MaxPoly As Integer

' Remove the polygon from the data structure...
MaxPoly% = UBound(Lev.Polys)
If MaxPoly < 1 Then Exit Sub

Dim Poly As Integer, Nm As Integer
For Poly% = Number To MaxPoly% - 1
  Lev.Polys(Poly%) = Lev.Polys(Poly% + 1)
Next Poly%
ReDim Preserve Lev.Polys(MaxPoly% - 1)

' ... and from the list box, if desired
If SyncFlags And Sync_UpdateLists Then
  Dim List As ListItems
  Set List = formLists.lstView(1).ListItems
  
  ' Remove the ListItem & update the rest
  List.Remove Number
  For Nm% = Number To List.Count - 1
    SetItemStat List(Nm), Lev.Polys(Nm).StatFlags
  Next Nm%
  
  formMaster.SetStatus 0, "1|Removed polygon #" & Number, FormMain.statBar

  If SyncFlags And Sync_SetCurrent Then
    formLists.SetCurPoly (Number - 1), SyncFlags
  End If
End If


End Sub


Public Function py_getCullFace(ByRef Lev As LevelType, ByVal Polygon As Integer) As Integer
'
' Evaluates the cull-face status of a polygon. Used primarily during draw routines.
'
' * This routine is time-critical
'
Syntax:
'
' py_getCullFace(Lev, Polygon)
'   Lev      - Submodel of interest
'   Polygon  - Polygon index # to evaluate
'
' Return value is a series of bitflags representing the visible
' status of the evaluated polygon's trifans:
'   +&H1  - Trifan #1
'   +&H2  - Trifan #2
'   +&H4  - Trifan #3
'   +&H8  - Trifan #4
'   +&H10 - Trifan #5
'   +&H20 - Trifan #6
'   +&H40 - Trifan #7
'   +&H80 - Trifan #8
' The polygon is visible when ANY flag is set (check for zero).
' Since a polygon can have up to 10 points, up to 8 flags may be returned.
'
' Function operates by evaluating the dot-product between the camera vector
' and each normal vector in the polygon.
' When dot-product is <0, culling will be performed for that trifan.


Dim V1 As D3DVECTOR, V2 As D3DVECTOR
Dim A As Integer, NumPts As Integer, Flag As Integer

' Grab a vector pointing towards the camera
V2 = v_ViewAngle(True)

py_getCullFace = 0

' Start with the LSB flag (for trifan #1) and evaluate the individual trifans.
Flag = 1
NumPts = UBound(Lev.Polys(Polygon).PtData)
For A = 2 To NumPts - 1
  V1 = py_getNormalV(Lev, Polygon, A, 0)
  
  If (V1.X * V2.X + V1.Y * V2.Y + V1.Z * V2.Z) >= 0 Then
    py_getCullFace = py_getCullFace Or Flag
  End If
  
  Flag = Flag * 2 ' shl 1 bit
Next A

End Function

Public Function getColorI(ByVal Depth As Single) As Long
'
' Calculates & returns a color-intensity value based upon
' the Depth parameter.  Kinda like z-fog, farther depth values return lower
' values, closer depth values return higher values.
'
Syntax:
'
' getColorI(Depth)
'
'   Depth - Z (depth) coordinate to use.
'
' Output value varies from 32 to 255.
'


Dim ColrV As Integer
If Depth < 0 Then
  ColrV = 192 + 4 * (Log(1 - Depth) / m_Log2)
Else
  ColrV = 192 - 4 * (Log(Depth + 1) / m_Log2)
End If

If ColrV < 32 Then getColorI = 32: Exit Function
If ColrV > 255 Then getColorI = 255: Exit Function
getColorI = ColrV

End Function


Public Sub SetMoveRate(Optional StepV As Integer = -1, Optional DoGUI As Boolean = False)
'
' Computes & assigns the editor move rate (Sys.MoveRate) factor.
'
Syntax:
'
' SetMoveRate([StepV], [DoGUI])
'
' StepV  - Move step to calculate from
'            * -1 = use Sys.MoveStep
' DoGUI  - Whether to update formTools.lblMove or not.
'

If StepV = -1 Then StepV = Sys.MoveStep
If StepV < MoveStep_Min Or StepV > MoveStep_Max Then Exit Sub

If User.Settings And Set_MoveBinary Then
  ' Binary scale
  Sys.MoveRate = 2 ^ StepV

Else
  ' Decimal (125) scale
  
  Select Case StepV Mod 3
  Case 0: Sys.MoveRate = 10 ^ (StepV \ 3)
  Case 1: Sys.MoveRate = 2 * 10 ^ (StepV \ 3)
  Case 2: Sys.MoveRate = 5 * 10 ^ (StepV \ 3)
  End Select

End If

If DoGUI Then formTools.lblMove.Caption = txt_Round1(Sys.MoveRate)

End Sub

Public Sub SetChanged(ByRef Pie As PieModel, ByVal File As Integer)
' Marks a PIE file, or the overall model, as having been changed.
'
' Also switches on the auto-save timer.

' When a PIE file or the model has been saved, clear the changes manually
' by assigning .Changed = .Loaded (or .Loaded = .Changed) .

Dim CTimer As Timer
Set CTimer = formMaster.timeSave

Dim St As Single
St = Timer

If File = 0 Then
  Pie.Changed = St
Else
  Pie.Files(File).Changed = St
End If

If Not CTimer.Enabled Or Not IsNumeric(CTimer.Tag) Then CTimer.Tag = (User.AutoSave * 60)
CTimer.Enabled = True

End Sub

Public Sub ClearAllChanges(File As Integer)
' Clears all changes made to the current model
' and all its constituent PIE files.
'
' In other words, the "Save Changes?" box will not appear
' again until the next change to a PIE file (or model) is made.


' To discourage calling this function recklessly,
' a file number must be specified.
If File = 0 Then
  Debug.Assert False
  Exit Sub
End If

Model.Changed = Model.Loaded
Dim F As Integer
For F = 1 To UBound(Model.Files)
  Model.Files(F).Changed = Model.Files(F).Loaded
Next F

End Sub

Public Sub SetThisFile(ByVal File As Integer, Optional SetCam As Boolean = False)
If File < 1 Or File > UBound(Model.Files) Then Exit Sub
'
' Switches to a specified file in the current PIE model.
'
Syntax:
'
' SetThisFile File, [SetCam]
'
'   File   -- PIE file of interest
'   SetCam -- Whether to set the camera to the PIE file's origin or not
'
' * Remember to do a Redraw on formMain after calling this!
'

With FormMain.lstFiles
  If .ListIndex <> File - 1 Then
    SkipEvents = True
      .ListIndex = File - 1
    SkipEvents = False
  End If
End With

ThisFile = File

' Update the "Save As" commands.
'
formMaster.UpdateSaveAsMenu Model.Files(File).FileName

' Fill formLists with the PIE file's data
'
formLists.LoadPointData
formLists.SetCurPoint 1

formLists.LoadPolyData
formLists.SetCurPoly 1

Dim PageN As Integer

SkipEvents = True
With Model.Files(ThisFile)
    
  Dim TexName As String
  PageN = GetTexPageNum(.TexPage, TexName)
  
  formTex.txtPage(0).Text = PageN
  formTex.txtPage(1).Text = TexName
  formTex.txtPage_Change 1

  If PageN > 0 Then
    If .TexFile = 0 Then ' If texturepage isn't in vidmem...
      formTex.SetDefaultTexture .TexPage
    End If
  End If
  
  formLists.lstPieType.ListIndex = Model.Files(ThisFile).PieType

End With
SkipEvents = False

' If requested, set the camera point (manually)
'
If SetCam Then
  Dim Mins As D3DVECTOR, Maxs As D3DVECTOR, V As D3DVECTOR
  getPieSizeBox Model.Files(ThisFile).Levs, Mins, Maxs
  CamCentre.X = (Mins.X + Maxs.X) \ 2
  CamCentre.Y = (Mins.Y + Maxs.Y) \ 2
  CamCentre.Z = (Mins.Z + Maxs.Z) \ 2
End If

'SetUndoStep Model.Files(ThisFile), 0

End Sub

Public Function getThisFile() As Integer
getThisFile = FormMain.lstFiles.ListIndex + 1
ThisFile = getThisFile
End Function


Public Function file_LoadPCK(ByVal FileName As String, ByRef Data As PieModel) As Long
Dim File As Integer, CurFile As Integer
Dim N As Integer, F As Integer
Dim Header As String, Version As Long
Dim PathName As String

' Basic syntax checks
If FileName = "" Or Dir(FileName) = "" Then
  Debug.Assert False
  file_LoadPCK = FE_LoadPck Or FE_FileNotFound
  Exit Function
End If

' Store pathname.  Pathname includes the backslash
N = InStrRev(FileName, "\")
PathName = Left$(FileName, N)

' Prep the file for loading
Dim TempData As PieType
Dim LineData As String
ReDim Data.Files(0)
ReDim Data.Tex_Files(0)

File = FreeFile
Open FileName For Input As #File

Seek #File, 1
Line Input #File, Header$

If Left$(Header, 5) <> "PieDX" Then
  file_LoadPCK = FE_LoadPck Or FE_BadFileHeader
  Exit Function
End If

Version = 100 * Mid(Header, 7, 1) + 10 * Mid(Header, 9, 1) + Mid(Header, 11, 1)

If Version > VerPCK Then
  MsgBox "Cannot load this model; it was saved in a newer version of Pie Slicer."
  file_LoadPCK = FE_LoadPck Or FE_NewerVersion
  
  Close #File
  Exit Function
End If

Do
  ' Data lines starting with "{" should be ignored
  '
  Line Input #File, LineData$
  
  ' Determine what keyword is used here
  '
  N = InStr(LineData, "=")
  If N = 0 Then N = 1
  Dim Key As String, Word As String
  Key = Left$(LineData, N - 1)
  Word = Mid$(LineData, N + 1)
  
  Select Case Key
  Case "File", "Body", "Prop", "Tur0" To "Tur3", "Flas", "Proj", "Hide"
    ' PIE file keywords
    '
    
    Dim Ptr As Long: Ptr = Seek(File)
    
    ' Resolve PIE file's full path/filename
    '
    Dim PieFileName As String
    If InStr(Word, ":") = 0 Then
      PieFileName = PathName & Word
    Else
      PieFileName = Word
    End If
    
    ' Determine whether to load the file internally or externally
    '
    Line Input #File, LineData
    TempData.FileName = ""
    
    If Left$(LineData, 9) = "FileData=" Then
      '
      ' Load an internal PIE file from within the PCK structure
      '
      Dim PieVer As Integer
      On Error Resume Next
        PieVer = Mid(LineData, 10)
        If Err.Number Then PieVer = 0
      On Error GoTo 0
      
      Dim Code As Long
      Code = file_LoadPCK_FileData(File, PieVer, TempData)

      If Code = FE_LoadPckPie Or FE_BadPckSrcFile Then
        TempData.FileName = ""
      Else
        TempData.FileName = PieFileName
      End If
      
    End If
    
    If TempData.FileName = "" Then
      ' Or, load an external PIE file.
      '
      ' * If an error occured while reading embedded PIE data,
      '   Pie Slicer will resort to loading the PIE file externally
      Select Case file_GetType(PieFileName)
      Case Format_PieFile
        ' Only PIE files currently allowed in model data
      
      Case Else
        Debug.Assert False
        file_LoadPCK = FE_LoadPck Or FE_BadPckSrcFile
        
        Close #File
        Exit Function
      End Select
      
      ' Load the file data
      Code = file_LoadPIE(PieFileName, TempData)
    
      ' Rewind the file pointer a little bit
      Seek #File, Ptr
    
    End If
    
    ' Verify the loaded data.
    '
    If Code < 1 Then
      Debug.Assert False
      file_LoadPCK = Code
      Exit Function
    End If
    
    ' Record the loaded data when successful
    '
    CurFile = UBound(Data.Files) + 1
    ReDim Preserve Data.Files(CurFile)
    Data.Files(CurFile) = TempData
  
    ' PCK v120 and up: Apply the PIE filetype here
    '
    Select Case Key
    Case "File": Data.Files(CurFile).PieType = TYPE_SINGLE ' Same as old PCK files
    Case "Body": Data.Files(CurFile).PieType = TYPE_BODY
    Case "Prop": Data.Files(CurFile).PieType = TYPE_PROP
    Case "Tur0": Data.Files(CurFile).PieType = TYPE_GNDTUR
    Case "Tur1": Data.Files(CurFile).PieType = TYPE_AIRTUR
    Case "Tur2": Data.Files(CurFile).PieType = TYPE_MUZL
    Case "Tur3": Data.Files(CurFile).PieType = TYPE_MUZLCYB
    Case "Flas": Data.Files(CurFile).PieType = TYPE_FLASH
    Case "Proj": Data.Files(CurFile).PieType = TYPE_PROJ
    Case "Hide": Data.Files(CurFile).PieType = TYPE_HIDE
    Case Else
      Debug.Assert False
    End Select
  
  Case "Texture"
    ' Texture (vidmem) management
    '
    Dim TexFile As String, TexF As Integer
    TexFile = Mid$(LineData, N + 1)
    
    ' Resolve the texture path/filename
    '
    If Mid$(LineData, N + 5, 1) = "(" Then
      TexFile = App.Path & "\" & TexFile
    Else
      TexFile = App.Path & "\TexData\" & TexFile
    End If
    
    ' Record the texpage entry
    TexF = UBound(Data.Tex_Files) + 1
    ReDim Preserve Data.Tex_Files(TexF)
    Data.Tex_Files(TexF) = TexFile
    
    If Version < 110 Then
      ' Original PCK: textures specified per each PIE file --
      '   assign texture index to only one PIE file
      Data.Files(CurFile).TexFile = TexF
    Else
      ' PCK v110: Textures specified after all PIE files --
      '   search & assign texture index as needed
      Dim Pg As Integer
      Pg = GetTexPageNum(TexFile, "")
      
      For F = 1 To UBound(Data.Files)
        If GetTexPageNum(Data.Files(F).TexPage, "") = Pg Then
          Data.Files(F).TexFile = TexF
        End If
      Next F
    End If
  
  End Select
Loop Until Seek(File) > LOF(File)

Close #File

' Finish loading the data
Data.FileName = FileName

' Declare it as "no changes made"
Data.Loaded = Timer
Data.Changed = Data.Loaded

' Resolve model composition info (PIE hookups)
SetPieOrigin Data, 0

file_LoadPCK = UBound(Data.Files)

End Function

Public Function py_Mirror(ByRef Lev As LevelType, ByVal Axis As Integer) As Integer
Dim Rota As D3DVECTOR
Rota.X = 1: Rota.Y = 1: Rota.Z = 1

' Mirrors the selected polygons about an axis and returns the # of polygons
' added.

' Set up our mirroring axis
Select Case Axis
Case 1: Rota.X = -1
Case 2: Rota.Y = -1
Case 3: Rota.Z = -1
Case Else: Exit Function
End Select

Dim PtMax As Integer, PyMax As Integer
PtMax = UBound(Lev.Points) ' Current # of points in submodel
PyMax = UBound(Lev.Polys) ' Current # of polygons in submodel

' This will identify which points to copy
Dim Selec() As Boolean
ReDim Selec(PtMax)

' This will hold the new indexes corresponding to copied points
Dim Transform() As Integer
ReDim Transform(PtMax)
  
Dim Po As Integer, Pt As Integer
Dim A As Integer, PtIndx As Integer

' Take note of points belonging to selected polygons.  Ignore others.
For Po = 1 To PyMax
  With Lev.Polys(Po)
    If .StatFlags = 1 Then
      For A = 1 To UBound(.PtData)
        Selec(.PtData(A).Index) = True
      Next A
    End If
  End With
Next Po

' Copy the points we designated above.
PtIndx = 0
For Pt = 1 To PtMax
  If Selec(Pt) Then
    PtIndx = PtIndx + 1

    ' Indicate what new index corresponds to the old point
    Transform(Pt) = PtMax + PtIndx
    
    ' Add the new point.
    ReDim Preserve Lev.Points(PtMax + PtIndx)
    Lev.Points(PtMax + PtIndx) = Lev.Points(Pt)
  
    ' Mirror the point about respective axis (simplified matrix math).
    With Lev.Points(PtMax + PtIndx)
      .Loc.X = .Loc.X * Rota.X
      .Loc.Y = .Loc.Y * Rota.Y
      .Loc.Z = .Loc.Z * Rota.Z
    End With
  End If
Next Pt

' Copy the selected polygons.
Dim PyIndx As Integer, Py As Integer, Polydx As Integer
For Py = 1 To PyMax
  If Lev.Polys(Py).StatFlags = 1 Then
    
    PyIndx = PyIndx + 1 ' # of polygons copied
    
    ' Copy the selected polygon
    
    Polydx = (PyMax + PyIndx) ' New index of polygon
    ReDim Preserve Lev.Polys(Polydx)
    Lev.Polys(Polydx) = Lev.Polys(Py)

    Dim PtIndex As Integer, PyPts As Integer
    PyPts = UBound(Lev.Polys(Py).PtData)
    
    ' Update the point indexes.  Note the reversed point order.
    
    For Pt = 1 To PyPts
      With Lev.Polys(Polydx)
        If Pt = 1 Then
          .PtData(1) = Lev.Polys(Py).PtData(1)
        Else
          .PtData(Pt) = Lev.Polys(Py).PtData(PyPts + 2 - Pt) ' Flips normal v.
        End If
        .PtData(Pt).Index = Transform(.PtData(Pt).Index) ' Update point index.
      End With
    Next Pt
  End If
Next Py

' Return # of mirrored polygons.
py_Mirror = PyIndx

End Function

Public Function py_isVertexPt(Polygon As PolygonData, ByVal Point As Integer) As Integer
'
' Tests whether a point is attached to a given polygon or not.
'
' May be performed on any polygon, even tempvars.
'
Syntax:
'
' py_isVertexPt(Polygon, Point)
'
'    Polygon - Polygon of interest.
'    Point   - Point index to check for.
'
' Returns the vertex # which represents that point,
' or zero if the point is not included in the polygon.

Dim A As Integer
For A = 1 To UBound(Polygon.PtData)
  If Polygon.PtData(A).Index = Point Then py_isVertexPt = A: Exit Function
Next A

End Function

Public Sub SetOnTop(ByRef Form As Form, ByVal IsOnTop As Integer)
' Applies the "Always On Top" option, usually for Pie Slicer DX's master window.

Dim Top As Long, Left As Long, Wide As Long, High As Long
Dim TX As Single: TX = Screen.TwipsPerPixelX
Dim TY As Single: TY = Screen.TwipsPerPixelY

Left = Form.Left \ TX
Top = Form.Top \ TY

Wide = Form.Width \ TX
High = Form.Height \ TY

If IsOnTop Then
  SetWindowPos Form.hWnd, HWND_TOPMOST, Left, Top, Wide, High, SWP_NOACTIVATE  'Or SWP_SHOWWINDOW
Else
  SetWindowPos Form.hWnd, HWND_NOTTOPMOST, Left, Top, Wide, High, SWP_NOACTIVATE  'Or SWP_SHOWWINDOW
End If

End Sub

Public Function py_SetFanPoint(ByRef Polygon As PolygonData, ByVal Vertex As Integer) As Integer
' Sets the triangle-fan point of a polygon.
'
'
' py_SetFanPoint(polygon, vertex)
'
' Polygon -- The polygon to affect
' Vertex  -- Index of the vertex which will be the new triangle-fan point

' Returns >0 when successful.

Dim OldPoly As PolygonData
OldPoly = Polygon

If Vertex < 1 Then
  Debug.Assert False
  Exit Function
End If

Dim NewV As Integer, OldV As Integer, NumPts As Integer
NumPts = UBound(Polygon.PtData)

If NumPts < 4 Then
  formMaster.SetStatus 0, "3|This function does not apply to triangles", FormMain.statBar
  py_SetFanPoint = -1
  Exit Function
End If

OldV = Vertex
For NewV = 1 To NumPts
  Polygon.PtData(NewV) = OldPoly.PtData(OldV)
  OldV = (OldV Mod NumPts) + 1
Next NewV

py_SetFanPoint = Vertex

End Function

Public Function py_isCoplanar(ByRef Lev As LevelType, ByVal Index As Integer) As Single
'
' Returns a value indicating whether a polygon is coplanar (flat) or not.
'
Syntax:
'
' py_isCoplanar(Lev, Index)
'   Lev   - Submodel of interest.
'   Index - Polygon index to check.
'
' Function computes normal vectors for all trianglefans in the polygon,
' sums them, then divides the magnitude by the # of trianglefans.
'
' If a polygon is 100% coplanar (flat), the return value will be 1.
'

Dim TempV As D3DVECTOR
Dim Vector1 As D3DVECTOR, Vector2 As D3DVECTOR
Dim Vec1 As D3DVECTOR, Vec2 As D3DVECTOR
Dim NormV As D3DVECTOR
Dim MP As Integer, A As Integer
Dim Pt(1 To 3) As Integer

' Grab # of points in current polygon for reference.
MP = UBound(Lev.Polys(Index).PtData)

' Triangles are always coplanar.
If MP < 4 Then
  py_isCoplanar = 1
  Exit Function
End If

Pt(1) = Lev.Polys(Index).PtData(1).Index

' Call py_getNormalV with no scalar to sum all trifan normals.
'
Vec2 = py_getNormalV(Lev, Index, 0, 0)

' Calculate the return value as a percentage^2.
'
' Example using a 4-sided polygon (2 trifans):
' - If the polygon is flat, the trifan normals will all be paralell and:
'     v_Magna(vector sum) = 2, result=1.00
' - If the polygon bends by 90-deg, the trifan normals will add up to:
'     v_Magna(vector sum) = 1.414, result=(1.414/2)^2 = (0.7)^2 = 0.5
' - If the polygon bends by 180-degrees, the trifan normals will add up to:
'     v_Magna(vector sum) = 0, result=0.00
'
py_isCoplanar = v_Magna(Vec2, 2) / (MP - 2) ^ 2

End Function

Public Function v_DoMatrix(ByRef Src As D3DVECTOR, ByRef Mat As D3DMATRIX) As D3DVECTOR
'
' Transforms a D3DVECTOR using a D3DMATRIX.
'
' (The fourth column of the matrix is ignored)
'
v_DoMatrix.X = Src.X * Mat.rc11 + Src.Y * Mat.rc21 + Src.Z * Mat.rc31 + Mat.rc41
v_DoMatrix.Y = Src.X * Mat.rc12 + Src.Y * Mat.rc22 + Src.Z * Mat.rc32 + Mat.rc42
v_DoMatrix.Z = Src.X * Mat.rc13 + Src.Y * Mat.rc23 + Src.Z * Mat.rc33 + Mat.rc43

End Function

Public Function pt_RemoveDuplis(ByRef Lev As LevelType, ByVal SrcPt As Integer) As Integer
'
' Scans a submodel for duplicate points sharing the same coordinates and
' consolidates (merges) them.  This will also attach any polygons using
' those points as well.
'
' Of course, if joining two points causes polygons from different
' submodels to become attached, Pie Slicer will automatically separate
' them when saving PIE data to disk.
'
Syntax:
'
' pt_RemoveDuplis(Lev, SrcPt)
'   Lev   - Submodel to check.
'   SrcPt - Point index to check.
'           * To check ALL points in the submodel, specify -1.
'
' Function returns the # of duplicate points eliminated,
' 0 for none, or -1 if error.
'

Dim FoundP() As Integer
Dim Min As Integer, Max As Integer

If SrcPt > 0 Then
  ' Check a single point & remove (merge) any duplicates of it.
  '
  
  ' FoundP() array will hold a list of all higher-index duplicate points.
  ' FoundP(n) -- Index of a point sharing the same location as SrcPt
  '
  Dim NumFound As Integer
  ReDim FoundP(0)
  NumFound = 0
    
  ' Compare this point to all other points in the submodel
  '
  Dim Pt As Integer, Pt1 As D3DVECTOR, Pt2 As D3DVECTOR
  Pt2 = Lev.Points(SrcPt).Loc
  
  For Pt = 1 To UBound(Lev.Points)
    If Pt <> SrcPt Then
      Pt1 = Lev.Points(Pt).Loc
      If Int(Pt1.X + 0.5) = Int(Pt2.X + 0.5) Then
      If Int(Pt1.Y + 0.5) = Int(Pt2.Y + 0.5) Then
      If Int(Pt1.Z + 0.5) = Int(Pt2.Z + 0.5) Then
        
        ' Duplicate point found.  Record its index in an array
        NumFound = NumFound + 1
        ReDim Preserve FoundP(NumFound) As Integer
        FoundP(NumFound) = Pt
      
      End If: End If: End If
    End If
  Next Pt
  
  If NumFound < 1 Then
    pt_RemoveDuplis = 0
    Exit Function
  End If
  
  ' Ask for user confirmation.
  Dim Mesg As String, T As Integer
  Mesg$ = "Would you like to merge point #" & SrcPt & " with"
  For T = 1 To NumFound
    Mesg$ = Mesg$ & IIf(NumFound > 1, vbCrLf & "  ", " ") & "point #" & FoundP(T)
  Next T
  Mesg$ = Mesg$ & " ?" & vbCrLf & vbCrLf & "All polygons using these points will become" & vbCrLf & "attached to the current point."
  If MsgBox(Mesg$, vbQuestion + vbYesNo, "PIE Slicer") = vbNo Then
    pt_RemoveDuplis = -1
    Exit Function
  End If
  
  ' Begin updating all polygons in the model
  Dim Py As Integer
  For Py = 1 To UBound(Lev.Polys)
  With Lev.Polys(Py)
    
    ' Iterate through each polygon's vertexes
    For Pt = 1 To UBound(.PtData)
    With .PtData(Pt)
      
      ' Update any old vertex references to reflect the new point
      Dim C As Integer
      For C = 1 To UBound(FoundP)
        If .Index = FoundP(C) Then
          .Index = SrcPt
        End If
      Next C
    
    End With
    Next Pt
  
  End With
  Next Py
     
  ' Remove the duplicate points in DESCENDING ORDER (critical!)
  ' NOTE:  Update the points list manually when routine is finished
  For C = UBound(FoundP) To 1 Step -1
    pt_Remove Lev, FoundP(C), True
  Next C
  
  pt_RemoveDuplis = NumFound

Else
  
  ' Find & remove all duplicate points in the entire submodel!
  '
  
  ' This time, each element in the FoundP() array will correspond
  ' to each point in the submodel.  We'll use it to keep track of
  ' which points are unique and which ones are duplicitous.
  '
  ' FoundP(n) = 0  Point #n is unique, or the lowest-index duplicate.
  '                Therefore, keep it.
  '
  ' FoundP(n) > 0  Point #n is a higher-index duplicate and can be removed.
  '                The exact value says which lower-index point it matches.
  '
  
  ' Initialize the FoundP() array to mirror the Points() array
  Max = UBound(Lev.Points)
  ReDim FoundP(Max)
  
  Dim NumBasePts As Integer
  
  ' Iterate through all the points
  '
  Dim TestPt As Integer
  For TestPt = 1 To Max - 1
    ' Test a given point only if it is NOT a
    ' duplicate of a lower-index point (FoundP(n) returns zero)
    '
    If FoundP(TestPt) = 0 Then
      
      Pt2 = Lev.Points(TestPt).Loc
      
      ' Check all higher-index points
      '
      Dim FirstPt As Boolean
      FirstPt = True
      For Pt = TestPt + 1 To Max
        Pt1 = Lev.Points(Pt).Loc
        If Int(Pt1.X + 0.5) = Int(Pt2.X + 0.5) Then
        If Int(Pt1.Y + 0.5) = Int(Pt2.Y + 0.5) Then
        If Int(Pt1.Z + 0.5) = Int(Pt2.Z + 0.5) Then
          ' Duplicate point found.
          ' Record that Pt is a duplicate of TestPt
          '
          If FirstPt Then
            NumBasePts = NumBasePts + 1
            FirstPt = False
          End If
          
          FoundP(Pt) = TestPt
          
          ' Keep track of how many duplicates have been found.
          NumFound = NumFound + 1
        End If: End If: End If
      Next Pt
    End If
  Next TestPt
  
  If NumFound < 1 Then
    pt_RemoveDuplis = 0
    Exit Function
  End If
  
  ' Ask for user confirmation.
  Mesg$ = "Found " & NumFound & " duplicate points in this model.  Would you like to combine them?"
  Mesg$ = Mesg$ & vbCrLf & vbCrLf & "All polygons using these points will become connected."
  If MsgBox(Mesg$, vbQuestion + vbYesNo, "PIE Slicer") = vbNo Then
    pt_RemoveDuplis = -1
    Exit Function
  End If
  
  ' Now, begin removing the points from the array,
  ' in Descending Order.
  '
  For TestPt = Max To 1 Step -1
    
    If FoundP(TestPt) > 0 Then
      
      For Py = 1 To UBound(Lev.Polys)
      With Lev.Polys(Py)
        
        For Pt = 1 To UBound(.PtData)
        With .PtData(Pt)
          
          ' Update all polygons, using the array of duplicate pt indexes
          '
          If .Index = TestPt Then
            .Index = FoundP(TestPt)
          End If
          
        End With
        Next Pt
        
      End With
      Next Py
      
      ' Remove the duplicate point & update all polygon references
      pt_Remove Model.Files(ThisFile).Levs, TestPt, True
    
    End If
  
  Next TestPt

  pt_RemoveDuplis = NumFound
End If

End Function

Public Sub SetCamPoint(ByVal How As Integer, RedrawCode As Long)
' Sets the camera point to a certain location in the model.

Select Case How
Case CamPt_Point
  If ThisPoint < 1 Then Exit Sub
  
  CamCentre = Model.Files(ThisFile).Levs.Points(ThisPoint).Loc
  formMaster.SetStatus 0, "Focused the camera on point #" & ThisPoint, FormMain.statBar

Case CamPt_Poly
  If ThisPoly < 1 Then Exit Sub
  
  CamCentre = pt_CenterOf(PTC_CurPy, Model.Files(ThisFile).Levs, ThisPoly, 0)
  formMaster.SetStatus 0, "Focused the camera on polygon #" & ThisPoly, FormMain.statBar

Case CamPt_None
  CamCentre.X = 0
  CamCentre.Y = 0
  CamCentre.Z = 0
  formMaster.SetStatus 0, "Reset the camera to (0,0,0)", FormMain.statBar

Case Else
  Debug.Assert How = -1
  Exit Sub
End Select

If RedrawCode > 0 Then FormMain.Redraw RedrawCode

End Sub


Public Sub ListPolyData(Polygon As PolygonData, ByVal Item As ListItem)
' Updates the subitem data for a polygon's listbox entry.

Item.SubItems(1) = Polygon.Submodel
Item.SubItems(2) = Hex$(Polygon.Flags)
Item.SubItems(3) = UBound(Polygon.PtData)
Item.SubItems(4) = IIf(Polygon.Flags And &H2000&, "=", "")
Item.SubItems(5) = IIf(Polygon.Flags And &H4000&, Polygon.TeamC.Animated, "")

SetItemStat Item, Polygon.StatFlags

End Sub


Public Function file_LoadData(ByVal FileName As String, Data As PieModel, ByVal Dialog As CommonDialog) As Long
' Determines what kind of file is requested, then loads its data.

' If successful, returns the # of PIE file structures available in the Model() array.

Dim Temp As PieModel

If FileName = "" Or Dir(FileName) = "" Then
  file_LoadData = FE_LoadFile Or FE_FileNotFound
  Exit Function
End If

ReDim Temp.Files(0)
ReDim Temp.Tex_Files(0)

Dim TypeFlag As Long
TypeFlag = file_GetType(FileName)

Select Case TypeFlag
Case Format_Unable, Format_Invalid
  file_LoadData = FE_LoadFile Or FE_BadFormat
  Exit Function
  
Case Format_PckFile
  file_LoadPCK FileName, Temp
  
Case Format_PieFile, Format_CakFile

  ReDim Temp.Files(1)
  ReDim Temp.Tex_Files(0)
  
  Dim Code As Long
  Select Case TypeFlag
  Case Format_PieFile
    Code = file_LoadPIE(FileName, Temp.Files(1))
  Case Format_CakFile
    Code = file_LoadCak(Dialog.FileName, Temp.Files(1))
  End Select
  
  If Code < 1 Then
    file_LoadData = Code
    Exit Function
  End If
  
Case Format_AniFile
  Dim AnimData As AnimData
  Dim PieFileName As String
  
  Code = file_LoadANI(FileName, Temp, Dialog)
  
  If Code < 1 Then
    file_LoadData = Code
    Exit Function
  End If
  
Case Else
  Debug.Assert False
  
End Select

vid_LoadPieTextures Temp

Data = Temp

file_LoadData = UBound(Temp.Files)

End Function

Public Function GetMaxiFrames(ByVal File As Integer, Optional SetIt As Boolean = True)

Select Case Model.Files(File).Anim.AnimType
Case ANIM3DFRAMES
  GetMaxiFrames = enumMaxLevels(Model.Files(File).Levs)
  If SetIt Then Model.Files(File).Anim.MaxFrames = GetMaxiFrames
Case ANIM3DTRANS
  GetMaxiFrames = Model.Files(File).Anim.MaxFrames
Case Else
  ' GetMaxiFrames = 0
End Select

End Function

Public Sub getCamera(File As PieType, Cam As D3DVECTOR, ByRef ZStep As Integer)
'
' Computes camera settings based on PIE file data.
' (i.e. position and zoom level)
'
Syntax:
'
' getCamera(File, Cam, ZStep)
'   File  - PIE file to evaluate.
'   Cam   - (output) Camera coordinates to return.
'   ZStep - (output) Zoom level to view it with.
'

' Measure the size of the PIE file.
' Then calculate and return the camera position, centered about it.
'
Dim Mins As D3DVECTOR, Maxs As D3DVECTOR
getPieSizeBox File.Levs, Mins, Maxs

Cam.X = (Mins.X + Maxs.X) \ 2
Cam.Y = (Mins.Y + Maxs.Y) \ 2
Cam.Z = (Mins.Z + Maxs.Z) \ 2

' Calculate and return a zoom factor appropriate for the model.
'
Dim Size1 As Integer, PieSize As Integer
PieSize = Abs(Mins.X - Maxs.X)
Size1 = Abs(Mins.Y - Maxs.Y)
If Size1 > PieSize Then PieSize = Size1
Size1 = Abs(Mins.Z - Maxs.Z)
If Size1 > PieSize Then PieSize = Size1
  
' Fetch maximum display size
Dim Box As PictureBox, X As Integer, Y As Integer
Set Box = FormMain.pic(2)
X = Box.ScaleWidth: Y = Box.ScaleHeight
If X > Y Then Size1 = X \ 2 Else Size1 = Y \ 2
  
' Calculate ZStep according to a dB scale.
'
ZStep = -Int(20 * Log((PieSize + 1) / Size1) / m_Log10)
If ZStep > 15 Then ZStep = 15
If ZStep < -25 Then ZStep = -25
  
End Sub

Public Function file_GetType(ByVal FileName As String) As Long
'
' Analyzes a data file to see if Pie Slicer supports it.
'

If Dir(FileName) = "" Then
  file_GetType = Format_Unable
  Exit Function
End If

Dim FileNum As Integer
FileNum = FreeFile
Open FileName For Binary Access Read Lock Write As #FileNum
  
Dim Lin As Integer
Do
  Dim HeaderLine As String
  Lin = Lin + 1
  Line Input #FileNum, HeaderLine
  
  ' Pie/PCK file test: first line MUST be the header.
  If Lin = 1 Then
    If Left$(HeaderLine, 5) = "PIE 2" Then
      Close #FileNum
      
      file_GetType = Format_PieFile
      Exit Function
      
    ElseIf Left$(HeaderLine, 4) = "CAKE" Then
      Close #FileNum
      
      file_GetType = Format_CakFile
      Exit Function
      
    ElseIf Left$(HeaderLine, 5) = "PieDX" Then
      Close #FileNum
      
      file_GetType = Format_PckFile
      Exit Function
    End If
  
  ' ANI file test: must have an ANIM3D tag.
  ElseIf InStr(UCase$(HeaderLine), "ANIM3D") Then
    Close #FileNum
    file_GetType = Format_AniFile
    
    Exit Function
    
  ' Support for ANI file test.  ANI header must be within the first 5 lines of file
  ' though it's not always on the first line.
  '
  ' Any preceding lines must either be comments or blank
  '
  ' Instr returns 2 = /* comment
  '               1 = empty line or // comment
  '               0 = something else (not allowed)
  ElseIf InStr("//*", Left$(HeaderLine, 2)) = 0 Then
    Close #FileNum
    file_GetType = Format_Invalid
    
    Debug.Assert False
    Exit Function
  End If
Loop While Lin < 5
  
' Unable to find a suitable header in first 5 lines: not a valid file.
Close #FileNum
Debug.Assert False
file_GetType = Format_Invalid

End Function

Public Function py_BuildExtrude(ByRef Source As LevelType, ByVal Record As Integer, Offset As Integer, Optional PasteData As Boolean = False) As LevelType
Dim Temp As LevelType
ReDim Temp.Points(0)
ReDim Temp.Polys(0)
'
' Builds an extrusion from the current polygon.
'
Syntax:
'
' py_BuildExtrude(source, poly, offset, [paste] )
'   source -- Submodel for reference.
'   poly   -- Polygon to extrude off of.
'   offset -- Height (distance) of the extrusion.
'   paste  -- False (default) - Return an extrusion submodel
'             True - Paste extrusion data into source submodel
'
py_BuildExtrude = Temp
If Offset = 0 Then Exit Function

' Calculate offset, based on parameters
Dim Vector As D3DVECTOR
With py_getNormalV(Source, Record, , 1)
  Vector.X = .X * Offset
  Vector.Y = .Y * Offset
  Vector.Z = .Z * Offset
End With

With Source.Polys(Record)
  Dim Num As Integer
  Num = UBound(.PtData)
  ReDim Temp.Points(2 * Num)
  ReDim Temp.Polys(1 + Num)

  ' Copy the current polygon and its points
  Temp.Polys(1) = Source.Polys(Record)
  Temp.Polys(1).StatFlags = 0
  Dim A As Integer, Pt As Integer
  
  Pt = 0
  For A = 1 To Num
    Pt = Pt + 1
    
    ' Pt records offset (new) points
    Temp.Points(Pt) = Source.Points(.PtData(A).Index)
    
    ' Pt+Num records original points from current poly (Status=-1 for clarity)
    Temp.Points(Pt + Num) = Temp.Points(Pt)
    Temp.Points(Pt + Num).StatFlags = -1
    
    With Temp.Points(Pt).Loc
      .X = .X + Vector.X
      .Y = .Y + Vector.Y
      .Z = .Z + Vector.Z
    End With
    
    
    ' Change the base (offset) polygon's vertex
    Temp.Polys(1).PtData(A).Index = Pt
    
    ' Create side polys (4 sides apiece)
    Dim A2 As Integer
    A2 = IIf(A = Num, 1, A + 1)
    With Temp.Polys(A + 1)
      Temp.Polys(A + 1).Flags = Source.Polys(1).Flags
      Temp.Polys(A + 1).TeamC = Source.Polys(1).TeamC
      
      ReDim .PtData(4)
      If PasteData Then
        Debug.Assert A <= Num
        Debug.Assert A2 <= Num
        .PtData(1).Index = A2
        .PtData(2).Index = A
        .PtData(3).Index = -A
        .PtData(4).Index = -A2
      Else
        .PtData(1).Index = A2
        .PtData(2).Index = A
        .PtData(3).Index = A + Num
        .PtData(4).Index = A2 + Num
      End If
    
      ' Prepare some texture information for the new polys
      .PtData(1).Loc = Temp.Polys(1).PtData(A2).Loc
      .PtData(2).Loc = Temp.Polys(1).PtData(A).Loc
      .PtData(3).Loc = Temp.Polys(1).PtData(A).Loc
      .PtData(4).Loc = Temp.Polys(1).PtData(A2).Loc
    
    End With
  Next A
End With

If PasteData Then
  ' paste point data
  Dim MaxPts As Integer, maxPys As Integer
  MaxPts = UBound(Source.Points)
  ReDim Preserve Source.Points(MaxPts + Num)
  For A = 1 To Num
    Debug.Assert Temp.Points(A).StatFlags = 0
    Source.Points(MaxPts + A) = Temp.Points(A)
  Next A
  
  ' paste poly data
  maxPys = UBound(Source.Polys)
  ReDim Preserve Source.Polys(maxPys + Num + 1)
  For A = 1 To UBound(Temp.Polys)
    Source.Polys(maxPys + A) = Temp.Polys(A)
    
    With Source.Polys(maxPys + A)
    Dim B As Integer
    For B = 1 To UBound(.PtData)
      Pt = .PtData(B).Index
      Debug.Assert Pt <= Num
      If Pt < 0 Then ' Negative value means use this vertex of current poly
        .PtData(B).Index = Source.Polys(Record).PtData(-Pt).Index
      Else
        .PtData(B).Index = Pt + MaxPts ' Adjust for existing points
      End If
    Next B
    End With
  Next A
Else
  py_BuildExtrude = Temp
End If

End Function


Public Sub SaveWindowSize(ByRef Source As Form, ByVal Index As Integer)
' Records a window's dimensions.
If Source.WindowState = 1 Then Exit Sub

With Sys.WindowPos(Index)
  .X1 = Source.Left
  .Y1 = Source.Top
  .X2 = Source.Width
  .Y2 = Source.Height
End With

End Sub

Public Function file_GetAttrs(ByVal FileName As String) As Long
'
' Performs a combination file-exists and attributes check.
'
' Commonly used when saving to test if an existing file is read only.
'

If FileName = "" Then Exit Function
If Dir(FileName) = "" Then Exit Function

file_GetAttrs = GetAttr(FileName)

End Function


Public Function file_CreateTempName(ByRef FileName As String) As Long
'
' Creates a name for a tempfile when saving.
'
Syntax:
'
' OpenTempFile(FileName)
'   FileName - (output) Name of temporary file to use.
'
' No return value.

Dim N As Long
N = &H1000000 * Rnd
FileName = App.Path & "\$" & Hex$(N) & ".tmp"

End Function

Public Function vid_LoadPieTextures(Data As PieModel) As Long
'
' Loads the texture surfaces specified in a PCK model file.
'


' Clear the vidmem first
ReDim Sys.VidMem(0)

Dim T As Integer
Dim FileName As String, TexPage As Integer
For T = 1 To UBound(Data.Tex_Files)
  
  ' Identify which texture this is.
  FileName = Data.Tex_Files(T)
  TexPage = GetTexPageNum(FileName, "")
  
  ' See if this texpage has been loaded into vidmem already or not
  '
  If TexPage > 0 Then
  If vid_CheckForPage(TexPage, False) = 0 Then
    
    Dim TexFiles() As String, NumAvail As Integer
    Dim DefFiles() As String, NumDefs As Integer
    Dim FileN As Integer, Matched As Integer
    
    ' Check custom textures first
    '
    NumAvail = GetAvailTextures(TexPage, False, TexFiles())
    If NumAvail > 0 Then
      For FileN = 1 To NumAvail
        If LCase$(FileN) = LCase$(App.Path & "\TexData\" & FileName) Then
          Matched = 3 ' Exact match
          FileName = TexFiles(FileN)
          Exit For
        ElseIf GetTexPageNum(FileN, "") = TexPage Then
          Matched = 1 ' Partial match with custom texture
          FileName = TexFiles(FileN)
        End If
      Next FileN
    End If
    
    ' Then check default textures
    '
    NumAvail = GetAvailTextures(TexPage, True, TexFiles())
    If NumAvail > 0 Then
      For FileN = 1 To NumAvail
        If LCase$(FileN) = LCase$(App.Path & "\" & FileName) Then
          Matched = 3 ' Exact match for default texture
          FileName = TexFiles(FileN)
          Exit For
        ElseIf GetTexPageNum(FileN, "") = TexPage Then
          Matched = 2 ' Partial match with default texture
          FileName = TexFiles(FileN)
          Exit For
        End If
      Next FileN
    End If
    
    ' Found our texture to use?
    '
    If FileName <> "" Then
      ' Load the texture into vidmem
      '
      'Debug.Assert False
      TexPage = UBound(Sys.VidMem) + 1
      ReDim Preserve Sys.VidMem(TexPage)
      
      Dim Ax As Integer
      For Ax = 1 To MAX_TexMaps
        If VidMem(Ax) Is Nothing Then
          Sys.VidMem(TexPage).SurfPtr = Ax
          ChangeTexture Sys.VidMem(TexPage), FileName
          Exit For
        End If
      Next Ax
    Else
      ' Couldn't find texture?
      '
      Dim Warned As Integer
      If Not Warned Then MsgBox "Unable to load all the textures for this PIE model.", vbInformation
      Warned = True
    End If
    
  End If
  End If
  
Next T

End Function


Public Function GetAvailTextures(ByVal TexPage As Integer, ByVal GetDefaults As Boolean, Results() As String) As Long
Dim FileNam As String, ResNum As Integer, N1 As Long
'
' Locates available Pie Slicer textures for a specific page ##.
'
Syntax:
'
' GetAvailTextures(TexPage, GetDefaults, Results())
'
' TexPage     -- Texturepage ## to search for.
' GetDefaults -- Whether to look for default or custom textures.
' Results()   -- String array to store the results in.
'
' Function returns the # of textures found
' and an array of filenames to refer to them by.
'

ReDim Results(0)

Select Case GetDefaults
Case True
  '
  ' Default textures:  Pie Slicer directory, "page(##).bmp"
  '
  FileNam$ = Dir(App.Path & "\page(" & TexPage & ")*.pcx")
  Do Until FileNam$ = ""
    N1& = InStr(FileNam$, ")")
    
    ResNum = UBound(Results) + 1
    ReDim Preserve Results(ResNum)
    Results(ResNum) = FileNam
    
    FileNam = Dir()
  Loop

Case False
  '
  ' Custom textures: Pie Slicer\TexData directory, WZ naming convention
  '
  
  Dim BaseDir As String, TexDir As String
  BaseDir = IIf(Right$(App.Path, 1) = "\", App.Path, App.Path & "\")
  TexDir = BaseDir & "TexData\"
  
'  If User.Settings And Set_PCXMask Then
    FileNam$ = Dir(TexDir & "page-" & TexPage & "*.*")
'  Else
'    FileNam$ = Dir(TexDir & "page-" & TexPage & "*.bmp")
'  End If
  
  Dim Ext As String * 4
  Ext = LCase$(Right$(FileNam, 4))
  
  Do Until FileNam$ = ""
  If Ext = ".pcx" Or Ext = ".bmp" Then
    
    If GetTexPageNum(FileNam$, "") = TexPage Then ' Must have, because of wildcards
      ResNum = UBound(Results) + 1
      ReDim Preserve Results(ResNum)
      Results(ResNum) = FileNam
    End If
    
  End If
  FileNam$ = Dir()
  Loop

End Select

GetAvailTextures = ResNum

End Function

Public Function vid_CheckForPage(ByVal Page As Integer, Optional AddIt As Boolean = False) As Long
'
' Vidmem array management 1 - checks to see if a vidmem page # is loaded
'
Syntax:
'
' AddVidmemPage(Page, AddIt)
'   Page  - Page ## for the texture as per texture's filename
'   AddIt - TRUE = creates a new entry for the texpage
'
' Function returns the vidmem index corresponding to the texturepage,
' or 0 if no matching texpage was found.
'

' N/A for page 0 (NoTexture)
If Page = 0 Then Exit Function

' Check to see if the page # is already loaded
'
Dim Num As Integer, VidPage As Integer
For Num = 1 To UBound(Sys.VidMem)
  If Sys.VidMem(Num).Page = Page Then
    vid_CheckForPage = Num
    Exit Function
  End If
Next Num

' If not, make array space for a new page #
'
If AddIt Then
  Num = UBound(Sys.VidMem) + 1
  ReDim Preserve Sys.VidMem(Num)
  Sys.VidMem(Num).Page = Page
  
  Dim Ax As Integer
  For Ax = 1 To MAX_TexMaps
    If VidMem(Ax) Is Nothing Then
      Sys.VidMem(Num).SurfPtr = Ax
      Exit For
    End If
  Next Ax
  
  vid_CheckForPage = Num
Else
  vid_CheckForPage = 0
End If

End Function

Public Function vid_CleanupPages() As Long
'
' Vidmem array management 2 - Unload textures not needed anymore.
'
' Not intended for use during Pie Slicer rendering.
'
Syntax:
'
' vid_CleanupPages()
'

' Begin iterating through the vidmem pages
'
Dim Ax As Integer, Bx As Integer
Dim Surf As Integer

Dim Max As Integer
Max = UBound(Sys.VidMem)

Ax = 1
Do Until Ax > UBound(Sys.VidMem)
  
  ' Iterate through the current model & test if the vidmem page is used
  Dim Clr As Integer
  For Clr = 1 To UBound(Model.Files)
    If Model.Files(Clr).TexFile = Ax Then
      Clr = 0
      Exit For
    End If
  Next Clr

  ' Unload the texture if it's not needed anymore
  '
  If Clr Then
  
    Surf = Sys.VidMem(Ax).SurfPtr
    Set VidMem(Surf) = Nothing
    
    Clr = UBound(Sys.VidMem) - 1
    For Bx = Ax To Clr
      Sys.VidMem(Bx) = Sys.VidMem(Bx + 1)
    Next Bx
    ReDim Preserve Sys.VidMem(Clr)
    
    ' Update the files' texinfo because the vidmem has changed
    For Bx = 1 To UBound(Model.Files)
      With Model.Files(Bx)
        If .TexFile > Ax Then .TexFile = .TexFile - 1
      End With
    Next Bx
    
  Else
    Ax = Ax + 1
  End If
Loop

End Function

Public Function DistanceLine(Base As D3DVECTOR, Normal As D3DVECTOR, Point As D3DVECTOR) As Single

' Textbook math equation for computing linear distance between a point and a line.
' D = mag((P-B) x D) / mag(D)

' Note the use of squared magnitude checks.
' Reduces # of flops by 1 sqr() call.
DistanceLine = Sqr(v_Magna(v_CrossP(v_Subtract(Point, Base), Normal), 2) / v_Magna(Normal, 2))

End Function


Public Function SetupToolbar(ByRef Ctrl As Toolbar, Optional Tag As String = "*")
'
' Common setup function - initialize a toolbar and its icons
'
Ctrl.Style = 1

Set Ctrl.ImageList = formMaster.lstTools

Dim Item As Button
For Each Item In Ctrl.buttons
  If Item.Tag = "" Then
    Item.Image = 0
  Else
    Item.Image = Item.Tag + 1
    formMaster.lstTools.ListImages(Item.Image).Tag = Tag
  End If
Next Item

'Ctrl.Height = formMaster.picTools.ClipHeight + 10
'Ctrl.ButtonHeight = formMaster.picTools.ClipHeight + 7
'Ctrl.ButtonWidth = formMaster.picTools.ClipWidth + 6

End Function

Public Function sel_Scale(ByRef Data As LevelType, Scalar As Single, ByVal Flags As Long) As Integer
Dim Offset As D3DVECTOR
'
' Performs linear scaling of selected data.
'
Syntax:
'
' sel_Scale(Data, Scalar, Flags)
'
'   Data   - Submodel of interest
'   Scalar - Scalar to multiply with.
'   Flags  - See "ScaRotaFlags" enum for details


' Scalar=1 has no effect
If Scalar = 1 Then sel_Scale = 0: Exit Function

' Axial flag takes priority over others
'
If Flags And SCR_DoAxial Then
  
  ' Check for a proper scaling axis
  '
  Dim AxisPt As D3DVECTOR, AxisRay As D3DVECTOR
  With AxisRay
    .X = Sys.PtAxis.X - Sys.PtCenter.X
    .Y = Sys.PtAxis.Y - Sys.PtCenter.Y
    .Z = Sys.PtAxis.Z - Sys.PtCenter.Z
  End With
  
  ' Analyze the scaling axis
  '
  Dim TempF As Integer
  If AxisRay.X = 0 Then TempF = TempF Or 1
  If AxisRay.Y = 0 Then TempF = TempF Or 2
  If AxisRay.Z = 0 Then TempF = TempF Or 4
  
  Select Case TempF
  Case 7
     ' Invalid axis? Then do all-axis scaling about that point.
     Flags = (Flags And SCR_DataMask) Or SCR_AxisMask Or SCR_UseCenterPt
  
  Case 6
    ' =X-axis - scale Y & Z axes only
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisY Or SCR_DoAxisZ And SCR_UseCenterPt
  
  Case 5
    ' =Y-axis - scale X & Z axes only
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisX Or SCR_DoAxisZ And SCR_UseCenterPt
  
  Case 3
    ' =Z-axis - scale X & Y axes only
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisX Or SCR_DoAxisY And SCR_UseCenterPt
  End Select

End If

' Linear scaling - adjust for center point if requested
'
Dim OffsetPt As D3DVECTOR
If Flags And SCR_UseCenterPt Then
  With OffsetPt
    .X = Sys.PtCenter.X * (1 - Scalar)
    .Y = Sys.PtCenter.Y * (1 - Scalar)
    .Z = Sys.PtCenter.Z * (1 - Scalar)
  End With
End If

  
' Enum the relevant point data
'
Dim PointsList() As Integer, DoScale As Integer
DoScale = sel_GetPointList(Data, Flags, PointsList())

If DoScale < 1 Then
  sel_Scale = 0
  Exit Function
End If
  
' Begin scaling the selected points
'
Dim Pt As Integer, SrcPt As D3DVECTOR

For Pt = 1 To UBound(PointsList)
With Data.Points(PointsList(Pt))
    
  ' Perform axial scaling if requested
  '
  If Flags And SCR_DoAxial Then
  
    SrcPt = .Loc
      GoSub DoAxialMath
    .Loc = SrcPt
  
  ' Otherwise, do linear scaling
  '
  Else
    If Flags And SCR_DoAxisX Then .Loc.X = .Loc.X * Scalar + OffsetPt.X
    If Flags And SCR_DoAxisY Then .Loc.Y = .Loc.Y * Scalar + OffsetPt.Y
    If Flags And SCR_DoAxisZ Then .Loc.Z = .Loc.Z * Scalar + OffsetPt.Z
  End If
  
  ' Count how many points have been processed
  '
  sel_Scale = sel_Scale + 1

End With
Next Pt

If Flags And SCR_DoCcs Then
  
  For Pt = 1 To UBound(Data.Ccs)
  With Data.Ccs(Pt)
  If (.StatFlags And stat_SelMask) = stat_Sel Then
  
    ' Perform axial scaling if requested
    '
    If Flags And SCR_DoAxial Then
    
      SrcPt = .Loc
        GoSub DoAxialMath
      .Loc = SrcPt
    
    ' Otherwise, do linear scaling
    '
    Else
      If Flags And SCR_DoAxisX Then .Loc.X = .Loc.X * Scalar + OffsetPt.X
      If Flags And SCR_DoAxisY Then .Loc.Y = .Loc.Y * Scalar + OffsetPt.Y
      If Flags And SCR_DoAxisZ Then .Loc.Z = .Loc.Z * Scalar + OffsetPt.Z
    End If
    
    ' Count how many items have been processed
    '
    sel_Scale = sel_Scale + 1
  
  End If
  End With
  Next Pt

End If


' Done!
Exit Function

' Sub-function: Perform axial scaling on SrcPt.
'
DoAxialMath:
' Basic formula:
' A,B - bases for axis vector
' C   - point to scale
'
' D = proj(C, AB)
' S = (AC-D)*Scalar
' C2 = A+D+S
'
  ' Vector AC (point, relative to axis base)
  Dim RayPt As D3DVECTOR
  RayPt.X = SrcPt.X - Sys.PtCenter.X
  RayPt.Y = SrcPt.Y - Sys.PtCenter.Y
  RayPt.Z = SrcPt.Z - Sys.PtCenter.Z
  
  ' Compute scaled vector
  ' = (AC-D)*Scalar + D + A
  With v_Project(RayPt, AxisRay, 0) ' Vector D
    SrcPt.X = (RayPt.X - .X) * Scalar + .X + Sys.PtCenter.X
    SrcPt.Y = (RayPt.Y - .Y) * Scalar + .Y + Sys.PtCenter.Y
    SrcPt.Z = (RayPt.Z - .Z) * Scalar + .Z + Sys.PtCenter.Z
  End With
Return

End Function

Public Function sel_GetPointList(ByRef Data As LevelType, ByVal Flags As Long, ByRef Results() As Integer) As Long
'
' Scans through pie data and returns a list of all selected points.
'
Syntax:
'
' sel_GetPointList(Data, Flags, Results)
'
'   Data   - PIE data to check
'   Flags  - Flags indicating data to check.
'               +SCR_DoPoints = Return selected points
'               +SCR_DoPolys  = Return vertexes for selected polys
'                                 * The polys themselves are not counted
'               +SCR_DoCcs    = Count selected cc's
'
'   Results - Array to store the findings in.
'               Results(0) = Number of selected points found
'               Results(#) = index of a selected point
'
' Returns the total # of selected items found.  This won't always
' equal the number of selected POINTS found, so use Ubound(Results) to check that.
'

Dim Temp() As Integer
ReDim Temp(UBound(Data.Points))

ReDim Results(UBound(Data.Points))

If Flags And SCR_DoPolys Then
  ' Enumerate points attached to selected polygons.
  '
  Dim Pt As Integer, Py As Integer, Vt As Integer
  For Py = 1 To UBound(Data.Polys)
  With Data.Polys(Py)
  
    If .StatFlags = 1 Then
      ' Note affected points
      For Vt = 1 To UBound(.PtData)
        Pt = .PtData(Vt).Index
        
        ' Record new points only.
        If Not Temp(Pt) Then
          sel_GetPointList = sel_GetPointList + 1
          Results(sel_GetPointList) = Pt
          Temp(Pt) = True
        End If
      
      Next Vt
    End If
  
  End With
  Next Py
End If

If Flags And SCR_DoPoints Then
  ' Enumerate all selected points.
  '
  For Pt = 1 To UBound(Data.Points)
  If (Data.Points(Pt).StatFlags And stat_SelMask) = stat_Sel Then
    
    ' Record new points
    If Not Temp(Pt) Then
      sel_GetPointList = sel_GetPointList + 1
      Results(sel_GetPointList) = Pt
      Temp(Pt) = True
    End If
  
  End If
  Next Pt

End If

' Crop the results array to fit
ReDim Preserve Results(sel_GetPointList)
Results(0) = sel_GetPointList

If Flags And SCR_DoCcs Then
  ' Count selected Cc's.
  '
  For Pt = 1 To UBound(Data.Ccs)
    If Data.Ccs(Pt).StatFlags = 1 Then sel_GetPointList = sel_GetPointList + 1
  Next Pt
End If

End Function

Public Function sel_Rotate(ByRef Data As LevelType, ByVal Angle As Integer, ByVal Flags As Long)
'
' Performs rotation of selected data (points or polygons) in a PIE file.
'
Syntax:
'
' sel_Rotate(Data, Angle, Flags)
'   Data  - Submodel of interest
'   Angle - Positive (ccw) angle of rotation
'   Flags - Defines the behavior of this function
'             SCR_DoPoints    = Rotate selected points
'             SCR_DoPolys     = Rotate selected polygons
'             SCR_DoCcs       = Rotate selected cc's
'             SCR_DoAxisX     = Rotate X axis
'             SCR_DoAxisY     = Rotate Y axis
'             SCR_DoAxisZ     = Rotate Z axis
'             SCR_DoAxial     = Rotate about custom axis
'                                 (Sys.PtCenter & Sys.PtAxis)
'             SCR_UseCenterPt = Rotate relative to Sys.PtCenter
'                                 (N/A for SCR_DoAxial)
'           See enumeration (Shift+F2) for full details.
'
' Function returns the # of points rotated when successful.
' Returns 0 if nothing was rotated (i.e. nothing selected or angle=0)
' or -1 for bad calling syntax.


' Syntax checks first
'
If (Flags And SCR_DataMask) = 0 Then sel_Rotate = -1: Exit Function

If Angle = 0 Then sel_Rotate = 0: Exit Function

' Custom axis requested?
'
If Flags And SCR_DoAxial Then
  Dim AxisRay As D3DVECTOR
  With AxisRay
    .X = Sys.PtAxis.X - Sys.PtCenter.X
    .Y = Sys.PtAxis.Y - Sys.PtCenter.Y
    .Z = Sys.PtAxis.Z - Sys.PtCenter.Z
  End With

  ' Analyze the rotation axis
  '
  Dim TempF As Integer
  If AxisRay.X = 0 Then TempF = TempF Or 1
  If AxisRay.Y = 0 Then TempF = TempF Or 2
  If AxisRay.Z = 0 Then TempF = TempF Or 4
  
  Select Case TempF
  Case 7 ' Invalid axis
    sel_Rotate = -1
    Exit Function
  Case 6 ' =X-axis
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisX Or SCR_UseCenterPt
  Case 5 ' =Y-axis
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisY Or SCR_UseCenterPt
  Case 3 ' =Z-axis
    Flags = (Flags And SCR_DataMask) Or SCR_DoAxisZ Or SCR_UseCenterPt
  End Select
  
' Normal (orthogonal axis) rotation
'
Else
  
  ' Allow only one axis flag set at a time
  '
  Select Case Flags And SCR_AxisMask
  Case SCR_DoAxisX, SCR_DoAxisY, SCR_DoAxisZ
  
  Case Else ' Combined axes not allowed
    sel_Rotate = -1
    Exit Function
  End Select
End If
  
' Compute rotational scalars before we start looping
'
Dim CS As Single, SN As Single
CS = Cos(Angle * m_Deg2Rad)
SN = Sin(Angle * m_Deg2Rad)

' Compute offsets to accomodate center-point of rotation.
'
' Rotated points = (point - center)*rotation + center
'                = (point*rotation) + center - (center*rotation)
'                = (point*rotation) + (center - center*rotation)
'
' Offset vector  = center - (center*rotation)
'
Dim OffP As D3DVECTOR
With Sys.PtCenter
  Select Case Flags And SCR_AxisMask
  Case SCR_DoAxisX
    OffP.X = 0
    OffP.Y = .Y - (.Y * CS + .Z * SN)
    OffP.Z = .Z - (.Z * CS - .Y * SN)
  Case SCR_DoAxisY
    OffP.X = .X - (.X * CS - .Z * SN)
    OffP.Y = 0
    OffP.Z = .Z - (.Z * CS + .X * SN)
  Case SCR_DoAxisZ
    OffP.X = .X - (.X * CS - .Y * SN)
    OffP.Y = .Y - (.Y * CS + .X * SN)
    OffP.Z = 0
  End Select
End With

' Enum which points need rotating
'
Dim PointList() As Integer, Rota As Long
Rota = sel_GetPointList(Data, Flags, PointList())

' If no valid data, return to caller
If Rota < 1 Then
  sel_Rotate = Rota
  Exit Function
End If

' Rotate all affected points
Dim Pt As Integer, RotaP As D3DVECTOR, ResP As D3DVECTOR
For Pt = 1 To UBound(PointList)
With Data.Points(PointList(Pt))
      
  RotaP = .Loc
  
  If Flags And SCR_DoAxial Then
    
    ' Rotate about a custom axis using vector math.
    '
    ' More computation-heavy than orthogonal axes.
    '
    GoSub VectorMath
    .Loc = ResP
  
  Else
  
    ' Rotate about an orthogonal (coordinate) axis.
    '
  
    ' Perform rotational math from RotaP to ResP
    '
    Select Case (Flags And SCR_AxisMask)
    Case SCR_DoAxisX: GoSub XMath
    Case SCR_DoAxisY: GoSub YMath
    Case SCR_DoAxisZ: GoSub ZMath
    Case Else
      Debug.Assert False
      Exit Function
    End Select
  
    ' Use offset vector to allow for rotation around a point
    '
    If Flags And SCR_UseCenterPt Then
      With .Loc
      
        .X = ResP.X + OffP.X
        .Y = ResP.Y + OffP.Y
        .Z = ResP.Z + OffP.Z
      
      End With
    Else
    ' Or not
    '
      .Loc = ResP
    End If
  End If
    
  sel_Rotate = sel_Rotate + 1

End With
Next Pt

' Rotate affected cc's if requested
If Flags And SCR_DoCcs Then
  For Pt = 1 To UBound(Data.Ccs)
  With Data.Ccs(Pt)
  If (.StatFlags And stat_Sel) = stat_Sel Then
  
    RotaP = .Loc
  
    If Flags And SCR_DoAxial Then
      
      ' Rotate about a custom axis using vector math.
      '
      GoSub VectorMath
      .Loc = ResP
    
    Else
    
      ' Perform orthogonal rotation from RotaP to ResP
      ' (relative to 0,0,0 )
      '
      Select Case (Flags And SCR_AxisMask)
      Case SCR_DoAxisX: GoSub XMath
      Case SCR_DoAxisY: GoSub YMath
      Case SCR_DoAxisZ: GoSub ZMath
      End Select
    
      ' Apply offset vector if rotating about a point
      '
      If Flags And SCR_UseCenterPt Then
        With .Loc
        
          .X = ResP.X + OffP.X
          .Y = ResP.Y + OffP.Y
          .Z = ResP.Z + OffP.Z
        
        End With
      Else
        .Loc = ResP
      End If
    End If
      
    sel_Rotate = sel_Rotate + 1
  
  End If
  End With
  Next Pt

End If


' Done!
'
Exit Function


' Inline subfunctions
'


XMath: ' X axis rotation
  ResP.X = RotaP.X
  ResP.Y = RotaP.Y * CS + RotaP.Z * SN
  ResP.Z = RotaP.Z * CS - RotaP.Y * SN
Return

YMath: ' Y axis rotation
  ResP.X = RotaP.X * CS - RotaP.Z * SN
  ResP.Y = RotaP.Y
  ResP.Z = RotaP.Z * CS + RotaP.X * SN
Return

ZMath: ' Z axis rotation
  ResP.X = RotaP.X * CS - RotaP.Y * SN
  ResP.Y = RotaP.Y * CS + RotaP.X * SN
  ResP.Z = RotaP.Z
Return

VectorMath:
  ' Math for custom axis rotation.
  '
  
  ' RayPt = Source point relative to axis base
  '
  Dim RayPt As D3DVECTOR
  RayPt.X = RotaP.X - Sys.PtCenter.X
  RayPt.Y = RotaP.Y - Sys.PtCenter.Y
  RayPt.Z = RotaP.Z - Sys.PtCenter.Z
  
  ' AxRay = RayPt projected onto axis
  '
  Dim AxRay As D3DVECTOR
  AxRay = v_Project(RayPt, AxisRay, 0)
  

' A cute alteration to make it 'twist' instead of rotate
'  Dim Ax As Single
'  Ax = Sqr(v_Magna(AxRay, 2) / v_Magna(v_Subtract(Sys.PtCenter, Sys.PtAxis), 2))
'  CS = Cos(Angle * Ax * m_Deg2Rad)
'  SN = Sin(Angle * Ax * m_Deg2Rad)
  
  
  ' VecC (cosine vector) = Source point relative to axis
  '
  Dim VecC As D3DVECTOR
  VecC.X = RayPt.X - AxRay.X
  VecC.Y = RayPt.Y - AxRay.Y
  VecC.Z = RayPt.Z - AxRay.Z
  
  ' VecS (sine vector) = orthogonal to axis & VecC AND same length as VecC
  '
  Dim VecS As D3DVECTOR
  VecS = v_CrossP(AxisRay, VecC, v_Magna(VecC))
  
  ' Compute final (rotated) vector
  ' ResP = (VecC*Cos + VecS*Sine) + AxRay + (axis base)
  '
  ResP.X = VecC.X * CS + VecS.X * SN + AxRay.X + Sys.PtCenter.X
  ResP.Y = VecC.Y * CS + VecS.Y * SN + AxRay.Y + Sys.PtCenter.Y
  ResP.Z = VecC.Z * CS + VecS.Z * SN + AxRay.Z + Sys.PtCenter.Z
Return


End Function

Public Function v_ViewAngle(Optional ToCam As Integer = 0) As D3DVECTOR
'
' Returns a normal vector describing the current camera angle.
'
Syntax:
'
' v_ViewAngle( [ToCam] )
'
'    ToCam - 0 = Result vector points AWAY from the camera
'            1 = Result vector points TOWARDS the camera

Dim Rad As Single, H As Integer, V As Integer
Dim S1 As Single, C1 As Single
Dim S2 As Single, C2 As Single

' Calculate axes scalars
'
getViewAngles H, V, 0
Rad = H * m_Deg2Rad
S1 = -Sin(Rad)
C1 = Cos(Rad)

Rad = V * m_Deg2Rad
S2 = -Sin(Rad)
C2 = Cos(Rad)

' Orthogonal scalars for debug purposes
'C1 = 1: S1 = 0
'C2 = 1: S2 = 0

' Invert 3D rotation to determine the camera angle relative to coordinate axes
' (-x rotation then -y rotation)
Dim P As D3DVECTOR

' Default viewing angle: camera lies along positive z axis
P.X = 0
P.Y = 0
P.Z = 10

' First step: vertical angle
'  ResP.Y = TempP.Y * CS + TempP.z * SN
'  ResP.z = TempP.z * CS - TempP.Y * SN
' Second step: horizontal angle
'  ResP.X = TempP.X * CS - TempP.z * SN
'  ResP.z = TempP.z * CS + TempP.X * SN

With v_ViewAngle
  ' View-to-world rotation
  .X = P.X * C1 - (P.Z * C2 - P.Y * S2) * S1
             .Y = (P.Y * C2 + P.Z * S2)
  .Z = P.X * S1 + (P.Z * C2 - P.Y * S2) * C1
  
  If ToCam Then
    .X = -.X
    .Y = -.Y
    .Z = -.Z
  End If
End With

End Function

Public Sub SetUndoStep(Data As PieModel, ByVal Step As Long)
'
' Manages Pie Slicer's Undo system.
'
Syntax:
'
' SetUndoStep(File, Step, [Continuous])
'   File       - PIE file data to cache in the undo buffer
'   Step       - The type of change made. See enum for details
'                  0         = Clear/reset the Undo buffer
'                  Undo_Undo = Go back one step
'                  Undo_Redo = Go forwards one step
'                  (other)   = Log a change made to the model
'



' This var records whether a new Undo step needs be created in memory or not
Dim ReCache As Boolean


Dim FetchIndex As Long

Dim MenuStep As Long, MenuItem As Integer
'Debug.Assert Step = 0

Select Case Step
Case 0 ' Clear/reset the Undo buffer
  
  Sys.MaxUndoStep = 0
  Sys.CurUndoStep = 0
    
  ' Disable the Undo/Redo commands
  MenuStep = 0
  MenuItem = -1
  GoSub SetMenu
  Exit Sub
  
Case Undo_Undo ' Perform an "Undo" step.
  
  If Sys.CurUndoStep = 0 Then
    MenuStep = 0
    MenuItem = 0
    GoSub SetMenu
    Exit Sub
  End If
  
  ' If we haven't undone anything yet, then stash a copy of the PIE data
  ' in Sys.Backups(MAX_UNDO) so that we can Redo up to there.
  '
  If Sys.CurUndoStep = Sys.MaxUndoStep Then
    FetchIndex = MAX_UNDO
    GoSub StoreUndoInfo
  End If
  
  ' Backpedal one Undo step.
  Sys.CurUndoStep = Sys.CurUndoStep - 1
  
  ' Rewind the array ptr, then fetch the PIE model's status
  Sys.BackupPtr = IIf(Sys.BackupPtr, Sys.BackupPtr - 1, MAX_UNDO - 1)
  FetchIndex = Sys.BackupPtr
  GoSub FetchUndoInfo
  
  ' Set the "Undo" menu command if there's another step available
  If Sys.CurUndoStep Then
    FetchIndex = IIf(Sys.BackupPtr, Sys.BackupPtr - 1, MAX_UNDO - 1)
    MenuStep = Sys.Backups(FetchIndex).Action
  Else
    MenuStep = 0
  End If
  MenuItem = 0
  GoSub SetMenu
  
  ' Set the "Redo" command to reflect this action
  MenuItem = 1
  MenuStep = Sys.Backups(Sys.BackupPtr).Action
  GoSub SetMenu
  
Case Undo_Redo ' Perform a "Redo" step, IF APPLICABLE.

  ' Go forwards one Undo step.  Sys.Backups(MAX_UNDO) stores
  ' the status of the model before we used the Undo command in the first place,
  ' meaning that we can undo an Undo.
  '
  If Sys.CurUndoStep = Sys.MaxUndoStep Then
    MenuStep = 0
    MenuItem = 0
    GoSub SetMenu
    Exit Sub
  End If
  
  ' Go forwards in the Undo history.
  Sys.CurUndoStep = Sys.CurUndoStep + 1
  
  ' Rotate the array ptr forwards one step also.
  Sys.BackupPtr = IIf(Sys.BackupPtr = MAX_UNDO - 1, 0, Sys.BackupPtr + 1)
  
  ' Fetch the PIE model's status.  Note that when we arrive
  ' at the front of the Undo history, we fetch the PIE data
  ' from Sys.Backups(MAX_UNDO) instead of the array ptr.
  '
  FetchIndex = IIf(Sys.CurUndoStep = Sys.MaxUndoStep, MAX_UNDO, Sys.BackupPtr)
  GoSub FetchUndoInfo
  
  ' Set the "Redo" menu command if there's another Redo available
  If Sys.CurUndoStep < Sys.MaxUndoStep Then
    MenuStep = Sys.Backups(FetchIndex).Action
  Else
    MenuStep = 0
  End If
  MenuItem = 1
  GoSub SetMenu
  
  ' Set the "Undo" command to reflect this action
  FetchIndex = IIf(Sys.BackupPtr, Sys.BackupPtr - 1, MAX_UNDO - 1)
  MenuStep = Sys.Backups(FetchIndex).Action
  MenuItem = 0
  GoSub SetMenu
  
Case Undo_ChangeSel
  
  ' User selected/de-selected data in the PIE file.
  '
  ' Not a change in itself, but it flags any
  ' previous change as having been completed, forcing
  ' a re-cache of the Undo buffer the next time a change is made.
  
  If Sys.CurUndoStep Then
    FetchIndex = IIf(Sys.BackupPtr, Sys.BackupPtr - 1, MAX_UNDO - 1)
    Sys.Backups(FetchIndex).Done = True
  End If
  
Case Else

  ' Set the "Undo" command appropriately
  MenuStep = Step
  MenuItem = 0
  GoSub SetMenu
  
  ' Set "Can't Redo"
  MenuStep = 0
  MenuItem = 1
  GoSub SetMenu
  
  ' Determine if we need to recache the PIE data,
  ' by comparing this to the previous edit.
  '
  ReCache = True
  If Sys.CurUndoStep Then
    
    FetchIndex = IIf(Sys.BackupPtr, Sys.BackupPtr - 1, MAX_UNDO - 1)
    If Step = Sys.Backups(FetchIndex).Action Then
      ReCache = Sys.Backups(FetchIndex).Done
    End If
  
  End If
    
  ' Determine if we need to re-cache the PIE data for this step.
  '
  ' Action (Step) codes are grouped according to their criteria:
  '   - Always recache
  '   - Don't recache if still on the same polygon
  '   - Don't recache until new selection is made
  '
  ' Refer to the enum for which actions are in which group.
  
  Select Case Step
  Case Undo_FirstSameSel To Undo_LastSameSel
    '
    ' All actions continuous for the same selection go here.
    '
    ' Changing the selection automatically flags a recache as necessary,
    ' so nothing special needs to be done here.
    '
  
  Case Undo_FirstReCache To Undo_LastReCache
    '
    ' All non-continuous (always re-cache) actions go here.
    '
    ReCache = True
  
  Case Undo_FirstSamePoly To Undo_LastSamePoly
    '
    ' All actions continuous for the same polygon go here.
    '
    If Sys.CurUndoStep Then
      If Sys.Backups(FetchIndex).CurPolygon <> ThisPoly Then ReCache = True
    End If
    
  Case Undo_EditPt
    ' Edited point coords.  Continuous for same point only
    '
    If Sys.CurUndoStep Then
      If ThisPoint <> Sys.Backups(FetchIndex).CurPoint Then ReCache = True
    End If
    
  Case Undo_EditCc
    ' Edited connector coords.  Continuous for same cc only
    '
    If Sys.CurUndoStep Then
      If ThisCc <> Sys.Backups(FetchIndex).CurCc Then ReCache = True
    End If
    
  Case Else ' Other/undefined steps do this
    Debug.Assert False
    
    ReCache = True
    
  End Select

  ' Cache new undo buffer data for changes made?
  If ReCache Then
    
    ' Increment the # defined Undo steps if possible.
    '
    ' * Sys.Backups(MAX_UNDO) is used exclusively for Undo/Redo actions.
    '   Other changes simply cannot be allowed to use it.
    '
    Dim N As Long
    N = Sys.CurUndoStep
    If N < MAX_UNDO Then
      
      ' Point to Sys.Backups(Sys.CurUndoStep)
'      FetchIndex = Sys.BackupPtr
      
      ' Increment the # valid steps in memory
      N = N + 1
      Sys.CurUndoStep = N
      Sys.MaxUndoStep = N
      
    End If
    
    ' Store the Undo data...
    FetchIndex = Sys.BackupPtr
    GoSub StoreUndoInfo
    
    ' ...then rotate the array ptr forwards
    Sys.BackupPtr = IIf(Sys.BackupPtr = MAX_UNDO - 1, 0, Sys.BackupPtr + 1)
    
  End If
  
  ' Allow continuous operations for a short time
  ' (assume that the operation is complete if no change is made for awhile)
  formMaster.UndoTimer = 7
  
End Select

Debug.Print "Undo buffer= "; Sys.CurUndoStep & "/" & Sys.MaxUndoStep & " Ptr=" & Sys.BackupPtr

Exit Sub

' StoreUndoInfo: Stores the current PIE file into the Undo buffer.
'
' * Set FetchIndex before calling
' * Normally, FetchIndex must be < MAX_UNDO
'
StoreUndoInfo:

  Debug.Assert Step <> Undo_Undo Imp FetchIndex < MAX_UNDO
  
  With Sys.Backups(FetchIndex)
    .PieStatus = Data
    .CurFile = ThisFile
    .CurPoint = ThisPoint
    .CurCc = ThisCc
    .CurPolygon = ThisPoly
  
    .Action = Step
    .Done = 0
  End With
Return

' FetchUndoInfo: Restores the current PIE file to how it was in the Undo buffer.
'
' Used by both the Undo and Redo commands.
'
' * Set FetchIndex before calling
' * Normally, FetchIndex must be < MAX_UNDO
'
FetchUndoInfo:

'Debug.Assert Step <> Undo_Redo Imp FetchIndex < Sys.MaxUndoStep

With Sys.Backups(FetchIndex)
  ' Recall PIE file before the last edit
  '
  Model = .PieStatus
 
  .Done = True
  
  ' Recall other settings
  '
  SetThisFile .CurFile, 0
  
  formLists.SetCurPoint .CurPoint, 0
  formLists.SetCurPoly .CurPolygon, 0
  formLists.SetCurCon .CurCc
  
  ' Indicate that this was another change to the file
  SetChanged Model, ThisFile
End With
Return

'End Sub

' Inline sub - sets the Undo or Redo menu item as applicable
'
' Before calling:
'   MenuStep = the action chosen (0 for "Can't Undo/Redo")
'   MenuItem = menu index to effect, 0 for Undo and 1 for Redo
'                * If MenuStep=0, can also be -1 for both
SetMenu:
  
  Dim Stri As String
  Select Case MenuStep
  Case 0 ' Disable the Undo/Redo commands.
    If MenuItem = 1 Or MenuItem = -1 Then
      FormMain.mnuUndo(1).Caption = "Can't Redo"
      FormMain.mnuUndo(1).Enabled = False
    End If
    
    If MenuItem = 0 Or MenuItem = -1 Then
      FormMain.mnuUndo(0).Caption = "Can't Undo"
      FormMain.mnuUndo(0).Enabled = False
    End If
    
    Return
    
  Case Undo_EditPt: Stri = "Edit Point"
  Case Undo_EditCc: Stri = "Edit Connector"
  Case Undo_EditPy: Stri = "Edit Polygon Flags"
  Case Undo_CopyPyFlags: Stri = "Copy Polygon Flags"
  Case Undo_CopyPyAni: Stri = "Copy Team Colors"
  Case Undo_MovePts: Stri = "Move Selection"
  Case Undo_AddPt: Stri = "New Point(s)"
  Case Undo_DelPt: Stri = "Delete Point(s)"
  Case Undo_DelPts, Undo_DelPys
     Stri = "Delete Selection"
  Case Undo_AddCc: Stri = "New Connector"
  Case Undo_DelCc: Stri = "Delete Connector"
  Case Undo_AddPy: Stri = "New Polygon"
  Case Undo_DelPy: Stri = "Delete Polygon(s)"
  Case Undo_AttachPt: Stri = "Attach Points"
  Case Undo_AttachPts: Stri = "Attach All Points"
  Case Undo_DoExtrude: Stri = "New Extrusion"
  Case Undo_ReorderPy: Stri = "Polygon Reordering"
  Case Undo_SetTriFan: Stri = "Set Vertex #1"
  Case Undo_AddVertex: Stri = "New Vertex(es)"
  Case Undo_MoveVertex: Stri = "Move Vertex(es)"
  Case Undo_DetachVertex: Stri = "Detach Vertex(es)"
  Case Undo_KillVertex: Stri = "Remove Vertex(es)"
  Case Undo_SlicePoly: Stri = "Slice Polygon"
  Case Undo_ShatterPoly: Stri = "Shatter Polygon"
  Case Undo_JoinPolys: Stri = "Join Polygons"
  Case Undo_ScaleSel: Stri = "Scale Selection"
  Case Undo_RotateSel: Stri = "Rotate Selection"
  Case Undo_Mirror: Stri = "Mirror Polygons"
  Case Undo_ChangeSub: Stri = "Assign Submodel"
  Case Undo_PasteData: Stri = "Paste"
  Case Else: Step = 0
  End Select
  
  Dim Item As Menu
  Debug.Assert MenuItem = 0 Or MenuItem = 1
  
  Set Item = FormMain.mnuUndo(MenuItem)
  Item.Caption = IIf(MenuItem, "Redo ", "Undo ") & Stri
  Item.Enabled = True
  
Return

End Sub


Public Function v_tempVector(ByVal X As Double, ByVal Y As Double, ByVal Z As Double) As D3DVECTOR
'
' Creates and returns a D3DVECTOR.
'
' Used primarily for creating temp vectors,
' i.e. those that are used only once and don't
' need permanent storage in the routine.
'
With v_tempVector
  .X = X
  .Y = Y
  .Z = Z
End With
    
End Function

Private Function file_SavePCK_FileData(File As Integer, ByRef Data As PieType, Optional AniFile As Boolean = False) As Boolean
'
' Packages PIE data into an open PCK file while saving.
'

' Output the FileData flag, as well as the texpage specified by the file
'
Print #File, IIf(AniFile, "// ", "") & "FileData=" & VerPCK_Pie
Print #File, IIf(AniFile, "// ", "") & "{Tex:" & Data.TexPage & "}"

Dim Stri As String
Dim MaxLevs As Integer, Lv As Integer
Dim MaxPt As Integer, Pt As Integer
Dim MaxPy As Integer, Py As Integer
Dim Coord As D3DVECTOR

MaxLevs = enumMaxLevels(Data.Levs)
'For Lv% = 1 To MaxLevs
With Data.Levs
  
  ' A submodels header is needed only when there is more than one
  '
  ' Revised PCK_PIE v110: submodel data now integrated with polygons array
  '
  If MaxLevs > 1 Then
'    If Lv = 1 Then
'      Print #File, "{Lv:" & MaxLevs & "}"
'    Else
      Print #File, "{Lv}"
'    End If
  End If
  
  ' Write # of points in submodel.
  MaxPt = UBound(.Points)
  Print #File, IIf(AniFile, "// ", "") & "{Pts:" & Hex$(MaxPt) & "}"
  
  ' Write the point data
  For Pt = 1 To MaxPt
  With .Points(Pt)
    
    Coord = .Loc
    GoSub ProcessLoc
    
    Print #File, IIf(AniFile, "// ", "") & "{" & Stri & "}"
    
  End With
  Next Pt%

  ' Output # of polygons in submodel
  '
  ' v111: Now stored as hex not decimal
  '
  MaxPy = UBound(.Polys)
  Print #File, IIf(AniFile, "// ", "") & "{Pys:" & Hex$(MaxPy) & "}"
  
  ' Output the polygon data
  Dim Poly As PolygonData
  For Py = 1 To MaxPy
  
    Poly = .Polys(Py)
  
    ' Bugfix: Output team-colors flag only if there are 2+ color frames
    '
    If (Poly.Flags And &H4000&) And (Poly.TeamC.Animated < 2) Then
      Poly.Flags = Poly.Flags And Not &H4000&
    End If
    
    ' Format the submodel (PCK_PIE v110) & type flags
    '
    Stri = Poly.Submodel & ":" & Hex$(Poly.Flags \ 256)
    
    ' Format team-color data if necessary (numframes,time,width,height)
    '
    If Poly.Flags And &H4000& Then
    With Poly.TeamC
      
      If .AnimTime = 0 Then .AnimTime = 1
      If .AnimW < 1 Then .AnimW = 1
      If .AnimH < 1 Then .AnimH = 1
      Stri = Stri & ":" & Hex$(.Animated) & ":" & Hex$(.AnimTime) & ":" & _
                                  Hex$(.AnimW) & ":" & Hex$(.AnimH)
    End With
    End If
        
    ' Then format the vertex data (index,tx,ty, index,tx,ty, etc.)
    '
    For Pt = 1 To UBound(Poly.PtData)
    With Poly.PtData(Pt)
        
        Stri = Stri & ":" & Hex$(.Index)
        Dim X As Integer, Y As Integer
        X = Int(.Loc.X + 0.5)
        Y = Int(.Loc.Y + 0.5)
        If X < 0 Then Stri = Stri & ":-" & Hex$(-X) Else Stri = Stri & ":" & Hex$(X)
        If Y < 0 Then Stri = Stri & ":-" & Hex$(-Y) Else Stri = Stri & ":" & Hex$(Y)
        
    End With
    Next Pt
    
    ' Finally, output the formatted polygon data
    '
    Print #File, IIf(AniFile, "// ", "") & "{" & Stri & "}"
  Next Py
  
  ' Write connectors data (if any).
  '
  ' All connectors data will be formatted into one line
  '
  Dim DataLine As String: DataLine = "{Cc"
  
  For Pt = 1 To UBound(.Ccs)
  With .Ccs(Pt)
      
    Coord = .Loc
    GoSub ProcessLoc
    
    DataLine = DataLine & ":" & Stri
    
  End With
  Next Pt

  If UBound(.Ccs) Then Print #File, IIf(AniFile, "//", "") & DataLine & "}"

End With
'Next Lv%

' Output a blank marker to identify the end of the embedded Pie data
'
Print #File, IIf(AniFile, "// ", "") & "{}"

Exit Function

' Inline subfunction - Format Coord data into output string, Stri.
'
ProcessLoc:
  Dim V As Long
  V = Int(Coord.X + 0.5)
  If V < 0 Then Stri = "-" & Hex$(-V) Else Stri = Hex$(V)
  
  V = Int(Coord.Y + 0.5)
  If V < 0 Then Stri = Stri & ":-" & Hex$(-V) Else Stri = Stri & ":" & Hex$(V)
  
  V = Int(Coord.Z + 0.5)
  If V < 0 Then Stri = Stri & ":-" & Hex$(-V) Else Stri = Stri & ":" & Hex$(V)
Return

End Function


Private Function file_LoadPCK_FileData(File As Integer, ByVal Ver As Integer, ByRef Data As PieType) As Integer
'
' Interprets packaged PIE data found while loading a PCK file.
'
' Returns the # submodels successfully loaded.
'
Syntax:
'
' file_LoadPCK_FileData(File, Ver, Data)
'   File  - File handle to read from.
'   Ver   - Version of embedded PIE data.
'   Data  - Place to store the PIE data.
'
'
' Other vars
Dim MaxSubs As Integer, CurSub As Integer
Dim MaxPts As Integer, PtID As Integer, Coord As D3DVECTOR
Dim maxPys As Integer, PyID As Integer, Poly As PolygonData
Dim CcID As Integer
Dim N As Integer

' Temporary var to hold the current submodel being loaded
Dim TempLev As LevelType
ReDim TempLev.Points(0): PtID = 0
ReDim TempLev.Polys(0): PyID = 0
ReDim TempLev.Ccs(0): CcID = 0

' Section ID markers
Dim Section As Long
Const Sec_Pts As Long = 1
Const Sec_Pys As Long = 2

On Error GoTo BadData

' Assume one submodel by default
'
'ReDim Data.Levs(1)
CurSub = 1

Dim LineData As String, Terms() As String
Do
  
  ' Input a line of data.
  '
  Line Input #File, LineData
  If Left$(LineData, 2) = "//" Then LineData = Mid$(LineData, 3)
  
  Debug.Print LineData
  
  ' Data lines must be contained in braces, or else...
  '
  If Left$(LineData, 1) <> "{" Or Right$(LineData, 1) <> "}" Then
    
    file_LoadPCK_FileData = FE_LoadPckPie Or FE_BadPckSrcFile
    ReDim Data.Levs.Points(0)
    ReDim Data.Levs.Polys(0)
    ReDim Data.Levs.Ccs(0)
    Exit Do
  End If
  
  ' End-of-data marker: Finish up
  '
  If LineData = "{}" Then Exit Do
  
  ' Strip the braces, then split it into terms
  '
  Terms = Split(Mid$(LineData, 2, Len(LineData) - 2), ":")

  ' Analyze the first term to see what to do with the data.
  '
  Select Case Terms(0)
  Case "Tex"
    ' Texpage identification
    Data.TexPage = Terms(1)
    
  Case "Lv"
    ' Submodels declaration/assignation
    '
    ' Data line format:
    ' {Lv:n} - set maximum submodels and begin submodel #1
    ' {Lv}   - begin next submodel
    '
    If UBound(Terms) Then
      ' {Lv:n} format - old
      '
      MaxSubs = CInt("&H" & Terms(1))
'      ReDim Data.Levs(MaxSubs)
    Else
      ' {Lv} format - switch to next submodel
      '
      CurSub = CurSub + 1
      
    End If
    
    ' Empty the reference vars
    '
    Section = 0
'    ReDim TempLev.Points(0): PtID = 0
'    ReDim TempLev.Polys(0): PyID = 0
'    ReDim TempLev.Ccs(0)
    
  Case "Pts"
    ' Points declaration
    '
    ' Data line format:
    ' {Pt:n} - Specify # of points
    '
    ' Subsequent untagged lines will be interpreted as point data
    '
    
    MaxPts = CInt("&H" & Terms(1))
    N = UBound(TempLev.Points)
    ReDim TempLev.Points(N + MaxPts)
    
    Section = Sec_Pts
  
  Case "Pys"
    ' Polygons declaration
    '
    ' Data line format:
    ' {Pys:n} - Specify # polygons
    '
    ' Subsequent untagged lines will be interpreted as polygon data
    '
    
    If Ver > 110 Then
      maxPys = CInt("&H" & Terms(1))
    Else
      maxPys = Terms(1)
    End If
    
    ReDim TempLev.Polys(maxPys)
    
    Section = Sec_Pys
    
  Case "Cc"
    ' Connectors declaration
    '
    ' Data line format:
    ' {Cc:x,y,z...} - Coordinates for one or more connectors
    '
    ' Multiple connectors can be specified in one line
    '
    N = 1
    Do While N <= UBound(Terms) - 2
      CcID = CcID + 1
      ReDim Preserve TempLev.Ccs(CcID)
        
      ' Parse x,y,z values into Coord
      GoSub ReadCoord
        
      ' Apply the connector data
      TempLev.Ccs(CcID).Loc = Coord
      
      N = N + 3
    Loop
  
  Case Else
    ' Numerical data - process as follows
    '
    Select Case Section
    Case Sec_Pts
      ' Point data lines
      '
      ' Data line format:
      ' {x,y,z...} - Coordinates for one or more points (3 terms per point)
      '
      
      ' Multiple points may be declared in one line, if necessary
      '
      N = 0
      Do While N <= UBound(Terms) - 2
        
        PtID = PtID + 1
        If PtID > MaxPts Then ReDim Preserve TempLev.Points(PtID)
        
        ' Parse coordinate data
        GoSub ReadCoord
        
        ' Apply the point data
        TempLev.Points(PtID).Loc = Coord
        
        ' Increment to next set, if any
        N = N + 3
      Loop
    
    Case Sec_Pys
      ' Polygon data lines
      '
      ' Data line format:
      '   VerPCK_Pie=0    - {Py:f:d:d:d:d...}
      '   VerPCK_Pie=110+ - {Py:s:f:d:d:d:d....}
      '
      '  s - Polygon submodel
      '  f - Polygon flags (multiply by &H100)
      '  d - Polygon data.
      '        * 4 terms first if flag &H4000 is set, then
      '          3 terms for each vertex (index, tx, ty)
      
      ' Allow for new submodel management
      If Ver >= 110 Then
        Poly.Submodel = Terms(0)
        N = 1
      Else
        Poly.Submodel = CurSub
        N = 0
      End If
      
      ' Buffer 10 vertexes for this poly - we'll crop later
      ReDim Poly.PtData(10)
    
      ' Apply polygon flags
      Poly.Flags = CLng("&H" & Terms(N)) * &H100&
      N = N + 1
      
      ' Parse team-color options, if supplied
      If Poly.Flags And &H4000& Then
        Poly.TeamC.Animated = CInt("&H" & Terms(N))
        Poly.TeamC.AnimTime = CInt("&H" & Terms(N + 1))
        Poly.TeamC.AnimW = CInt("&H" & Terms(N + 2))
        Poly.TeamC.AnimH = CInt("&H" & Terms(N + 3))
        
        N = N + 4
'      Else
'        N = N + 1
      End If
      
      ' Parse the vertex data.
      ' One vertex for every 3 terms will be created,
      ' up to 10 vertexes for one polygon
      '
      Dim V As Integer
      V = 0
      Do While N <= UBound(Terms) - 2
        V = V + 1
        
        ' Parse three terms into Coord
        GoSub ReadCoord
        
        Debug.Assert Coord.X And Coord.X <= MaxPts
        
        ' Apply the vertex data
        With Poly.PtData(V)
          .Index = Coord.X ' Terms(N)
          .Loc.X = Coord.Y ' Terms(N+1)
          .Loc.Y = Coord.Z ' Terms(N+2)
        End With
        
        If V >= 10 Then Exit Do
        
        N = N + 3
      Loop
      
      ' Crop the polygon's vertex array to fit the loaded data
      ReDim Preserve Poly.PtData(V)
      
      ' Store the polygon data in the submodel
      ' if 3+ vertexes have been loaded
      '
      If V > 2 Then
        PyID = PyID + 1
        If PyID > maxPys Then Debug.Assert False: ReDim Preserve TempLev.Polys(PyID)
        TempLev.Polys(PyID) = Poly
        
      End If
      
    Case Else
      Debug.Assert False
    
    End Select
  End Select

Loop

' Array housekeeping
If PyID < maxPys Then Debug.Assert False: ReDim Preserve TempLev.Polys(PyID)

' Finish up and return to file_LoadPCK()
'
Data.Levs = TempLev

'If CurSub < UBound(Data.Levs) Then ReDim Preserve Data.Levs(CurSub)

' Mark the data as unchanged
'
Data.Loaded = Timer
Data.Changed = Data.Loaded

file_LoadPCK_FileData = 1
Exit Function

'
' Inline subfunctions begin here
'

' ReadCoord:  Parses 3 terms into Coord variable.
'
' Each term is hexadecimal and may or not include a negative sign.
'
ReadCoord:
  ' Parse x coordinate.
  If Left$(Terms(N), 1) = "-" Then
    Coord.X = -CLng("&H" & Mid$(Terms(N), 2))
  Else
    Coord.X = CLng("&H" & Terms(N))
  End If
  
  ' Parse y coordinate.
  If Left$(Terms(N + 1), 1) = "-" Then
    Coord.Y = -CLng("&H" & Mid$(Terms(N + 1), 2))
  Else
    Coord.Y = CLng("&H" & Terms(N + 1))
  End If
  
  ' Parse z coordinate.
  If Left$(Terms(N + 2), 1) = "-" Then
    Coord.Z = -CLng("&H" & Mid$(Terms(N + 2), 2))
  Else
    Coord.Z = CLng("&H" & Terms(N + 2))
  End If
Return

' ApplySub: Crops empty space (if any) from the temp submodel
' and applies it to the temp model data
'
' Called before creating a new submodel and when finished loading
'
ApplySub:
  If CurSub < 1 Then Return
  
  ' Crop point/polygon data to fit if necessary
  '
  If PtID < UBound(TempLev.Points) Then ReDim Preserve TempLev.Points(PtID)
  If PyID < UBound(TempLev.Polys) Then ReDim Preserve TempLev.Polys(PyID)
Return

BadData:
' Error-trapping routine
'
Debug.Assert False

On Error GoTo 0
Resume
' Declare no valid data loaded
file_LoadPCK_FileData = FE_LoadPckPie Or FE_BadPckSrcFile

End Function

Private Sub file_BuildPath(ByVal PathFile As String)
'
' Creates necessary directories for a file path.
'
Dim N As Integer
Do
  N = InStr(N + 1, PathFile, "\")
  If N = 0 Then Exit Do
  If Dir(Left$(PathFile, N - 1), vbDirectory) = "" Then
    MkDir Left$(PathFile, N - 1)
  End If
Loop

End Sub

Public Function py_AddToSubmodel(ByRef Lev As LevelType, Index As Integer) As Integer
'
' Assigns polygon data to a specified submodel.
'
Syntax:
'
' py_AssignToSubmodel(Lev, Index)
'   Lev   - Model data to search.  Only selected polygons will be used.
'   Index - Submodel # to assign to.
'             * Usually ranges from 1-10
'             * -1=auto (will be set to the first available submodel)
'
' Function returns the # of polygons assigned, or -1 for error.
'
Dim PtFlags() As Integer, pyFlags As Integer
Dim Count As Integer
Count = enumSubmodels(Lev, PtFlags, pyFlags, 0)

If Index < 1 Then
  If Count > 9 Then
    
    ' Hunt for an empty submodel.
    Dim Mask As Integer
    Mask = stat_Sub1
    
    For Count = 1 To 10
      If pyFlags And Mask Then Index = Count: Exit For
      Mask = Mask * 2
    Next Count
    
    If Index < 1 Then
      py_AddToSubmodel = -1
      Exit Function
    End If
  End If
  Index = Count + 1
End If

Dim Py As Integer

' Iterate through all selected polygons
For Py = 1 To UBound(Lev.Polys)
With Lev.Polys(Py)
If (.StatFlags And stat_SelMask) = stat_Sel Then
  py_AddToSubmodel = py_AddToSubmodel + 1
  
  If .Submodel <> Index Then
    .Submodel = Index
  
    formLists.lstView(1).ListItems(Py).SubItems(1) = .Submodel
    
  End If
  
End If
End With
Next Py

End Function


Public Function sel_ToggleSel(Lev As LevelType, ByVal Query As Long, ByVal Index As Long) As Long
'
' Toggles the select status of various data in a PIE file.
'
' Does not update the GUI or other fancy stuff.
'
Syntax:
'
' sel_ToggleSel(Lev, Query, Index)
'   Lev          - PIE data to use.
'   Query,Index  - Query indicating what data to toggle.
'                    * See enumSelection() for details
'
' Returns the # of items effected/toggled.
'   0 = no data found
'  -1 = Bad calling syntax
'

If (Query And Get_QueryMask) = 0 Then
  Debug.Assert False
  sel_ToggleSel = -1
  Exit Function
End If
 
' Get_AllSel query param cannot be used here.
If (Query And Get_HowMask) = Get_AllSel Then
  Debug.Assert False
  sel_ToggleSel = -1
  Exit Function
End If

Dim Status As Long, DataList() As Integer

' Enumerate selected data.
Status = enumSelection(Lev, Query, Index, DataList)
If Status < 0 Then
  Debug.Assert False
  sel_ToggleSel = Status
  Exit Function
End If

' Check that data is available
If UBound(DataList) = 0 Then Exit Function

Dim NewV As Integer
NewV = IIf(Status, 0, stat_Sel)

' Iterate through the results list
Select Case DataList(0)
Case Get_Pts
  ' Change affected points
  '
  Dim P1 As Integer
  For P1 = 1 To UBound(DataList)
  With Lev.Points(DataList(P1))
    
    .StatFlags = (.StatFlags And stat_NotSelMask) Or NewV
  
  End With
  Next P1

Case Get_Pys
  ' Change affected polys
  '
  For P1 = 1 To UBound(DataList)
  With Lev.Polys(DataList(P1))
    
    .StatFlags = (.StatFlags And stat_NotSelMask) Or NewV
  
  End With
  Next P1

Case Else
  Debug.Assert False
  sel_ToggleSel = -1
  Exit Function
End Select

sel_ToggleSel = UBound(DataList)

End Function


Public Function SetPieOrigin(Data As PieModel, ByVal File As Integer) As Long
'
' Calculates the assembly data for a PIE model.
'
' Should be called whenever the model's assembly has changed.
'
' This includes:
'   - Successful loading of a PCK file from disk
'   - Removing a PIE file from the model
'   - Specifically assigning a PIE file's model type
'
Syntax:
'
' SetPieOrigin(Data, File)
'   Data - PIE model to evaluate.
'   File - File index in that model to evaluate.
'            * 0 or -1 = compute all
'
' No return value.

Dim F As Integer, IsVTOL As Long, TempMat As D3DMATRIX


For F = IIf(File < 1, 1, File) To IIf(File < 1, UBound(Data.Files), File)
With Data.Files(F)

  TempMat = GetRotaMatrix(.PieType, IsVTOL)
  
  .P3DFlags = IIf(IsVTOL And 2, P3D_IsVTOL, 0)
  .PieOrigin.X = TempMat.rc41
  .PieOrigin.Y = TempMat.rc42
  .PieOrigin.Z = TempMat.rc43

End With
Next F

End Function
